;;; xtla-core.el --- Core of xtla

;; Copyright (C) 2003-2004 by Stefan Reichoer

;; Author: Stefan Reichoer, <xsteve@nit.at>
;; Contributions from:
;;    Matthieu Moy <Matthieu.Moy@imag.fr>
;;    Masatake YAMATO <jet@gyve.org>
;;    Milan Zamazal <pdm@zamazal.org>
;;    Martin Pool <mbp@sourcefrog.net>
;;    Robert Widhopf-Fenk <hack@robf.de>
;;    Mark Triggs <mst@dishevelled.net>

;; This file is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This file is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; This file provides the low-level functions used by xtla.el

;;; Code:

(require 'xtla-defs)
(require 'ewoc)

;; ----------------------------------------------------------------------------
;; Compatibility stuff
;; ----------------------------------------------------------------------------
(eval-when-compile
  (require 'cl)
  (if (featurep 'xemacs)
      (load-library "xtla-xemacs")))

(require 'pp)

(defvar tla--buffer-type-alist
  '((changes "*tla-changes*")
    (inventory "*tla-inventory*")
    (missing "*tla-missing*")
    (cat-log "*tla-cat-log*")
    (file-diff "*tla-file-diff*")
    (changelog "*tla-changelog*")
    (tree-lint "*tla-tree-lint*")
    (logs "*tla-logs*")
    (errors "*tla-error*")
    (generic "*tla-process*"))
  "List of pairs (type name) used to generate a name for a buffer")

(defvar tla--buffers-tree nil
  "List in the form
 ((type1 (\"path1\" buffer)
         (\"path2\" buffer2))
  (type2 (\"path1\" buffer3)
         (\"path3\" buffer4)))
Used to keep track of all the tla related buffers.")

(defun tla--buffers-tree-remove (buffer)
  "Removes BUFFER from the buffers tree"
    (dolist (type-cons tla--buffers-tree)
      (dolist (path-buffer (cdr type-cons))
        (when (eq (cadr path-buffer) buffer)
          (setcdr type-cons (delete path-buffer (cdr type-cons)))))))

(defun tla--buffers-tree-add (type path buffer)
  "Adds BUFFER to the buffers tree"
  (let ((current-assoc (assoc type tla--buffers-tree)))
    (if current-assoc
        (setcdr current-assoc
                (cons (list path buffer)
                      (cdr current-assoc)))
      (setq tla--buffers-tree
            (cons `(,type (,path ,buffer)) tla--buffers-tree)))))

(defun tla--uniquify-file-name (path)
  "Returns a unique string designating PATH. If PATH is a directory,
the returned contains one and exactly one trailing slash. If PATH is
nil, then nil is returned."
  (and path
       (let ((expanded (expand-file-name (if (file-directory-p path)
                                             (concat path "/")
                                           path))))
         (if (featurep 'xemacs)
             (replace-regexp-in-string "/+$" "/" expanded)
           expanded))))

(defun tla--get-buffer-create (type &optional path)
  "Gets a buffer of type TYPE for the path PATH. Reuse one if it
exists, otherwise, call `create-file-buffer' to create the buffer.

See also `tla--get-buffer'"
  ;; Inspired from `cvs-get-buffer-create'
  (or (tla--get-buffer type path)
      (let ((path (tla--uniquify-file-name path))
            (name (cadr (assoc type tla--buffer-type-alist))))
        (with-current-buffer
            (let* ((default-directory (or path default-directory))
                   (buffer (create-file-buffer (or name ""))))
              (if (featurep 'xemacs)
                  (save-excursion
                    (set-buffer buffer)
                    (tla--install-buffer-menu)))
              (tla--buffers-tree-add type path buffer)
              buffer)))))

(add-hook 'kill-buffer-hook 'tla--kill-buffer-function)

(defun tla--kill-buffer-function ()
  "Function run when a buffer is killed."
  (tla--buffers-tree-remove (current-buffer))
  (tla--kill-process-maybe (current-buffer)))

(defun tla--get-buffer (type &optional path)
  "Gets a buffer of type TYPE for the path PATH. Reuse one if it
exists, otherwise, return nil. See also `tla--get-buffer-create'."
  (let ((path (and path
                   (tla--uniquify-file-name path))))
    (let* ((list-path (assoc type tla--buffers-tree))
           (buffer (cadr (assoc path
                                (cdr list-path)))))
      (when buffer
        (if (buffer-live-p buffer) buffer
          (setcdr list-path
                  (delq (assoc path (cdr list-path))
                        (cdr list-path)))
          (tla--get-buffer type path))))))

(defun tla--add-buffer-type (type name)
  (unless (assoc type tla--buffer-type-alist)
    (push (list type name) tla--buffer-type-alist)))

(defun tla--last-visited-inventory-buffer ()
  "Return the last visited xtla's inventory buffer."
  (let ((inventories (remove nil (mapcar
                                  (lambda (elt)
                                    (when (buffer-live-p (cadr elt))
                                      elt))
                                  (cdr (assoc 'inventory tla--buffers-tree)))))
        (bl (buffer-list)))
    (cadr (car (sort inventories (lambda (a b)
                                   (let ((aindex (position (cadr a) bl))
                                         (bindex (position (cadr b) bl)))
                                     (< aindex bindex))))))))

(defun tla-show-inventory-buffer ()
  "Switch to the last visited inventory buffer."
  (interactive)
  (tla-switch-to-buffer (tla--last-visited-inventory-buffer)))

;; Lint trap
(eval-when-compile (unless (fboundp 'read-directory-name)
                     (defun read-directory-name (&optional a b c d e))))

(defun tla--read-directory-name (prompt &optional dir default-dirname
                                        mustmatch initial)
  "Read directory name, prompting with PROMPT and completing in directory DIR.
Value is not expanded---you must call `expand-file-name' yourself.
Default name to DEFAULT-DIRNAME if user exits with the same
non-empty string that was inserted by this function.
 (If DEFAULT-DIRNAME is omitted, the current buffer's directory is used,
  except that if INITIAL is specified, that combined with DIR is used.)
If the user exits with an empty minibuffer, this function returns
an empty string.  (This can only happen if the user erased the
pre-inserted contents or if `insert-default-directory' is nil.)
Fourth arg MUSTMATCH non-nil means require existing directory's name.
 Non-nil and non-t means also require confirmation after completion.
Fifth arg INITIAL specifies text to start with.
DIR should be an absolute directory name.  It defaults to
the value of `default-directory'."
  (if (fboundp 'read-directory-name)
      (read-directory-name prompt dir default-dirname mustmatch initial)
    ;; The same as the definition of `read-directory-name'
    ;; in GNU Emacs in CVS.
    (unless dir
      (setq dir default-directory))
    (unless default-dirname
      (setq default-dirname
            (if initial (concat dir initial) default-directory)))
    (read-file-name prompt dir default-dirname mustmatch initial)))

(defun tla--add-to-list (list-var element &optional append)
  "Same behavior as GNU Emacs's add-to-list, but also works on XEmacs."
  (if (featurep 'xemacs)
      (if append
          (when (not (member element (eval list-var)))
	    (set list-var (append (eval list-var) (list element))))
        (add-to-list list-var element))
    (add-to-list list-var element append)))

;; ----------------------------------------------------------------------------
;; Process buffers
;; ----------------------------------------------------------------------------

;;;###autoload
(defcustom tla-process-buffer " *tla-process*"
  "*Name of the process buffer"
  :type 'string
  :group 'xtla-internal)

;;;###autoload
(defcustom tla-error-buffer " *tla-errors*"
  "*Name of the buffer to which tla's stderr is redirected"
  :type 'string
  :group 'xtla-internal)

(defcustom tla-number-of-dead-process-buffer 20
  "Number of process buffers to keep after process termination. When the
number of process buffers exceeds this number, the most ancient
is killed. This includes both the process buffer and the error
buffer (to which stderr is redirected).

A nil value here means \"Never kill any process buffer\". Useful for
debugging, but this will eat the memory of your computer ;-)"
  :type 'integer
  :group 'xtla-internal)

(defcustom tla-show-internal-buffers-on-menu t ; till 1.0
  "Control whether or not buffers in `tla--dead-process-buffer-queue' appear
in the buffer menu."
  :type 'boolean
  :group 'xtla-internal)

(defvar tla--dead-process-buffer-queue nil
  "List of process buffers belonging to terminated processes. When the
list is greater than `tla-number-of-dead-process-buffer', the last
ones are killed.")

(defun tla--kill-process-buffer (buffer)
  "Don't actually kill BUFFER, but add it to
`tla--dead-process-buffer-queue'. It will eventually be killed
when the number of buffers in `tla--dead-process-buffer-queue'
exceeds `tla-number-of-dead-process-buffer'."
  (add-to-list 'tla--dead-process-buffer-queue buffer t)
  (when tla-number-of-dead-process-buffer
    (while (> (length tla--dead-process-buffer-queue)
              (max 2 tla-number-of-dead-process-buffer))
      (kill-buffer (car tla--dead-process-buffer-queue))
      (setq tla--dead-process-buffer-queue
            (cdr tla--dead-process-buffer-queue)))))

(defvar tla--last-process-buffer nil
  "Last created process-buffer")

(defvar tla--last-error-buffer nil
  "Last created process-buffer")

(defun tla--new-process-buffer (to-be-deleted)
  "Creates a new process buffer"
  (let ((buffer (create-file-buffer tla-process-buffer)))
    (setq tla--last-process-buffer buffer)
    (when to-be-deleted (tla--kill-process-buffer buffer))
    buffer))

(defun tla--new-error-buffer (to-be-deleted)
  "Creates a new error buffer"
  (let ((buffer (create-file-buffer tla-error-buffer)))
    (setq tla--last-error-buffer buffer)
    (when to-be-deleted (tla--kill-process-buffer buffer))
    buffer))

;; ----------------------------------------------------------------------------
;; Process management
;; ----------------------------------------------------------------------------

;; Candidates for process handlers
(defun tla--default-error-function (output error status arguments)
  "Default function called when a tla process ends with a non-zero
status."
  (tla--show-error-buffer error)
  (error "`tla %s' failed with code %d"
         (mapconcat 'identity arguments " ")
         status))

(defvar tla--default-killed-function-noerror 0
  "If > 0, `tla--default-killed-function' won't make any error next
time, and will decrement it by one.")

(defun tla--default-killed-function (output error status arguments)
  "Default function called when a tla process is killed."
  (if (> tla--default-killed-function-noerror 0)
      (setq tla--default-killed-function-noerror
            (- tla--default-killed-function-noerror 1))
    (tla-switch-to-buffer error)
    (error "`tla %s' process killed !"
           (mapconcat 'identity arguments " "))))

(defun tla--null-handler (output error status arguments)
  "Function doing nothing. Candidate as an argument for one of the
keywords :finished, :error or :killed in `tla--run-tla-sync' or
`tla--run-tla-async'."
  nil
  )

(defun tla--status-handler (output error status arguments)
  "returns an integer value represented by STATUS. Candidate as an
argument for one of the keywords :finished, :error or :killed in
`tla--run-tla-sync' or `tla--run-tla-async'."
  (cond ((numberp status) status)
        ((string-match "^exited abnormally with code \\(.*\\)" status)
         (string-to-int (match-string 1)))
        (t (error status))))

(defun tla--output-buffer-handler (output error status arguments)
  "returns the content of the buffer OUTPUT, stripping the final
newline if necessary."
  (tla--buffer-content output))

(defun tla--output-buffer-split-handler (output error status arguments)
  "returns the content of the buffer OUTPUT as a list of lines."
  (split-string (tla--buffer-content output) "\n"))

(defun tla--default-finish-function (output error status arguments)
  "Default function called when a tla process terminates."
  (with-current-buffer output
    (tla-process-buffer-mode))
  (tla-switch-to-buffer output)
  (message "`tla %s' process finished !"
           (mapconcat 'identity arguments " "))
  status)

(defun tla--finish-function-without-buffer-switch (output error status arguments)
  "Similar to `tla--default-finish-function' but no buffer switch."
  (message "`tla %s' process finished !"
           (mapconcat 'identity arguments " "))
  status)

(defvar tla--log-cookie nil)

(defstruct (tla--event) output-buffer error-buffer related-buffer
  command tree event time)

(defun tla--strip-final-newline (string)
  "Strip the final newline from string if there's one"
  (if (eq (aref string (- (length string) 1)) ?\n)
      (substring string 0 (- (length string) 1))
    string))

(defun tla--log-printer (elem)
  "tla event printer"
  (insert "Command: " (tla--event-command elem)
          "\nDirectory: " (tla--event-tree elem)
          "\nDate: " (format-time-string "%c" (tla--event-time elem))
          "\nBuffer: " (or (buffer-name (tla--event-related-buffer elem))
                           (pp-to-string (tla--event-related-buffer elem))))
  (when (not (string= (tla--event-event elem)
                      "started"))
    (insert "\nEvent: " (tla--event-event elem)))
  (insert "\n"))

(defmacro tla--switch-to-buffer-macro (function accessor)
  `(defun ,function ()
     "In a log buffer, pops to the output or error buffer corresponding to the
process at point"
     (interactive)
     (let ((buffer (,accessor
                    (ewoc-data (ewoc-locate tla--log-cookie)))))
       (if (buffer-live-p buffer)
           (tla-switch-to-buffer buffer)
         (error "Buffer has been killed")))))

(tla--switch-to-buffer-macro tla-switch-to-output-buffer
                             tla--event-output-buffer)

(tla--switch-to-buffer-macro tla-switch-to-error-buffer
                             tla--event-error-buffer)

(tla--switch-to-buffer-macro tla-switch-to-related-buffer
                             tla--event-related-buffer)

(defun tla--log-event (output error command tree event)
  "Logs an event in the `tla-log-buffer' buffer. Returns that event."
  (unless (and tla--log-cookie
               (buffer-live-p (ewoc-buffer tla--log-cookie)))
    (with-current-buffer (get-buffer-create tla-log-buffer)
      (setq tla--log-cookie
            (ewoc-create 'tla--log-printer))
      (tla-log-buffer-mode)))
  (let ((related-buffer (current-buffer)))
    (with-current-buffer (ewoc-buffer tla--log-cookie)
      (let ((elem (make-tla--event :output-buffer output
                                   :error-buffer error
                                   :related-buffer related-buffer
                                   :command command
                                   :tree tree
                                   :event event
                                   :time (current-time)))
            buffer-read-only)
        (ewoc-enter-last tla--log-cookie elem)
        (ewoc-refresh tla--log-cookie)
        elem))))

(defun tla-log-next () (interactive)
  (let ((next (ewoc-next tla--log-cookie
                         (ewoc-locate tla--log-cookie))))
    (when next (goto-char (ewoc-location next)))))

(defun tla-log-prev () (interactive)
  (let ((prev (ewoc-prev tla--log-cookie
                         (ewoc-locate tla--log-cookie))))
    (when prev (goto-char (ewoc-location prev)))))

(define-derived-mode tla-log-buffer-mode fundamental-mode "Xtla Log"
  "Major mode for Xtla's internal log buffer. You can open this buffer
with `tla-open-internal-log-buffer'."
  (toggle-read-only 1))

(define-derived-mode tla-process-buffer-mode fundamental-mode
  "Xtla Process"
  "Major mode for process buffers. Mainly defines \\[bury-buffer]
to quit the buffer"
  (tla--install-buffer-menu)
  (toggle-read-only 1))

(defmacro tla--with-keywords (keywords plist &rest body)
  (flet ((keyword-to-symbol (keyword)
                            (intern (substring (symbol-name keyword) 1))))
    (let ((keyword (make-symbol "keyword"))
          (default (make-symbol "default")))
      `(let ,(mapcar (lambda (keyword-entry)
                       (keyword-to-symbol (if (consp keyword-entry)
                                              (car keyword-entry)
                                            keyword-entry)))
                     keywords)
         (dolist (keyword-entry ',keywords)
           (let ((,keyword (if (consp keyword-entry)
                               (car keyword-entry)
                             keyword-entry))
                 (,default (if (consp keyword-entry)
                               (cadr keyword-entry)
                             nil)))
             (set (intern (substring (symbol-name ,keyword) 1))
                  (or (cadr (member ,keyword ,plist))
                      ,default))))
         ,@body))))
(put 'tla--with-keywords 'lisp-indent-function 1)

(defun tla--build-tla-command (list-args)
  "Builds a string representing the shell command to execute to run
tla with args LIST-ARGS"
  (mapconcat 'shell-quote-argument
             (cons tla-tla-executable
                   (delq nil list-args))
             " "))

(defun tla--run-tla-async (arguments &rest keys)
  "Run tla asynchronously. ARGUMENTS is a list of arguments. nil
values in this list are removed. KEYS is a list of keywords and
values. Possible keywords are

 :finished ....... Function run when the process finishes. If none
                   specified, `tla--default-finish-function' is run.

 :killed ......... Function run when the process is killed. If none
                   specified, `tla--default-killed-function' is run.

 :error .......... Function run when the process exits with a non 0
                   status. If none specified,
                   `tla--default-error-function' is run.

All these functions take 4 arguments : output, error, status, and
arguments.

   - \"output\" is the output buffer
   - \"error\" is the buffer where standard error is redirected
   - \"status\" is the numeric exit-status or the signal number
   - \"arguments\" is the list of arguments, as a list of strings,
              like '(\"changes\" \"--diffs\")

   `tla--null-handler' can be used here if there's nothing to do.

 :output-buffer .. Buffer where the output of the process should be
                   redirected. If none specified, a new one is
                   created, and will be entered in
                   `tla--dead-process-buffer-queue' to be killed
                   later.

 :error-buffer ... Buffer where the standard error of the process
                   should be redirected.

 :related-buffer . Defaults to `current-buffer'. This is the buffer
                   where the result of the process will be used. If
                   this buffer is killed before the end of the
                   execution, the user is prompted if he wants to kill
                   the process.

Example:
  (tla--run-tla-async `(\"changes\" ,(unless arg \"--diffs\"))
                      :finished
                      (lambda (output error status arguments)
                        (message \"No changes in this working copy\"))
                      :error
                      (lambda (output error status arguments)
                        (tla-show-changes-buffer output)))
"
  (tla--with-keywords
      (:finished :killed :error :output-buffer :error-buffer :related-buffer)
    keys
    (let* ((output-buf (or (and output-buffer (get-buffer-create output-buffer))
                           (tla--new-process-buffer nil)))
           (error-buf  (or (and error-buffer (get-buffer-create error-buffer))
                           (tla--new-error-buffer nil)))
           (error-file (make-temp-name "/tmp/arch-errors"))
           (command (tla--build-tla-command arguments))
           ;; Make the `default-directory' unique. The trailing slash
           ;; may be necessary in some cases.
           (default-directory (tla--uniquify-file-name default-directory))
           (process (start-process
                     tla-tla-executable output-buf
                     "sh" "-c"
                     (format "%s 2> %s"
                             command error-file)))
           (process-event
            (list process
                  (tla--log-event output-buf
                                  error-buf
                                  command
                                  default-directory "started"))))
      (with-current-buffer (or related-buffer (current-buffer))
        (message "running process `%s' in `%s'" command default-directory)
        (add-to-list 'tla-process-running process-event)
        (set-process-sentinel
         process
         `(lambda (process event)
            (let ((default-directory ,default-directory))
              (tla--log-event ,output-buf ,error-buf ,command
                              ,default-directory
                              (tla--strip-final-newline event))
              (setq tla-process-running
                    (delq ',process-event tla-process-running))
              (when (file-exists-p ,error-file)
                (with-current-buffer ,error-buf
                  (insert-file-contents ,error-file))
                (delete-file ,error-file))
              (let ((state (process-status process))
                    (status (process-exit-status process)))
                (unwind-protect
                    (cond ((and (eq state 'exit) (= status 0))
                           (funcall (or (quote ,finished)
                                        'tla--default-finish-function)
                                    ,output-buf ,error-buf
                                    status (quote ,arguments)))
                          ((eq state 'signal)
                           (funcall (or (quote ,killed)
                                        'tla--default-killed-function)
                                    ,output-buf ,error-buf status
                                    (quote ,arguments)))
                          ((eq state 'exit) ;; status != 0
                           (funcall (or (quote ,error)
                                        'tla--default-error-function)
                                    ,output-buf ,error-buf status
                                    (quote ,arguments)))))
                ;; Schedule any buffers we created for killing
                (unless ,output-buffer (tla--kill-process-buffer ,output-buf))
                (unless ,error-buffer (tla--kill-process-buffer ,error-buf))))))
        process))))

(defun tla--run-tla-sync (arguments &rest keys)
  "Run tla synchronously. See `tla--run-tla-async' for details"
  (tla--with-keywords
      (:finished :killed :error :output-buffer :error-buffer :related-buffer)
    keys
    (let ((output-buf (or (and output-buffer (get-buffer-create output-buffer))
                          (tla--new-process-buffer t)))
          (error-buf  (or (and error-buffer (get-buffer-create error-buffer))
                          (tla--new-error-buffer t)))
          (command (tla--build-tla-command arguments))
          (error-file (make-temp-name "/tmp/arch-errors"))
          ;; Make the `default-directory' unique. The trailing slash
          ;; may be necessary in some cases.
          (default-directory (tla--uniquify-file-name default-directory)))
      (with-current-buffer (or related-buffer (current-buffer))
        (tla--log-event output-buf error-buf command default-directory "started")
        (let ((status (call-process "sh" nil output-buf nil "-c"
                                    (format "%s 2> %s"
                                            command
                                            error-file))))
          (when (file-exists-p error-file)
            (with-current-buffer error-buf
              (insert-file-contents error-file))
            (delete-file error-file))
          (unwind-protect
              (cond ((stringp status)
                     (when (string= status "Terminated")
                       (funcall (or killed 'tla--default-killed-function)
                                output-buf error-buf status arguments)))
                    ((numberp status)
                     (if (zerop status)
                         (funcall (or finished 'tla--default-finish-function)
                                  output-buf error-buf status arguments)
                       (funcall (or error 'tla--default-error-function)
                                output-buf error-buf status arguments)))
                    (t (message "Unknown status - %s" status)))
            ;; Schedule any buffers we created for killing
            (unless output-buffer (tla--kill-process-buffer output-buf))
            (unless error-buffer (tla--kill-process-buffer error-buf))))))))

(defvar tla-process-running nil
  "List of tla processes running. nil if none.

The list is a list of pairs (process event) where EVENT is the event
corresponding to the beginning of the execution of process. It can be
used to get more info about the process.")

(defun tla--kill-process-maybe (buffer)
  "Prompts and possibly kill process whose related buffer is BUFFER"
  (let ((process-list nil))
    (dolist (process-buffer tla-process-running)
      (when (eq (tla--event-related-buffer (cadr process-buffer))
                buffer)
        (add-to-list 'process-list (car process-buffer))))
    (let ((l (length process-list)))
      (when (and process-list
                 (y-or-n-p (format "%s process%s running in buffer %s. Kill %s? "
                                   l (if (> l 1) "es" "")
                                   (buffer-name buffer)
                                   (if (> l 1) "Them" "it"))))
        (dolist (process process-list)
          (setq tla--default-killed-function-noerror
                (1+ tla--default-killed-function-noerror))
          (if (eq (process-status process) 'run)
              (kill-process process)))))))

;;;###autoload
(add-to-list 'minor-mode-alist
             '(tla-process-running
               (:eval (if (equal (length tla-process-running) 1)
                          " Tla running"
                        (concat " Tla running("
                                (int-to-string (length tla-process-running))
                                ")")))))

(defun tla-open-internal-log-buffer ()
  "Switch to the Xtla's internal log buffer. This buffer contains a
list of all the tla commands previously executed. The buffer uses the
mode `tla-log-buffer-mode'"
  (interactive)
  (tla-switch-to-buffer tla-log-buffer))

(defun tla--buffer-content (buffer)
  "Returns the content of BUFFER as a string, striping the final
newline if there is one"
  (with-current-buffer buffer
    (buffer-substring-no-properties
     (point-min)
     (progn (goto-char (point-max))
            (if (eq (char-before) ?\n)
                (- (point) 1)
              (point))))))

(defun tla--get-process-output ()
  "Returns the content of the last process buffer, stripping the final
newline if necessary"
  (tla--buffer-content tla--last-process-buffer))

(defun tla--get-error-output ()
  "Returns the content of the last error buffer, stripping the final
newline if necessary"
  (tla--buffer-content tla--last-error-buffer))

(defvar tla--switched-buffer nil)
(defvar tla--switched-from-buffer nil)

(defun tla-switch-to-buffer (buffer)
  "Allow customizable buffer switching.
See `tla-switch-to-buffer-mode' for possible settings."
  (setq tla--switched-from-buffer (current-buffer))
  (cond ((eq tla-switch-to-buffer-mode 'pop-to-buffer)
         (pop-to-buffer buffer))
        ((eq tla-switch-to-buffer-mode 'single-window)
         (switch-to-buffer buffer))
        ((eq tla-switch-to-buffer-mode 'show-in-other-window)
         (pop-to-buffer buffer)
         (setq tla--switched-buffer (current-buffer))
         (pop-to-buffer tla--switched-from-buffer))
        ;; TODO : dedicated frame.
        (t
         (error "Switch mode %s not implemented" tla-switch-to-buffer-mode))))

(defun tla-post-switch-to-buffer ()
  (when tla--switched-buffer
    (pop-to-buffer tla--switched-buffer)
    (setq tla--switched-buffer nil)
    (goto-char (point-min))
    (pop-to-buffer tla--switched-from-buffer)))


(defun tla-show-process-buffer ()
  "Show the process buffer of the last started tla command."
  (interactive)
  (tla-switch-to-buffer tla--last-process-buffer)
  (tla-process-buffer-mode))

(defun tla--show-last-process-buffer (&optional type mode)
  "Switch to the last used process buffer in a new buffer of type
TYPE. If MODE is specified, it is a function that will be run in the
new buffer. Otherwise, the buffer will remain in fundamental mode, in
read-only."
  (when (buffer-live-p tla--last-process-buffer)
    (let ((content (with-current-buffer tla--last-process-buffer
                     (buffer-string))))
      (tla-switch-to-buffer (tla--get-buffer-create (or type 'generic)))
      (let ((inhibit-read-only t))
        (erase-buffer)
        (insert content)))
    (if mode
        (funcall mode)
      (tla-process-buffer-mode))))

(defun tla--show-error-buffer (buffer &optional type mode)
  "Pops up the error buffer. Works like
`tla--show-last-process-buffer', but displays BUFFER, of type 'errors
if TYPE is not specified."
  (when (buffer-live-p buffer)
    (let ((content (with-current-buffer buffer
                     (buffer-string))))
      (tla-switch-to-buffer (tla--get-buffer-create
                             (or type 'errors)))
      (let ((inhibit-read-only t))
        (erase-buffer)
        (insert content)))
    (if mode
        (funcall mode)
      (tla-process-buffer-mode))))

;; ----------------------------------------------------------------------------
;; Arch name manipulators
;; ======================
;;
;; Normally in xtla, a name, a revision specifier is represented as a
;; list like:
;;
;;    ("archive" "category" "branch" "version" "revision")
;;
;; Nil is permitted as the element. However the list length must be 5
;; like:
;;
;;    (nil "category" "branch" nil nil)
;;
;; In other hand, in tla command, the name must be represented as a
;; string like:
;;
;;    "archive/category--branch--version--revision"
;;
;; So we have to convert a name in different representation in many
;; cases.
;;
;; * tla--name-split-* is for converting from a string representation
;;   to a list representation. There are semi-qualified version and
;;   fully-qualified version.
;;
;;   - semi-qualified: "category--branch--version--revision".
;;     `tla--name-split-semi-qualified' expects a name string without
;;     archive component. The archive field of returned list is filled
;;     with nil.
;;
;;   - fully-qualified: "archive/category--branch--version--revision".
;;     `tla--name-split' expects a name string including archive.
;;
;; * tla--name-construct-* is for converting from a list
;;   representation to a string representation. The functions accept
;;   arguments two ways.
;;
;;   - normall passing: (tla--name-construct "archive" "category"...)
;;   - packed passing: (tla--name-construct '("archive" "category"...))
;;
;;   There are semi-qualified version and fully-qualified version.
;;   - semi-qualified: `tla--name-construct-semi-qualified' connects
;;     arguments with "--".
;;   - fully-qualified: `tla--name-construct" connects the first argument
;;     and the rest with "/". About the rest,
;;     `tla--name-construct-semi-qualified' is applied.
;;
;; * tla--name-{archive|category|branch|version|revision} is for
;;   extracting a component from a name. The both representations are
;;   acceptable.
;;
;; ----------------------------------------------------------------------------

;;
;; String representation -> List representation
;;
(defun tla--name-split-semi-qualified (name &optional archive)
  "Split \"--\" connected string into 5 elements list.
The first element is always nil if ARCHIVE is not given.
If ARCHIVE is given, use it as the first.
Even if the elements in name are less than 5, the list is filled by nil
to make the length 5.

    ELISP> (tla--name-split-semi-qualified \"branch--category--version--revision\" \"archive\")
    (\"archive\" \"branch\" \"category\" \"version\" \"revision\")

    ELISP> (tla--name-split-semi-qualified \"branch--category--version--revision\")
    (nil \"branch\" \"category\" \"version\" \"revision\")

    ELISP> (tla--name-split-semi-qualified \"branch--category--version\")
    (nil \"branch\" \"category\" \"version\" nil)

    ELISP> (tla--name-split-semi-qualified \"branch--category--version\" \"archive\")
    (\"archive\" \"branch\" \"category\" \"version\" nil)

    ELISP> (tla--name-split-semi-qualified \"branch--category\" \"archive\")
    (\"archive\" \"branch\" \"category\" nil nil)

    ELISP> (tla--name-split-semi-qualified \"branch--category\"nil)
    (nil \"branch\" \"category\" nil nil)

    ELISP> (tla--name-split-semi-qualified \"branch--category--\" nil)
    (nil \"branch\" \"category\" \"\" nil)

"
  (let ((list (tla--name-split-semi-qualified-internal name)))
    (while (> 4 (length list))
        (setq list (cons nil list)))
    (cons archive (nreverse list))))

(defun tla--name-split-semi-qualified-internal (name)
  (if (string-match "^\\(.+\\)--\\(.*\\)" name)
      (cons (match-string 2 name)
            (tla--name-split-semi-qualified-internal
             (match-string 1 name)))
    (cons name nil)))

(defun tla--name-split (name)
  "Parses a fully qualified revision name, but possibly incomplete.
email@address.com--arch/cat--branch--ver ->
  (\"email@address.com--arch\" \"cat\" \"branch\" \"ver\" nil)
email@address.com--arch/cat ->
  (\"email@address.com--arch\" \"cat\" nil nil nil)
email@address.com--arch ->
  (\"email@address.com--arch\" nil nil nil nil)"
  (if (string-match "\\(.*\\)/\\(.*\\)" name)
      (tla--name-split-semi-qualified (match-string 2 name) (match-string 1 name))
    (if (string= name "")
        (list nil nil nil nil nil)
      (list name nil nil nil nil))))


;;
;; List representation -> string
;;
(defun tla--name-construct-semi-qualified (&rest comp)
  ""
  "Concatenate comp with \"--\".
This function can accept strings or a list which contains strings.

    ELISP> (tla--name-construct-semi-qualified \"a\" \"b\" \"c\")
    \"a--b--c\"
    ELISP> (tla--name-construct-semi-qualified (list \"a\" \"b\" \"c\"))
    \"a--b--c\"
"
  (if (consp (car comp)) (setq comp (car comp)))
  (mapconcat 'identity (remove nil comp) "--"))

(defun tla--name-construct (archive &optional
                                    category
                                    branch
                                    version
                                    revision)
  "Creates the fully qualified revision name :
archive/category--branch--version--revision.
The arguments may be nil. If ARCHIVE is a revision name list like
 (archive category branch version revision), the list element is mapped
to arguments before creating the fully qualified revision name."
  (when (consp archive)
    (setq category (tla-category-name archive)
          branch   (tla-branch-name archive)
          version  (tla-version-name archive)
          revision (tla-revision-name archive)
          ;; archive must be last
          archive  (tla-archive-name archive)))
  (let ((semi (tla--name-construct-semi-qualified
               category branch version revision)))
    (concat
     (and archive (not (string= archive ""))
          (concat archive (when category "/")))
     semi)))

;;
;; Get a component from a list or string.
;;
(defun tla--name-archive (target)
  "Get archive component from TARGET.
Both representation of TARGET, a string and a list is acceptable."
  (when (stringp target)
    (setq target (tla--name-split target)))
  (car target))

(defun tla--name-category (target)
  "Get category component from TARGET.
Both representation of TARGET, a string and a list is acceptable."
  (when (stringp target)
    (setq target (tla--name-split target)))
  (cadr target))

(defun tla--name-branch (target)
  "Get branch component from a TARGET.
Both representation of TARGET, a string and a list is acceptable."
  (when (stringp target)
    (setq target (tla--name-split target)))
  (caddr target))

(defun tla--name-version (target)
  "Get version component from TARGET.
Both representation of TARGET, a string and a list is acceptable."
  (when (stringp target)
    (setq target (tla--name-split target)))
  (cadddr target))

(defun tla--name-revision (target)
  "Get revision component from TARGET.
Both representation of TARGET, a string and a list is acceptable."
  (when (stringp target)
    (setq target (tla--name-split target)))
  (cadddr (cdr target)))

;; Older functions set
(defun tla-archive-name (list)
  "LIST must be a full revision in the form of a list."
  (car list))

(defun tla-category-name (list)
  "LIST must be a full revision in the form of a list."
  (cadr list))

(defun tla-branch-name (list)
  "LIST must be a full revision in the form of a list."
  (caddr list))

(defun tla-version-name (list)
  "LIST must be a full revision in the form of a list."
  (cadddr list))

(defun tla-revision-name (list)
  "LIST must be a full revision in the form of a list."
  (cadddr (cdr list)))

;;
;; Buffers menu
;;
(defun tla--buffers-menu ()
  "Return menus for buffers managed in xtla."
  (let ((menu (make-sparse-keymap "Tla-Buffers"))
        label
        submenu
        (i tla-number-of-dead-process-buffer))
    ;; Debug QUEUE
    (setq submenu (make-sparse-keymap "Queue"))
    (mapcar
     (lambda (buffer)
       (when (buffer-live-p buffer)
         (define-key submenu (vector (make-symbol (buffer-name buffer)))
           `(menu-item ,(format "%d: %s%s"
                                i
                                (if (zerop (buffer-size buffer)) "[empty] " "")
                                (buffer-name buffer))
                       (lambda () (interactive) (switch-to-buffer ,buffer))
                       :enable t)))
       (setq i (1- i)))
     tla--dead-process-buffer-queue)
    (define-key menu [queue]
      `(menu-item "Queue(DEBUG)"
                  ,submenu
                  :enable tla-show-internal-buffers-on-menu))
    (mapcar (lambda (item)
              (setq label (capitalize (symbol-name (car item))))
              (setq submenu (make-sparse-keymap label))
              (mapcar
               (lambda (subitem)
                 (let ((path (car subitem))
                       (buffer (cadr subitem)))
                   (when (buffer-live-p buffer)
                     (unless path
                       (setq path (buffer-name buffer)))
                     (define-key submenu (vector (make-symbol path))
                       `(menu-item ,path
                                   (lambda () (interactive) (switch-to-buffer ,buffer))
                                   :enable t)))))
               (cdr item))
              (define-key menu (vector (car item))
                `(menu-item ,label
                            ,submenu
                            :enable t))
              )
            tla--buffers-tree)
    menu))

(eval-when-compile
  (unless (functionp 'add-submenu)
    (defun add-submenu (&rest arg)
      "Avoids a byte-compiler warning for GNU Emacs")))

(defun tla--install-buffer-menu ()
  (if (featurep 'xemacs)
      ;; See tla-xemacs-buffers-menu in xtla-xemacs.el
      (add-submenu nil '("Tla-Buffers" :filter tla-xemacs-buffers-menu) nil)
    ;; GNU Emacs
    (define-key-after
      (or (lookup-key global-map [menu-bar tools XTLA])
          (lookup-key global-map [menu-bar tools xtla]))
      [tla-buffers]
      (cons "Tla-Buffers" (tla--buffers-menu)))
    (let ((map (and
                (current-local-map)
                (or (lookup-key (current-local-map) [menu-bar])
                    (define-key (current-local-map) [menu-bar]
                      (make-keymap))))))
      (when map
        (apply (if (functionp 'define-key-after)
                   'define-key-after
                 'define-key)
               map
               [tla-buffers]
               (cons "Tla-Buffers" (tla--buffers-menu))
               nil)))
    (add-hook 'menu-bar-update-hook 'tla--install-buffer-menu nil t)))

;; TODO: Use tla--archive-tree.
(defun tla--version-head (archive category branch version)
  "Return the newest revision ofr ARCHIVE/CATEGORY--BRANCH--VERSION."
  (tla--run-tla-sync (list "revisions"
                           (tla--name-construct
                            archive
                            category
                            branch
                            version))
                     :finished (lambda (output error status arguments)
                                 (with-current-buffer output
                                   (goto-char (point-max))
                                   (re-search-backward "^.")
                                   (buffer-substring-no-properties
                                    (point) (line-end-position))))))

;; ----------------------------------------------------------------------------
;; Archive tree manipulators
;; ----------------------------------------------------------------------------

(defvar tla--archive-tree nil
  "Arch archive/category/branch/version/revision are stored in assoc list:

 ((\"xsteve@nit.at--public\" \"http://xsteve.nit.at/tla\")
 [...]
  (\"mbp@sourcefrog.net--2004\" \"http://sourcefrog.net/arch/mbp@sourcefrog.net--2004\"
   (\"xtla\")
   (\"tilly\")
 [...]
   (\"dupes\"
    (\"mainline\"
     (\"0.1\")))
 [...]
   (\"archzoom\"))
  (\"mark@dishevelled.net--2003-mst\" \"http://members.iinet.net.au/~mtriggs/arch/\")
  (\"lord@emf.net--2004\" \"http://regexps.srparish.net/{archives}/lord@emf.net--2004\")
 [...]
  (\"Matthieu.Moy@imag.fr--public\" \"http://www-verimag.imag.fr/webdav/moy/public\"
   (\"xtla\"
    (\"main\"
     (\"0.1\"
      (\"patch-228\"
       \"Merged from Robert (patch87-91), Milan (patch21-22), Stefan (patch58-22)\"
       \"Matthieu Moy <Matthieu.Moy@imag.fr>\"
       \"2004-06-03 20:13:11 GMT\")
      (\"patch-227\"
       \"Fix default-directory in tla--run-tla-sync, fix in tla-changes-ediff\"
       \"Matthieu Moy <Matthieu.Moy@imag.fr>\"
       \"2004-06-03 15:26:15 GMT\")
 [...]
      (\"patch-1\"
       \"typo\"
       \"Matthieu Moy <Matthieu.Moy@imag.fr>\"
       \"2004-04-07 22:57:00 GMT\")
      (\"base-0\"
       \"tag of xsteve@nit.at--public/xtla--main--0.1--patch-5\"
       \"Matthieu Moy <Matthieu.Moy@imag.fr>\" \"2004-04-07 22:52:39 GMT\")))))
 [...]
   )

This list is initially empty, and is built/rebuilt on demand.
")

;; Utilities
(defun tla--archive-tree-setcdr (parent value &optional rest)
  (let* ((current (cdr parent))
         (list-details (assoc value current)))
    (if (or (null current) (null list-details))
        ;; rest is '("summary" "creator" "date") when value is "patch-N"
        (setcdr parent (cons (cons value rest) current))
      (if (and list-details rest)
          ;; Field already there. update details.
          (setcdr list-details rest)))))

(defun tla--archive-tree-setcddr (parent value)
  (let ((current (cddr parent)))
    (if (or (null current) (null (assoc value current)))
        (setcdr (cdr parent) (cons (cons value nil) current)))))

;; Archive
(defun tla--archive-tree-add-archive (archive location &optional old)
  "Adds ARCHIVE to the archive tree. If OLD is provided, it is an old
archive tree from which some information can be found (this is usefull
to keep the category/branch/version info for existing archives)."
  (if (tla--archive-tree-get-archive archive)
      (let* ((a (tla--archive-tree-get-archive archive))
             (val (cdr a))
             (oldlocation (car val))
             (category (cdr val)))
        (setcdr a (cons (or location oldlocation) category)))
    (let ((oldinfo (tla--archive-tree-get-archive archive old))
          (newinfo (list archive location)))
      (when oldinfo
        (setcdr (cdr newinfo) (cddr oldinfo))) ;; list of versions.
      (setq tla--archive-tree (cons newinfo
                                    tla--archive-tree)))))

(defun tla--archive-tree-get-archive (archive &optional archive-tree)
  (assoc archive (or archive-tree tla--archive-tree)))

;; Category
(defun tla--archive-tree-add-category (archive category)
  (tla--archive-tree-add-archive archive nil)
  (tla--archive-tree-setcddr
   (tla--archive-tree-get-archive archive)
   category))

(defun tla--archive-tree-get-category (archive category)
  (assoc category (cdr (cdr (tla--archive-tree-get-archive archive)))))

;; Branch
(defun tla--archive-tree-add-branch (archive category branch)
  (tla--archive-tree-add-category archive category)
  (tla--archive-tree-setcdr
   (tla--archive-tree-get-category archive category)
   branch))

(defun tla--archive-tree-get-branch (archive category branch)
  (assoc branch (cdr (tla--archive-tree-get-category
                      archive category))))

;; Version
(defun tla--archive-tree-add-version (archive category branch version)
  (tla--archive-tree-add-branch archive category branch)
  (tla--archive-tree-setcdr
   (tla--archive-tree-get-branch archive category branch )
   version))

(defun tla--archive-tree-get-version (archive category branch version)
  (assoc version (cdr (tla--archive-tree-get-branch
                       archive category branch))))

;; Revision
(defun tla--archive-tree-add-revision (archive category branch version revision
                                               &optional summary creator date)
  (tla--archive-tree-add-version archive category branch version)
  (tla--archive-tree-setcdr
   (tla--archive-tree-get-version archive category branch version)
   revision (list summary creator date)))

(defun tla--archive-tree-get-revision (archive category branch version revision)
  (assoc revision (cdr (tla--archive-tree-get-version
                        archive category branch version))))

;; Archive tree builders
(defun tla--archive-tree-build (basename &optional use-cache ignore-error)
  "Generic version of tla--archive-tree-build-*."
  (when (stringp basename)
    (setq basename (tla--name-split basename)))
  (let ((archive (tla--name-archive basename))
        (category (tla--name-category basename))
        (branch (tla--name-branch basename))
        (version (tla--name-version basename)))
  (cond
   (version
    (tla--archive-tree-build-revisions archive
                                       category
                                       branch
                                       version
                                       use-cache
                                       ignore-error))
   (branch
    (tla--archive-tree-build-versions archive
                                      category
                                      branch
                                      use-cache
                                      ignore-error))
   (category
    (tla--archive-tree-build-branches archive
                                      category
                                      use-cache
                                      ignore-error))
   (archive
    (tla--archive-tree-build-categories archive
                                        use-cache
                                        ignore-error))
   (t
    (tla--archive-tree-build-archives use-cache
                                      ignore-error)))))

(defun tla--archive-tree-build-archives (&optional use-cache ignore-error)
  "Builds the list of archives"
  (when (or (not use-cache)
            (not tla--archive-tree))
    (tla--run-tla-sync '("archives")
                       :finished 'tla--null-handler
                       :error
                       (if ignore-error
                           'tla--null-handler
                         'tla--default-error-function))
    (let ((old-archive-tree tla--archive-tree))
      (setq tla--archive-tree nil)
      (save-excursion
        (let ((archive-name)
              (archive-location))
          (set-buffer tla--last-process-buffer)
          (goto-char (point-min))
          (while (> (line-end-position) (line-beginning-position))
            (setq archive-name (buffer-substring-no-properties
                                (line-beginning-position)
                                (line-end-position)))
            (beginning-of-line-text 2)
            (setq archive-location (buffer-substring-no-properties
                                    (point) (line-end-position)))
            (forward-line 1)
            (tla--archive-tree-add-archive archive-name
                                           archive-location
                                           old-archive-tree)))))))

(defun tla--archive-tree-build-categories (archive &optional
                                                   use-cache
                                                   ignore-error)
  "Builds the list of categories for an archive in `tla--archive-tree'"
  (when (or (not use-cache)
            (not (cddr (tla--archive-tree-get-archive archive))))
    (let ((basename archive))
      (message "building categories for `%s'..." basename)
      (tla--run-tla-sync (list "categories" "-A" basename)
                         :finished 'tla--null-handler
                         :error
                         (if ignore-error
                             'tla--null-handler
                           'tla--default-error-function))
      (message "building categories for `%s'...done" basename)
      (sit-for 0)
      (message nil))
    (with-current-buffer tla--last-process-buffer
      (let (category)
        (goto-char (point-min))
        (while (> (line-end-position) (line-beginning-position))
          (setq category (buffer-substring-no-properties
                          (line-beginning-position)
                          (line-end-position)))
          (forward-line 1)
          (tla--archive-tree-add-category archive category)
          )))))

(defun tla--archive-tree-build-branches (archive category
                                                 &optional
                                                 use-cache
                                                 ignore-error)
  "Builds the list of branches for a category in `tla--archive-tree'"
  (when (or (not use-cache)
            (not (cdr (tla--archive-tree-get-category archive category))))
    (let ((basename (tla--name-construct archive category)))
      (message "building branches for `%s'..." basename)
      (tla--run-tla-sync (list "branches" basename)
                         :finished 'tla--null-handler
                         :error
                         (if ignore-error
                             'tla--null-handler
                           'tla--default-error-function))
      (message "building branches for `%s'...done" basename)
      (sit-for 0)
      (message nil))
    (with-current-buffer tla--last-process-buffer
      (let (branch)
        (goto-char (point-min))
        (while (> (line-end-position) (line-beginning-position))
          (setq branch (buffer-substring-no-properties
                        (line-beginning-position)
                        (line-end-position)))
          (forward-line 1)
          (tla--archive-tree-add-branch
           archive
           category
           (tla-branch-name (tla--name-split-semi-qualified branch))))))))

(defun tla--archive-tree-build-versions (archive category branch
                                                 &optional
                                                 use-cache
                                                 ignore-error)
  "Builds the list of versions in a branch in `tla--archive-tree'"
  (when (or (not use-cache)
            (not (cdr (tla--archive-tree-get-branch archive category
                                                    branch))))
    (let ((basename (tla--name-construct archive category branch)))
      (message "building versions for `%s'..." basename)
      (tla--run-tla-sync (list "versions" basename)
                       :finished 'tla--null-handler
                       :error
                       (if ignore-error
                           'tla--null-handler
                         'tla--default-error-function))
      (message "building versions for `%s'...done" basename)
      (sit-for 0)
      (message nil))
    (with-current-buffer tla--last-process-buffer
      (let (version)
        (goto-char (point-min))
        (while (> (line-end-position) (line-beginning-position))
          (setq version (buffer-substring-no-properties
                         (line-beginning-position)
                         (line-end-position)))
          (forward-line 1)
          (tla--archive-tree-add-version
           archive
           category
           branch
           (tla-version-name (tla--name-split-semi-qualified version))))))))

(defun tla--archive-tree-build-revisions (archive category branch version
                                                  &optional
                                                  use-cache
                                                  ignore-error)

  "Builds the list of revisions in a version in `tla--archive-tree'"
  (when (or (not use-cache)
            (not (cdr (tla--archive-tree-get-version archive category branch
                                                     version))))
    (let ((details (or tla-revisions-shows-summary
                       tla-revisions-shows-date
                       tla-revisions-shows-creator))
          (basename (tla--name-construct
                     archive category branch version)))
      (message "building revisions for `%s'..." basename)
      (if details
          (progn
            (tla--run-tla-sync (list "revisions"
                                     "--summary" "--date" "--creator"
                                     basename)
                               :finished 'tla--null-handler
                               :error (if ignore-error
                                          'tla--null-handler
                                        'tla--default-error-function)))
        (progn
          (tla--run-tla-sync (list "revisions" basename)
                             :finished 'tla--null-handler
                             :error (if ignore-error
                                        'tla--null-handler
                                      'tla--default-error-function))))
      (message "building revisions for `%s'...done" basename)
      (sit-for 0)
      (message nil)
      (with-current-buffer tla--last-process-buffer
        (let (revision date creator summary)
          (goto-char (point-min))
          (while (> (line-end-position) (line-beginning-position))
            (setq revision (buffer-substring-no-properties
                            (line-beginning-position)
                            (line-end-position)))
            (forward-line 1)
            (when details
              (skip-chars-forward " ")
              (setq date (buffer-substring-no-properties (point)
                                                         (line-end-position)))
              (forward-line 1)
              (skip-chars-forward " ")
              (setq creator (buffer-substring-no-properties (point)
                                                            (line-end-position)))
              (forward-line 1)
              (skip-chars-forward " ")
              (setq summary (buffer-substring-no-properties (point)
                                                            (line-end-position)))
              (forward-line 1))
            (tla--archive-tree-add-revision
             archive
             category
             branch
             version
             revision
             summary
             creator
             date)))))))


(defun tla--revisions-tree-contains-details
  (archive category branch version)
  "Whether VERSION has already been listed with summary lines, dates,
and creator in the archive tree."
  (let ((vtree (tla--archive-tree-get-version archive category branch
                                              version)))
    (and (cdr vtree) ;; revision list is here
         (cadr (cadr vtree))))) ;; summary line also

;; ----------------------------------------------------------------------------
;; Revlib tree manipulators
;; ----------------------------------------------------------------------------
(defvar tla--revlib-tree nil)
(defun tla--revlib-tree-get-archive (archive &optional archive-tree)
  (assoc archive (or archive-tree tla--revlib-tree)))
(defun tla--revlib-tree-build-archives (&optional use-cache ignore-error)
  "Builds the list of archives for an archive in `tla--revlib-tree'
If IGNORE-ERROR is non-nil, error is not reported.
Return non-nil if the tree entry for archives are updated."
  (when (or (not use-cache)
            (not tla--revlib-tree))
    (tla--run-tla-sync '("library-archives")
                       :finished 'tla--null-handler
                       :error
                       (if ignore-error
                           'tla--null-handler
                         'tla--default-error-function))
    (let ((old-revlib-tree tla--revlib-tree) )
      (setq tla--revlib-tree nil)
      (save-excursion
        (let ((archive-name)
              (tmp tla--archive-tree)
              (tla--archive-tree tla--revlib-tree)
              result)
          (set-buffer tla--last-process-buffer)
          (goto-char (point-min))
          (while (> (line-end-position) (line-beginning-position))
            (setq result t)
            (setq archive-name (buffer-substring-no-properties
                                (line-beginning-position)
                                (line-end-position)))
            (forward-line 1)
            (tla--archive-tree-add-archive archive-name
                                           nil
                                           old-revlib-tree))
          (setq tla--revlib-tree tla--archive-tree
                tla--archive-tree tmp)
          result)))))

(defun tla--revlib-tree-get-category (archive category)
  (assoc category (cdr (cdr (tla--revlib-tree-get-archive archive)))))

(defun tla--revlib-tree-build-categories (archive &optional
                                                  use-cache
                                                  ignore-error)
  "Builds the list of categories for an archive in `tla--revlib-tree'
If IGNORE-ERROR is non-nil, error is not reported.
Return non-nil if the tree entry for categories are updated."
  (when (or (not use-cache)
            (not (cddr (tla--revlib-tree-get-archive archive))))
    (tla--run-tla-sync (list "library-categories" "-A" archive)
                       :finished 'tla--null-handler
                       :error
                       (if ignore-error
                           'tla--null-handler
                         'tla--default-error-function))
    (with-current-buffer tla--last-process-buffer
      (let (category
            (tmp tla--archive-tree)
            (tla--archive-tree tla--revlib-tree)
            result)
        (goto-char (point-min))
        (while (> (line-end-position) (line-beginning-position))
          (setq result t)
          (setq category (buffer-substring-no-properties
                          (line-beginning-position)
                          (line-end-position)))
          (forward-line 1)
          (tla--archive-tree-add-category archive category))
        (setq tla--revlib-tree tla--archive-tree
              tla--archive-tree tmp)
        result))))

(defun tla--revlib-tree-get-branch (archive category branch)
  (assoc branch (cdr (tla--revlib-tree-get-category
                      archive category))))

(defun tla--revlib-tree-build-branches (archive category
                                                &optional
                                                use-cache
                                                ignore-error)
  "builds the list of branches for a category in `tla--revlib-tree'
If IGNORE-ERROR is non-nil, error is not reported.
Return non-nil if the tree entry for branches are updated."
  (when (or (not use-cache)
            (not (cdr (tla--revlib-tree-get-category archive category))))
    (tla--run-tla-sync (list "library-branches" "-A" archive category)
                       :finished 'tla--null-handler
                       :error
                       (if ignore-error
                           'tla--null-handler
                         'tla--default-error-function))
    (with-current-buffer tla--last-process-buffer
      (let (branch
            (tmp tla--archive-tree)
            (tla--archive-tree tla--revlib-tree)
            result)
        (goto-char (point-min))
        (while (> (line-end-position) (line-beginning-position))
          (setq result t)
          (setq branch (buffer-substring-no-properties
                        (line-beginning-position)
                        (line-end-position)))
          (forward-line 1)
          (tla--archive-tree-add-branch
           archive
           category
           (tla-branch-name (tla--name-split-semi-qualified branch))))
        (setq tla--revlib-tree tla--archive-tree
              tla--archive-tree tmp)
        result))))

(defun tla--revlib-tree-get-version (archive category branch version)
  (assoc version (cdr (tla--revlib-tree-get-branch
                       archive category branch))))

(defun tla--revlib-tree-build-versions (archive category branch
                                                &optional
                                                use-cache
                                                ignore-error)
  "Builds the list of versions in a branch in `tla--archive-tree'
If IGNORE-ERROR is non-nil, error is not reported.
Return non-nil if the tree entry for versions are updated."
  (when (or (not use-cache)
            (not (cdr (tla--revlib-tree-get-branch archive category
                                                   branch))))
    (tla--run-tla-sync (list "library-versions"
                             (tla--name-construct
                              archive category branch))
                       :finished 'tla--null-handler
                       :error
                       (if ignore-error
                           'tla--null-handler
                         'tla--default-error-function))
    (with-current-buffer tla--last-process-buffer
      (let (version
            (tmp tla--archive-tree)
            (tla--archive-tree tla--revlib-tree)
            result)
        (goto-char (point-min))
        (while (> (line-end-position) (line-beginning-position))
          (setq result t)
          (setq version (buffer-substring-no-properties
                         (line-beginning-position)
                         (line-end-position)))
          (forward-line 1)
          (tla--archive-tree-add-version
           archive
           category
           branch
           (tla-version-name (tla--name-split-semi-qualified version))))
        (setq tla--revlib-tree tla--archive-tree
              tla--archive-tree tmp)
        result))))

(defun tla--revlib-tree-get-revision (archive category branch version revision)
  (assoc revision (cdr (tla--revlib-tree-get-version
                        archive category branch version))))

(defun tla--revlib-tree-build-revisions (archive category branch version
                                                 &optional
                                                 use-cache
                                                 ignore-error)

  "Builds the list of revisions in a version in `tla--revlib-tree'
If IGNORE-ERROR is non-nil, error is not reported.
Return non-nil if the tree entry for revisions are updated."
  (when (or (not use-cache)
            (not (cdr (tla--revlib-tree-get-version archive category branch
                                                    version))))
    (tla--run-tla-sync (list "library-revisions"
                             "--summary" "--date" "--creator"
                             (tla--name-construct
                              archive category branch version))
                       :finished 'tla--null-handler
                       :error (if ignore-error
                                  'tla--null-handler
                                'tla--default-error-function))
    (with-current-buffer tla--last-process-buffer
      (let (revision
            date
            creator
            summary
            (tmp tla--archive-tree)
            (tla--archive-tree tla--revlib-tree)
            result)

        (goto-char (point-min))
        (while (> (line-end-position) (line-beginning-position))
          (setq result t)
          (setq revision (buffer-substring-no-properties
                          (line-beginning-position)
                          (line-end-position)))
          (forward-line 1)
          (skip-chars-forward " ")
          (setq date (buffer-substring-no-properties (point)
                                                     (line-end-position)))
          (forward-line 1)
          (skip-chars-forward " ")
          (setq creator (buffer-substring-no-properties (point)
                                                        (line-end-position)))
          (forward-line 1)
          (skip-chars-forward " ")
          (setq summary (buffer-substring-no-properties (point)
                                                        (line-end-position)))
          (forward-line 1)
          (tla--archive-tree-add-revision
           archive
           category
           branch
           version
           revision
           summary
           creator
           date))
        (setq tla--revlib-tree tla--archive-tree
              tla--archive-tree tmp)
        result
        ))))

;; ----------------------------------------------------------------------------
;; Name reading engine
;; ----------------------------------------------------------------------------
;;Currently only able to read a full revision starting from nothing.

(defvar tla--name-read-minibuf-map
  (let ((map (copy-keymap minibuffer-local-completion-map)))
    (define-key map [(control r)] 'tla--name-read-refresh-cache)
    map))

(defun tla--name-read-refresh-cache ()
  "Function to be called from the minibuffer while reading a name."
  (interactive)
  (tla--archive-tree-build
   (tla--name-construct
    (butlast (delete nil (tla--name-split (minibuffer-contents))))))
  (setq tla--archive-tree nil))

(defun tla--name-read-complete (string predicate what)
  "Completion function for name reading.

This function uses the free variables archive, category, branch,
version, and revision. If one of these variables is non-nil, it means
the corresponding value must be read from keyboard.

REMINDER: this function may be called several times, with different
values for WHAT:

 - nil : The function must return the longest prefix
 - t : The function must return the list of completions
 - 'lambda : The function must return t if the completion correspond
   to an exact match, nil otherwise. (so that Emacs can distinguish
   between \"sole completion\" and \"complete, but not unique\"."
  (if (and (eq what 'lambda)
           (string-match "/\\(.*--\\)?$" string))
      ;; The caller just want to know wether this is a full
      ;; completion. This can not be the case with such suffix.
      nil
    (let* ((use-cache (not current-prefix-arg))
           (splited (tla--name-split string))
           (archive-loc  (tla-archive-name  splited))
           (category-loc (tla-category-name splited))
           (branch-loc   (tla-branch-name   splited))
           (version-loc  (tla-version-name  splited))
           (revision-loc (tla-revision-name splited))
           (completions
            (cond
             ;; If the user started to write a revision ...
             (revision-loc
              ;; ... and if the user is supposed to be prompted a
              ;; revision
              (when revision
                (let ((tla-revisions-shows-summary nil)
                      (tla-revisions-shows-date nil)
                      (tla-revisions-shows-creator nil))
                  (tla--archive-tree-build-revisions
                   archive-loc category-loc branch-loc version-loc use-cache t))
                (cdr (tla--archive-tree-get-version
                      archive-loc category-loc branch-loc version-loc))))
             (version-loc
              (when version
                (tla--archive-tree-build-versions
                 archive-loc category-loc branch-loc use-cache t)
                (cdr (tla--archive-tree-get-branch
                      archive-loc category-loc branch-loc))))
             (branch-loc
              (when branch
                (tla--archive-tree-build-branches
                 archive-loc category-loc use-cache t)
                (cdr (tla--archive-tree-get-category
                      archive-loc category-loc))))
             (category-loc
              (when category
                (tla--archive-tree-build-categories archive-loc use-cache t)
                (cddr (tla--archive-tree-get-archive archive-loc))))
             (t
              (when archive
                (tla--archive-tree-build-archives use-cache t)
                tla--archive-tree)))))
      (let ((completions
             (funcall 'all-completions
                      string
                      (mapcar (lambda (x)
                                (list (tla--name-construct
                                       (delete nil (list (when category-loc archive-loc)
                                                         (when branch-loc category-loc)
                                                         (when version-loc branch-loc)
                                                         (when revision-loc version-loc)
                                                         (car x))))))
                              completions)
                      predicate)))
        (let ((result
               (cond ((eq what t)
                      ;; We just want the list of completions
                      completions)
                     ((eq (length completions) 1)
                      ;; There's only one completion
                      (if (eq what 'lambda)
                          (string= (car completions) string)
                        (let ((suffix (cond
                                       ((and category (not category-loc) "/"))
                                       ((and branch   (not branch-loc)   "--"))
                                       ((and version  (not version-loc)  "--"))
                                       ((and revision (not revision-loc) "--"))
                                       (t nil))))
                          (cond (suffix (concat (car completions) suffix))
                                ((string= (car completions) string) t)
                                (t (car completions))))))
                     ;; there are several possible completions
                     (t (if (eq what 'lambda)
                            ;; complete, but not unique ?
                            (member string completions)
                          (try-completion string (mapcar 'list
                                                         completions)))))))
;;          (tla--trace "string=%s predicate=%S what=%s ==> result=%S\ncompletions=%S"
;;                      string predicate what result completions)
          result)))))

;; Test cases
;; (tla--name-read "enter category: " "Matthieu.Moy@imag.fr--public" 'prompt)
;; (tla--name-read "branch: " "lord@emf.net--2004" 'prompt 'prompt)
;; (tla--name-read "revision: " 'prompt 'prompt 'prompt 'prompt 'prompt)
;; (tla--name-read "revision or version: " 'prompt 'prompt 'prompt 'prompt 'maybe)
(defvar tla--name-read-history nil)
(defun tla--name-read (&optional prompt archive category
                                 branch version revision)
  "Function reading an archive location from keyboard.
Read name is expressed in a list built by `tla--name-split'.

First argument PROMPT is the prompt the user will get. Next arguments
ARCHIVE CATEGORY BRANCH VERSION and REVISION are either the default
value, or a request for a value. They can take four values:

 - A string means the default value, and will be used as an initial
   input.

 - The symbol 'prompt means the value will be prompted from the user.
   The user will HAVE to give this value.

 - The symbol 'maybe means the value will be prompted, but is optional
   for the user.

 - nil means the value won't be prompted.

They should appear in the abover order.

Example:
- Read a category in archive \"Matthieu.Moy@imag.fr--public\":
 (tla--name-read \"enter category: \" \"Matthieu.Moy@imag.fr--public\" 'prompt)
- Read a revision, anywhere:
 (tla--name-read \"revision: \" 'prompt 'prompt 'prompt 'prompt 'prompt)
- Read either a revision or a version:
 (tla--name-read \"revision: \" 'prompt 'prompt 'prompt 'prompt 'maybe)

At any time, the user may refresh the cache by pressing C-r.

"
  (let* ((minibuffer-local-completion-map tla--name-read-minibuf-map)
         (result (tla--name-construct
                        (delete
                         'maybe
                         (delete 'prompt (list archive category
                                               branch revision ""))))))
    (while (tla--prompt-not-finished result archive category branch
                                     version revision)
        (setq result
              (completing-read
               (or prompt "Location: ")
               'tla--name-read-complete
               nil nil result
               'tla--name-read-history)))
    (if result
        (tla--name-split result)
      result)))

(defun tla--prompt-not-finished (result archive category branch
                                        version revision)
  "True if RESULT (a string) is not sufficient when the user is
prompted for ARCHIVE CATEGORY BRANCH VERSION REVISION."
  (let ((res-split (tla--name-split result)))
    (or (and (eq archive 'prompt)
             (not (tla-archive-name res-split)))
        (and (eq category 'prompt)
             (not (tla-category-name res-split)))
        (and (eq branch 'prompt)
             (not (tla-branch-name res-split)))
        (and (eq version 'prompt)
             (not (tla-version-name res-split)))
        (and (eq revision 'prompt)
             (not (tla-revision-name res-split))))))


(defun tla-compute-direct-ancestor (&optional revision)
  "Computes the direct ancestor of REVISION. REVISION must be provided
as a list, and a list is returned. If revision is nil, return the
ancestor of the last revision of the local tree."
  (interactive
   (list (tla--name-read "Compute direct ancestor of: "
                         'prompt 'prompt 'prompt 'prompt 'prompt)))
  (let ((ancestor
         (tla--run-tla-sync (list "ancestry-graph" "--immediate"
                                  (and revision
                                       (tla--name-construct revision)))
                            :finished (lambda (output error status arguments)
                                        (tla--name-split
                                         (tla--buffer-content
                                          output))))))
    (when (interactive-p)
      (message "Ancestor of: %s\n         is: %s"
               (tla--name-construct ancestor)
               (tla--name-construct revision)))
    ancestor))


;; Debugging facilities
(defvar tla--debug t)

(defun tla--trace (&rest msg)
  "Same as `message' if `tla--debug' is non-nil. Does nothing
otherwise. Please use it for your debug messages."
  (when tla--debug
    (apply 'message (concat "xtla: " (car msg)) (cdr msg))))

(defun tla-reload ()
  "Reloads xtla."
  (interactive)
  (when (featurep 'xtla)      (unload-feature 'xtla t))
  (when (featurep 'xtla-core) (unload-feature 'xtla-core t))
  (when (featurep 'xtla-defs) (unload-feature 'xtla-defs t))
  (when (featurep 'xtla-xemacs) (unload-feature 'xtla-xemacs t))
  (require 'xtla))

;; (setq tla--archive-tree nil)
;; (setq tla--revlib-tree nil)
(provide 'xtla-core)
;; arch-tag: c9e35f5a-6aea-409d-a157-c0d73d92f9b0
;;; xtla-core.el ends here
