;;; xtla.el --- Arch interface for emacs

;; Copyright (C) 2003-2004 by Stefan Reichoer

;; Author: Stefan Reichoer, <xsteve@nit.at>
;; Contributions from:
;;    Matthieu Moy <Matthieu.Moy@imag.fr>
;;    Masatake YAMATO <jet@gyve.org>
;;    Milan Zamazal <pdm@zamazal.org>
;;    Martin Pool <mbp@sourcefrog.net>
;;    Robert Widhopf-Fenk <hack@robf.de>
;;    Mark Triggs <mst@dishevelled.net>

;; xtla.el is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; xtla.el is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, Inc., 59 Temple Place - Suite 330,
;; Boston, MA 02111-1307, USA.

;;; Commentary:

;; Some documentation can be found on the wiki here
;;     http://wiki.gnuarch.org/moin.cgi/xtla

;; There is a project page at
;;     https://gna.org/projects/xtla-el
;; You can subscribe to the mailing list via
;;     https://mail.gna.org/listinfo/xtla-el-dev

;; Usage:
;; put the following in your .emacs: (require 'xtla)

;; The main commands are available with the prefix key C-x T.
;; Type C-x T C-h for a list.

;; M-x tla-inventory shows a tla inventory
;; In this inventory buffer the following commands are available:
;; e ... tla-edit-log
;; = ... tla-changes
;; l ... tla-changelog
;; L ... tla-logs

;; To Edit a logfile issue: M-x tla-edit-log
;; In this mode you can hit C-c C-d to show the changes
;; Edit the log file
;; After that you issue M-x tla-commit (bound to C-c C-c) to commit the files

;; M-x tla-archives starts the interactive archive browser

;; M-x tla-make-archive creates a new archive directory
;; Many commands are available from here. Look at the menus, they're
;; very helpful to begin.

;; M-x tla-bookmarks RET
;; Is another good starting point. This is the place where you put the
;; project you work on most often, and you can get a new version, see
;; the missing patches, and a few other useful features from here.
;; Use `a' to add a bookmark. Add your own projects, and your
;; contributor's projects too. Select several related projects with
;; `m' (unselect with M-u or M-del). Make them partners with 'M-p'.
;; Now, with your cursor on a bookmark, view the uncommitted changes,
;; the missing patches from your archive and your contributors with
;; 'M'.

;; M-x tla-file-ediff RET
;; Is an wrapper to tla file-diff, ediff to view the changes
;; interactively.

;; Misc commands:
;; tla-insert-arch-tag inserts a arch-tag entry generated with uuidgen

;; There is an arch archive for xtla.el at: http://xsteve.nit.at/tla
;; Just type: M-x tla-register-archive <RET> http://xsteve.nit.at/tla <RET>
;;            M-x tla-archives <RET>
;; With these commands you can start to browse the archive

;; If you find xtla.el useful, and you have some ideas to improve it
;; please share them with us (Patches are preferred :-))

;;; Todo:
;; See docs/Too

;;; Code:

(eval-and-compile
  (require 'cl)
  (require 'ediff)
  (require 'font-lock))

(require 'sendmail)
(require 'pp)
(require 'ewoc)
(require 'diff-mode)
(require 'avoid)

(eval-and-compile
  (require 'xtla-defs)
  (require 'xtla-core))

(eval-and-compile
  (when (locate-library "xtla-version")
    (load-library "xtla-version"))

  (when (locate-library "smerge-mode")
    (load-library "smerge-mode"))

  (when (locate-library "hl-line")
    (load-library "hl-line"))

  (if (featurep 'xemacs)
      (load-library "xtla-xemacs")))

(autoload 'dired-recursive-delete-directory "dired")

;; ----------------------------------------------------------------------------
;; Internal variables
;; ----------------------------------------------------------------------------
(defvar tla-edit-arch-command nil)
(defvar tla-pre-commit-window-configuration nil)
(defvar tla-log-edit-file-name nil)
(defvar tla-log-edit-file-buffer nil)
(defvar tla-my-id-history nil)

(defvar tla-buffer-archive-name nil)
(defvar tla-buffer-category-name nil)
(defvar tla-buffer-branch-name nil)
(defvar tla-buffer-version-name nil)
(defvar tla-buffer-refresh-function nil
  "Variable should be local to each buffer. Function used to refresh
the current buffer")
(defvar tla-buffer-marked-file-list nil
  "List of marked files in the current buffer.")
(defvar tla-get-revision-info-at-point-function nil
  "Variable should be local to each buffer.
Function used to get the revision info at point")

(defvar tla-mode-line-process "")
(defvar tla-mode-line-process-status "")

(defvar tla-partner-select-was-interactive nil)

;; Overlay category
(put 'tla-default-button 'mouse-face 'highlight)
(put 'tla-default-button 'evaporate t)
;;(put 'tla-default-button 'rear-nonsticky t)
;;(put 'tla-default-button 'front-nonsticky t)

;;;###autoload
(defun xtla ()
  (interactive)
  (let* ((name "*xtla-welcome*")
         (buffer (get-buffer name)))
    (if buffer (tla-switch-to-buffer buffer)
      (tla-switch-to-buffer
       (setq buffer (get-buffer-create name)))
      (insert "            *** Welcome to Xtla ! ***

Xtla is the Emacs interface to the revision control system Arch, using
the reference implementation tla. (Therefore, xtla commands are
prefixed by \"tla-\" and not \"xtla-\")

As a starting point, you should look at the \"Tools\" menu, there is a
\"XTLA\" entry with a lot of interesting commands.

There is also a manual for Xtla. It should be available using the Info
system, and you can find it on the web from here:
http://wiki.gnuarch.org/moin.cgi/xtla

Hope you'll enjoy it !
")
      (toggle-read-only t)
      (local-set-key [?q] (lambda () (interactive)
                            (kill-buffer (current-buffer)))))))

;; ----------------------------------------------------------------------------
;; Face manipulators
;; ----------------------------------------------------------------------------
(defun tla--face-add (str face &optional keymap menu)
  "Add FACE, KEYMAP and MENU to STR as text properties and return it."
  (if tla-highlight
      (let ((strcpy (format "%s" str)))
        (add-text-properties 0 (length strcpy)
                             `(face ,face
;;; Even if we define a face in a buffer, it seems that
;;; font-lock mode just ignore it or remove the face property.
;;; I don't know the detail but in tla-inventory buffer,
;;; I cannot make both font-lock keywords and faces put by tl--face-add
;;; highlight at once. When font-lock-face is defined, I can do.
;;; See "Special Properties" subsection in the emacs lisp reference manual.
;;; `font-lock-face' property is new in Emacs 21.4. However, I guess there is
;;; no wrong side effect if I define font-lock-face property here.
                                    font-lock-face ,face
                                    ,@(when keymap
                                        `(mouse-face highlight
                                                     keymap ,keymap))
                                    ,@(when menu
                                        `(menu ,menu))
                                    )
                             strcpy)
        strcpy)
    str))

(defun tla--face-add-with-condition (condition text face1 face2)
  "If condition then add face1 to text, else add face2 to text."
  (if condition
      (tla--face-add text face1)
    (tla--face-add text face2)))

(defun tla--face-set-temporary-during-popup (begin end face menu &optional prefix)
  "Put FACE on BEGIN and END in the buffer during Popup MENU.
PREFIX is passed to `popup-menu'."
  (let (o)
    (unwind-protect
        (progn
          (setq o (make-overlay begin end))
          (overlay-put o 'face face)
          (sit-for 0)
          (popup-menu menu prefix))
      (delete-overlay o))))

;; ----------------------------------------------------------------------------
;; Macros
;; ----------------------------------------------------------------------------
(defmacro tla-toggle-list-entry (list entry)
  "Either add or remove the entry from list"
  `(if (member ,entry ,list)
       (setq ,list (delete ,entry ,list))
     (add-to-list ',list ,entry)))

(defun tla-update-mode-line ()
  (setq tla-mode-line-process tla-mode-line-process-status)
  ;; The following is not yet needed:
  ;; (concat tla-status-mode-line-process-edit-flag tla-mode-line-process-status))
  (force-mode-line-update))


;; ----------------------------------------------------------------------------
;; Common used functions for many xtla modes
;; ----------------------------------------------------------------------------
(defun tla-kill-all-buffers ()
  "Kill all xtla buffers."
  (interactive)
  (let ((number 0))
    (dolist (type-cons tla--buffers-tree)
      (dolist (path-buffer (cdr type-cons))
        (setq number (1+ number))
        (kill-buffer (cadr path-buffer))))
    (message "Killed %d buffer%s" number
             (if (> number 1) "s" "")))
  (setq tla--buffers-tree nil))

(defun tla-buffer-quit ()
  "Kill the current buffer."
  (interactive)
  (kill-buffer (current-buffer)))

(defun tla-edit-=tagging-method ()
  "Edit the {arch}/=tagging-method file."
  (interactive)
  (find-file (expand-file-name "{arch}/=tagging-method" (tla-tree-root))))

(defun tla-edit-.arch-inventory (&optional dir)
  "Edit DIR/.arch-inventory file.
default-directory is used as DIR if DIR is nil."
  (interactive)
  (unless dir
    (setq dir default-directory))
  (let* ((file (expand-file-name ".arch-inventory" dir))
         (newp (not (file-exists-p file))))
  (find-file file)
  (save-excursion
    (when (and newp (y-or-n-p
                     (format "Insert arch tag to \"%s\"" file)))
      (tla-insert-arch-tag)))))

(defun tla--ewoc-delete (cookie elem)
  "Remove element ELEM from COOKIE"
  (ewoc-filter cookie
               '(lambda (x) (not (eq x (ewoc-data elem))))))

(defun tla-generic-refresh ()
  "Calls the function specified by `tla-buffer-refresh-function'"
  (interactive)
  (let ((tla-read-directory-mode 'never)
        (tla-read-project-tree-mode 'never))
    (funcall tla-buffer-refresh-function)))

(defun tla--get-info-at-point ()
  "Get the version information that point is on."
  (when (fboundp tla-get-revision-info-at-point-function)
    (funcall tla-get-revision-info-at-point-function)))

(defvar tla-window-config nil
  "Used for inter-function communication.")

(defun tla--ediff-buffers (bufferA bufferB)
  "Wrapper around `ediff-buffers'"
  (let ((tla-window-config (current-window-configuration)))
    (ediff-buffers bufferA bufferB
                   '(tla-ediff-startup-hook) 'tla-ediff)))

(defun tla--insert-right-justified (string count &optional face)
  "Inserts STRING preceded by spaces so that the line ends exactly at
COUNT characters (or after if STRING is too long."
  (insert-char ?\  (max 0 (- count (length string))))
  (insert (if face (tla--face-add string face) string))
  )

(defun tla--generic-popup-menu (event prefix)
  "Generic function to popup a menu defined in the text property under the point which is given by mouse."
  (interactive "e\nP")
  (mouse-set-point event)
  (tla--generic-popup-menu-by-keyboard prefix))

(defun tla--generic-popup-menu-by-keyboard (prefix)
  "Generic function to popup a menu defined in the text property under the point."
  (interactive "P")
  (if (interactive-p)
      (let* ((pos (mouse-avoidance-point-position))
             (object (car pos))
             (x (cadr pos))
             (y (cddr pos)))
        (set-mouse-position object x y)))
  (let* ((b (1- (previous-single-property-change (point) 'menu)))
         (e (next-single-property-change (point) 'menu))
         (menu (get-text-property (point) 'menu)))
    (tla--face-set-temporary-during-popup b e
                                          'highlight
                                          menu
                                          prefix)))
;;
;; Functions to read archive/category/branch/version/revision names
;; from keyboard
;;

;; archive
(defvar tla--read-archive-history nil)
(defun tla--read-archive-name (&optional archive-prompt default)
  "Reads an archive name from keyboard. Returned value is a
one-element list: (archive)."
  (tla--archive-tree-build-archives t)
  (let* ((my-default-archive (or default (tla-my-default-archive)))
         (archive-prompt (or archive-prompt "Archive name: "))
         (archive (or     ;(tla--get-archive (tla--get-info-at-point))
                   (completing-read
                    (concat archive-prompt
                            "(default: " my-default-archive ") ")
                    tla--archive-tree
                    nil nil nil
                    'tla--read-archive-history
                    my-default-archive))))
    (list (when (not (string= archive ""))
            archive))))

;; category
(defun tla--completion-category (string predicate what)
  (tla--archive-tree-build-categories archive t t)
  (funcall (if (eq what t) 'all-completions 'try-completion)
           string
           (cddr (tla--archive-tree-get-archive archive))
           predicate))

(defvar tla--read-archive-category-history nil)
(defun tla--complete-category (archive &optional prompt)
  "See doc for `tla--complete-revision'"
  (let ((category (or ;(cadr (tla--get-category (tla--get-info-at-point)))
                   (completing-read
                    (or prompt (concat "category for " archive ": "))
                    (or (cddr (tla--archive-tree-get-archive archive))
                        'tla--completion-category)
                    nil nil nil
                    'tla--read-archive-category-history))))
    (when (not (string= category ""))
      category)))

(defun tla--read-archive-category-name (&optional archive-prompt
                                                  category-prompt
                                                  archive)
  "See doc for `tla--read-archive-category-branch-version-revision-name'"
  (let* ((l (or (and archive (list archive))
                (tla--read-archive-name archive-prompt)))
         (a (car l))
         (c (and a (tla--complete-category a category-prompt))))
    (list a c)))

;; branch
(defun tla--completion-branch (string predicate what)
  (tla--archive-tree-build-branches archive category t t)
  (funcall (if (eq what t) 'all-completions 'try-completion)
           string
           (cdr (tla--archive-tree-get-category archive category))
           predicate))

(defvar tla--read-archive-category-branch-history nil)
(defun tla--complete-branch (archive category &optional prompt)
  "See doc for `tla--complete-revision'"
  (let ((branch (completing-read
                 (or prompt (concat "branch for " archive "/" category ": "))
                 (or (cdr (tla--archive-tree-get-category archive category))
                     'tla--completion-branch)
                 nil nil nil
                 'tla--read-archive-category-branch-history)))
    (when (not (string= branch ""))
      branch)))

(defun tla--read-archive-category-branch-name (&optional archive-prompt
                                                         category-prompt
                                                         branch-prompt
                                                         archive category)
  "See doc for `tla--read-archive-category-branch-version-revision-name'"
  (let* ((l (when (not category)
              (tla--read-archive-category-name
               archive-prompt
               category-prompt
               archive)))
         (a (or archive (car l)))
         (c (or category (cadr l)))
         (b (and c (tla--complete-branch a c branch-prompt))))
    (list a c b)))

;; version
(defun tla--completion-version (string predicate what)
  (tla--archive-tree-build-versions archive category branch t t)
  (funcall (if (eq what t) 'all-completions 'try-completion)
           string
           (cdr (tla--archive-tree-get-branch archive category branch))
           predicate))

(defvar tla--read-archive-category-branch-version-history nil)
(defun tla--complete-version (archive category branch &optional
                                      prompt)
  "See doc for `tla--complete-revision'"
  (let ((version (completing-read
                  (or prompt (concat "version for " (tla--name-construct
                                                     archive category
                                                     branch) ": "))
                  (or (cdr (tla--archive-tree-get-branch archive category branch))
                      'tla--completion-version)
                  nil nil nil
                  'tla--read-archive-category-branch-version-history)))
    (when (not (string= version ""))
      version)))

(defun tla--read-archive-category-branch-version-name (&optional
                                                       archive-prompt
                                                       category-prompt
                                                       branch-prompt
                                                       version-prompt
                                                       archive
                                                       category
                                                       branch)
  "See doc for `tla--read-archive-category-branch-version-revision-name'"
  (let* ((l (when (not branch)
              (tla--read-archive-category-branch-name
               archive-prompt
               category-prompt
               branch-prompt
               archive category)))
         (a (or archive (car l)))
         (c (or category (cadr l)))
         (b (or branch (caddr l)))
         (v (and b (tla--complete-version a c b version-prompt))))
    (list a c b v)))

;; revision
(defun tla--completion-revision (string predicate what)
  (tla--archive-tree-build-revisions archive category branch version t t)
  (funcall (if (eq what t) 'all-completions 'try-completion)
           string
           (cdr (tla--archive-tree-get-version archive category branch version))
           predicate))

(defvar tla--read-archive-category-branch-version-revision-history nil)
(defun tla--complete-revision (archive category branch version
                                       &optional prompt)
  "Reads a revision name in the archive
ARCHIVE/CATEGORY--BRANCH--VERSION. The string entered is returned if
not empty. Otherwise, nil is returned.

Completion uses the `tla--archive-tree' variable. If the archive tree is
already built for that revision, read from the archive tree.
Otherwise, call tla to build the archive tree."
  (let ((revision (completing-read
                   (or prompt (concat "revision for " (tla--name-construct
                                                       archive category branch version) ": "))
                   (or (cdr (tla--archive-tree-get-version archive category branch version))
                       'tla--completion-revision)
                   nil nil nil
                   'tla--read-archive-category-branch-version-revision-history)))
    (when (not (string= revision ""))
      revision)))

(defun tla--read-archive-category-branch-version-revision-name (&optional archive-prompt
                                                                          category-prompt
                                                                          branch-prompt
                                                                          version-prompt
                                                                          revision-prompt
                                                                          archive
                                                                          category
                                                                          branch
                                                                          version)
  "Reads a full revision name from keyboard. Arguments ARCHIVE-PROMPT,
CATEGORY-PROMPT, VERSION-PROMPT and REVISION-PROMPT are the prompt
strings that will be displayed to the user. Say nil here to keep the
default value.

The return value is a list: (archive category branch version revision)

The user will be prompted, in order, the archive, category, branch and
version names. Entering an empty string will stop prompting, and the
values not entered will be nil in the returned list.

If ARCHIVE, CATEGORY, BRANCH or VERSION are provided then the
provided values are used, and the user isn't prompted. The values must
be given in order: If you specify BRANCH, then you must also specify
ARCHIVE and CATEGORY, but not necessarily VERSION."
  (let* ((l (when (not version)
              (tla--read-archive-category-branch-version-name
               archive-prompt category-prompt branch-prompt
               version-prompt
               archive category branch)))
         (a (or archive (tla-archive-name l)))
         (c (or category (tla-category-name l)))
         (b (or branch (tla-branch-name l)))
         (v (or version (tla-version-name l)))
         (r (and v (tla--complete-revision a c b v revision-prompt))))
    (list a c b v r)))

(defun tla--read-archive-category-branch-version-revision-name-for-tree (tree &optional
                                                                              archive-prompt
                                                                              category-prompt
                                                                              branch-prompt
                                                                              version-prompt
                                                                              revision-prompt)
  "Similar to `tla--read-archive-category-branch-version-revision-name'.
But the archive associated with TREE is used as default archive in reading."
  (tla--read-archive-category-branch-version-revision-name
   archive-prompt
   category-prompt
   branch-prompt
   version-prompt
   revision-prompt
   (tla-archive-name (tla--read-archive-name archive-prompt
                                             (tla-archive-name
                                              (tla-tree-version-list tree))))))

;;
;; String fully qualified revision manipulators
;;
(defun tla--get-archive (rev)
  "Get the archive part of a fully qualified revision"
  (if rev
      (or (nth 0 (tla--name-split (cadr rev))) (cadr rev))
    (car (tla--read-archive-name))))

(defun tla--get-category (rev)
  "Get the archive/category part of a fully qualified revision"
  (let* ((revs (tla--name-split (cadr rev)))
         (archive (tla-archive-name revs))
         (category (and archive
                        (or (tla-category-name revs)
                            (tla--complete-category archive)))))
    (list archive category)))

(defun tla--get-branch (rev)
  "Get the archive/category-branch part of a fully qualified revision"
  (let* ((revs (tla--name-split rev))
         (archive (tla-archive-name revs))
         (category (and archive
                        (or (tla-category-name revs)
                            (tla--complete-category archive))))
         (branch (and category
                      (or (tla-branch-name revs)
                          (tla--complete-branch archive category)))))
    (tla--name-construct archive category branch)))

(defun tla--get-version (rev)
  "Get the archive/category-branch-version part of a fully qualified revision"
  (let* ((revs (tla--name-split rev))
         (archive (tla-archive-name revs))
         (category (and archive
                        (or (tla-category-name revs)
                            (tla--complete-category archive))))
         (branch (and category
                      (or (tla-branch-name revs)
                          (tla--complete-branch archive category))))
         (version (and branch
                       (or (tla-version-name revs)
                           (tla--complete-version archive category branch)))))
    (tla--name-construct archive category branch version)))

(defun tla--get-revision (rev)
  "Get the archive/category-branch-version-revision part of a fully qualified revision"
  (let* ((revs (tla--name-split rev))
         (archive (tla-archive-name revs))
         (category (and archive
                        (or (tla-category-name revs)
                            (tla--complete-category archive))))
         (branch (and category
                      (or (tla-branch-name revs)
                          (tla--complete-branch archive category))))
         (version (and branch
                       (or (tla-version-name revs)
                           (tla--complete-version archive category branch))))
         (revision (and version
                        (or (tla-revision-name revs)
                            (tla--complete-revision archive category branch version)))))
    (tla--name-construct archive category branch version revision)))

(defun tla-tree-root (&optional location no-error)
  "Returns the tree root for LOCATION, nil if not in a local tree.
Computation is done from withing Emacs, by looking at an {arch}
directory in a parent buffer of LOCATION. This is therefore very
fast."
  (setq location (or location default-directory))
  (let ((pwd location))
    (while (not (or (string= pwd "/")
                    (file-exists-p (concat pwd "/{arch}"))))
      (setq pwd (expand-file-name (concat pwd "/.."))))
    (if (file-exists-p (concat pwd "/{arch}/=tagging-method"))
        (expand-file-name
         (replace-regexp-in-string "/+$" "/" pwd))
      (if no-error
          nil
        (error "%S is not in an arch-managed tree!" location)))))

(defun tla--read-project-tree-maybe (&optional prompt directory)
  "Returns a directory name which is the root of some project tree.
Either prompt from the user or use the current directory. The
recommended usage is

 (defun tla-some-feature (...)
   (let ((default-directory (tla--read-project-tree-maybe
                             \"Run some feature in\")))
      (code-for-some-feature))

The behavior can be changed according to the value of
`tla-read-project-tree-mode'."
  (let ((root (tla-tree-root (or directory default-directory) t))
        (default-directory (or (tla-tree-root
                                (or directory default-directory) t)
                               directory
                               default-directory))
        (prompt (or prompt "Use directory: ")))
    (case tla-read-project-tree-mode
      (always (tla-tree-root (tla--read-directory-name prompt)))
      (sometimes (or root
                     (tla-tree-root (tla--read-directory-name prompt))))
      (never (or root
                 (error "Not in a project tree")))
      (t (error "wrong value for tla-prompt-for-directory.")))))

(defun tla--read-directory-maybe (&optional prompt directory)
  "Returns a directory name which is a subdirectory or the root of
some project tree. Works in a way similar to
`tla--read-project-tree-maybe', but is customized with the variable
`tla-read-directory-mode'."
  (let ((root (tla-tree-root (or directory default-directory) t))
        (default-directory (or directory default-directory))
        (prompt (or prompt "Use directory: ")))
    (case tla-read-directory-mode
      (always (tla--read-directory-name prompt))
      (sometimes (if root (or directory default-directory)
                   (tla--read-directory-name prompt)))
      (never (if root (or directory default-directory)
               (error "Not in a project tree")))
      (t (error "wrong value for tla-read-directory-mode.")))))

(defun tla-save-some-buffers (&optional tree)
  "Saves all buffers visiting a file in TREE"
  (let ((ok t)
        (tree (or (tla-tree-root tree t)
                  tree)))
    (unless tree
      (error "Not in a project tree."))
    (dolist (buffer (buffer-list))
      (with-current-buffer buffer
        (when (buffer-modified-p)
          (let ((file (buffer-file-name)))
            (when file
              (let ((root (tla-tree-root (file-name-directory file) t))
                    (tree-exp (expand-file-name tree)))
                (when (and (string= root tree-exp)
                           ;; buffer is modified and in the tree TREE.
                           (or tla-do-not-prompt-for-save
                               (y-or-n-p (concat "Save buffer "
                                                 (buffer-name)
                                                 "? "))
                               (setq ok nil)))
                  (save-buffer))))))))
    ok))

(defun tla-revert-some-buffers (&optional tree)
  "Reverts all buffers visiting a file in TREE that aren't modified.
To be run after an update or a merge."
  (let ((tree (tla-tree-root tree)))
    (dolist (buffer (buffer-list))
      (with-current-buffer buffer
        (when (not (buffer-modified-p))
          (let ((file (buffer-file-name)))
            (when file
              (let ((root (tla-tree-root (file-name-directory file) t))
                    (tree-exp (expand-file-name tree)))
                (when (and (string= root tree-exp)
                           ;; buffer is modified and in the tree TREE.
                           tla-automatically-revert-buffers)
                  ;; TODO
                  ;; what should we do if file is not existed? >> Matthieu
                  (if (file-exists-p file)
                      (revert-buffer t t)))))))))))

;; ----------------------------------------------------------------------------
;; tla help system for commands that get input from the user via the minibuffer
;; ----------------------------------------------------------------------------

;; When the user is asked for input in the minibuffer, a help for the
;; command will be shown, if the user hits f1
;; This functionality is not only for xtla commands available
;; it is available for all emacs commands
;; to check: we should use some other binding for this, perhaps f1 C-m

;; GENERIC: This functionality should be in emacs itself. >> Masatake

(defun tla-display-command-help (command &optional current-prompt)
  (with-electric-help
   (lambda ()
     (let ((cmd-help (when (fboundp command)
                       (documentation command))))
       (delete-region (point-min) (point-max))
       (insert (if cmd-help
                   (format "Help for %S:\n%s" command cmd-help)
                 (format "No help available for %S" command)))))
   " *tla-command-help*"))

(defvar tla-command-stack nil)

(defun tla-minibuffer-setup ()
  (push  this-command tla-command-stack))

(defun tla-minibuffer-exit ()
  (pop tla-command-stack))

(defun tla-show-command-help ()
  (interactive)
  (tla-display-command-help (car tla-command-stack)))

(when tla-install-command-help-system
  (define-key minibuffer-local-map [f1] 'tla-show-command-help)
  (define-key minibuffer-local-completion-map [f1] 'tla-show-command-help)
  (define-key minibuffer-local-must-match-map [f1] 'tla-show-command-help)
  (add-hook 'minibuffer-setup-hook 'tla-minibuffer-setup)
  (add-hook 'minibuffer-exit-hook 'tla-minibuffer-exit))

;; ----------------------------------------------------------------------------
;; Top level tla commands
;; ----------------------------------------------------------------------------
(defcustom tla-make-log-function 'tla-default-make-log-function
  "*function used to create the log buffer. Must return a string which
is the absolute name of the log file. This function is called only
when the log file doesn't exist already. The default is
`tla-default-make-log-function', which just calls \"tla make-log\". If
you want to override this function, you may just write a wrapper
around `tla-default-make-log-function'."
  :type 'function
  :group 'xtla)

(defun tla-make-log ()
  (interactive)
  (let* ((version (tla-tree-version-list))
         (file (concat (tla-tree-root) "++log."
                       (tla-category-name version) "--"
                       (tla-branch-name   version) "--"
                       (tla-version-name  version) "--"
                       (tla-archive-name  version))))
    (if (file-exists-p file)
        file
      (funcall tla-make-log-function))))

(defun tla-default-make-log-function ()
  (tla--run-tla-sync '("make-log")
                     :finished
                     (lambda (output error status arguments)
                       (tla--buffer-content output))))

(defun tla-pop-to-inventory ()
  (interactive)
  (tla-inventory nil t))

(defvar tla-inventory-cookie nil)
(defvar tla-inventory-list nil
  "Full list for the inventory.")

(defun tla-inventory-goto-file (file)
  "Put cursor on FILE. nil return means the file hasn't been found"
  (goto-char (point-min))
  (let ((current (ewoc-locate tla-inventory-cookie)))
    (while (and current (not (string= (caddr (ewoc-data current))
                                      file)))
      (setq current (ewoc-next tla-inventory-cookie current)))
    (when current (tla-inventory-cursor-goto current))
    current))


(defun tla-inventory-make-toggle-fn-and-var (variable function)
  "Defines the variable and the toggle function for type TYPE."
  (make-variable-buffer-local variable)
  (eval `(defun ,function ()
           (interactive)
           (setq ,variable (not ,variable))
           (tla-inventory-redisplay))))

(dolist (type-arg tla-inventory-file-types-manipulators)
  (tla-inventory-make-toggle-fn-and-var (cadr type-arg) (caddr type-arg)))

(defun tla-inventory-redisplay ()
  (let* ((elem (ewoc-locate tla-inventory-cookie))
         (file (when elem (caddr (ewoc-data elem))))
         (pos (point)))
    (tla-inventory-display)
    (or (and file
             (tla-inventory-goto-file file))
        (goto-char pos))
    (tla-inventory-cursor-goto (ewoc-locate tla-inventory-cookie))))


(defun tla-inventory-set-toggle-variables (new-value)
  "Set all tla-inventory-display-* variables.
If NEW-VALUE is 'toggle set the values to (not tla-inventory-display-*
Otherwise set it to NEW-VALUE."
  (dolist (type-arg tla-inventory-file-types-manipulators)
    (eval `(setq ,(cadr type-arg)
                 (if (eq new-value 'toggle)
                     (not ,(cadr type-arg))
                   new-value)))))

(defun tla-inventory-set-all-toggle-variables ()
  (interactive)
  (tla-inventory-set-toggle-variables t)
  (tla-inventory-redisplay))

(defun tla-inventory-reset-all-toggle-variables ()
  (interactive)
  (tla-inventory-set-toggle-variables nil)
  (tla-inventory-redisplay))

(defun tla-inventory-toggle-all-toggle-variables ()
  (interactive)
  (tla-inventory-set-toggle-variables 'toggle)
  (tla-inventory-redisplay))


;;;###autoload
(defun tla-inventory (&optional directory arg)
  "Show a tla inventory at DIRECTORY.
When called with a prefix arg, pop to the inventory buffer.
DIRECTORY defaults to the current one when within an arch managed tree,
unless you gave two prefix args."
  (interactive (list (tla--read-directory-maybe
                      "Run inventory in (directory): ")
                     current-prefix-arg))
  (let ((directory (or directory default-directory)))
    (if arg
        (pop-to-buffer (tla--get-buffer-create 'inventory directory))
      (switch-to-buffer (tla--get-buffer-create 'inventory directory)))
    (cd directory))
  (tla-inventory-mode)
  (tla--run-tla-sync
   ;; We have to provide all file types or tla inventory won't display
   ;; junk files
   '("inventory" "--both" "--kind" "--source" "--backups" "--junk"
     "--unrecognized" "--precious")
   :finished
   (lambda (output error status arguments)
     (let ((list (split-string (tla--buffer-content output) "\n"))
           (inventory-list '()))
       (mapc
        (lambda (item)
          (when (string-match "\\([A-Z]\\)\\([\\? ]\\) +\\([^ ]\\) \\(.*\\)"
                              item)
            (push (list (string-to-char (match-string 1 item))
                        (string= (match-string 2 item) "?")
                        (match-string 4 item)
                        (string-to-char (match-string 3 item)))
                  inventory-list)))
        list)
       (setq inventory-list (reverse inventory-list))
       (set (make-local-variable 'tla-inventory-list)
            inventory-list)
       (tla-inventory-display)))))

(defun tla-inventory-display ()
  (interactive)
  (let (buffer-read-only)
    (erase-buffer)
    (set (make-local-variable 'tla-inventory-cookie)
         (ewoc-create 'tla-inventory-printer))
    (tla-inventory-insert-headers)
    (dolist (elem tla-inventory-list)
      (let ((type (car elem)))
        (if (eval (cadr (assoc type
                               tla-inventory-file-types-manipulators)))
            (ewoc-enter-last tla-inventory-cookie elem)))))
  (goto-char (point-min)))

(defun tla--inventory-chose-face (type)
  "Returns a face adapted to TYPE, which can be J, S, P, T or U"
  (case type
    (?P 'tla-precious)
    (?U 'tla-unrecognized)
    (?S 'tla-source)
    (?J 'tla-junk)
    (?T 'tla-nested-tree)))

(defun tla-inventory-printer (elem)
  (let* ((type (nth 0 elem))
         (question (nth 1 elem))
         (file (nth 2 elem))
         (file-type (nth 3 elem))
         (face (tla--inventory-chose-face type)))
    (insert (if (member file tla-buffer-marked-file-list)
                " * " "   "))
    (insert (tla--face-add (format "%c%s  "
                                   type
                                   (if question "?" " "))
                           face))
    (let ((intermediate (point)))
      (insert (tla--face-add
               (format "%s%s" file
                       (case file-type (?d "/") (?> "@") (t "")))
               face
               tla-inventory-item-map
               tla-inventory-item-menu)))))

(defun tla-inventory-mark-file ()
  (interactive)
  (let ((current (ewoc-locate tla-inventory-cookie))
        (file (tla--get-file-info-at-point)))
    (add-to-list 'tla-buffer-marked-file-list file)
    (ewoc-refresh tla-inventory-cookie)
    (tla-inventory-cursor-goto (or (ewoc-next tla-inventory-cookie
                                              current)
                                   current))))

(defun tla-inventory-unmark-file ()
  (interactive)
  (let ((current (ewoc-locate tla-inventory-cookie))
        (file (tla--get-file-info-at-point)))
    (setq tla-buffer-marked-file-list
          (delete file tla-buffer-marked-file-list))
    (ewoc-refresh tla-inventory-cookie)
    (tla-inventory-cursor-goto (or (ewoc-next tla-inventory-cookie
                                              current)
                                   current))))

(defun tla-inventory-unmark-all ()
  (interactive)
  (let ((current (ewoc-locate tla-inventory-cookie)))
    (setq tla-buffer-marked-file-list nil)
    (ewoc-refresh tla-inventory-cookie)
    (tla-inventory-cursor-goto current)))

(defvar tla-get-file-info-at-point-function nil
  "Function used to get the file at point, anywhere")

(defun tla--get-file-info-at-point ()
  (funcall tla-get-file-info-at-point-function))

(defun tla-inventory-get-file-info-at-point ()
  (caddr (ewoc-data (ewoc-locate tla-inventory-cookie))))

(defun tla-inventory-insert-headers ()
  (let* ((tree-version (tla--name-construct
                        (tla-tree-version-list nil 'no-error)))
         (separator
          (tla--face-add (make-string
                          (max (+ (length "Directory: ")   (length default-directory))
                               (+ (length "Default Tree Version: ") (length tree-version)))
                          ?\ )
                         'tla-separator)))
    (ewoc-set-hf
     tla-inventory-cookie
     (concat
      "Directory: "    (tla--face-add default-directory 'tla-local-directory) "\n"
      "Default Tree Version: " (tla--face-add tree-version 'tla-archive-name) "\n"
      separator "\n")
     (concat "\n" separator))))

(defvar tla-buffer-source-buffer nil
  "Buffer from where a command was called")

;;;###autoload
(defun tla-edit-log (&optional insert-changelog source-buffer)
  "Edit the tla log file.
With an optional prefix argument, insert the last group of entries from the
ChangeLog file. SOURCE-BUFFER, if non-nil, is the buffer from which
the function was called. It is used to get the list of marked files,
and potentially run a selected file commit."
  (interactive "P")
  (setq tla-pre-commit-window-configuration (current-window-configuration))
  (when (and tla--last-process-buffer (get-buffer tla--last-process-buffer))
    (kill-buffer tla--last-process-buffer))
  (setq tla-log-edit-file-name (tla-make-log))
  (tla-switch-to-buffer
   (find-file-noselect tla-log-edit-file-name))
  (when insert-changelog
    (goto-char (point-max))
    (let ((buf (find-file-noselect (find-change-log))))
      (insert-buffer buf))
    (when (re-search-forward "^2" nil t)
      (delete-region (line-beginning-position)
                     (line-beginning-position 3)))
    (when (re-search-forward "^2" nil t)
      (delete-region (line-beginning-position) (point-max)))
    (goto-char (point-min)))
  (tla-log-edit-mode)
  (set (make-local-variable 'tla-buffer-source-buffer)
       source-buffer)
  (end-of-line))

;;;###autoload
(defun tla-add-log-entry ()
  "Add new tla log ChangeLog style entry."
  (interactive)
  (save-restriction
    (tla-add-log-entry-internal)))

(defun tla-add-log-entry-internal ()
  ;; This is mostly copied from add-log.el.  Perhaps it would be better to
  ;; split add-change-log-entry into several functions and then use them, but
  ;; that wouldn't work with older versions of Emacs.
  (require 'add-log)
  (let* ((defun (add-log-current-defun))
         (buf-file-name (if (and (boundp 'add-log-buffer-file-name-function)
                                 add-log-buffer-file-name-function)
                            (funcall add-log-buffer-file-name-function)
                          buffer-file-name))
         (buffer-file (if buf-file-name (expand-file-name buf-file-name)))
         (file-name (tla-make-log))
         ;; Set ENTRY to the file name to use in the new entry.
         (entry (add-log-file-name buffer-file file-name))
         beg
         bound
         narrowing)
    (tla-edit-log)
    (undo-boundary)
    (goto-char (point-min))
    (when (re-search-forward "^Patches applied:" nil t)
      (narrow-to-region (point-min) (match-beginning 0))
      (setq narrowing t)
      (goto-char (point-min)))
    (re-search-forward "\n\n\\|\\'")
    (setq beg (point))
    (setq bound
          (progn
            (if (looking-at "\n*[^\n* \t]")
                (skip-chars-forward "\n")
              (if (and (boundp 'add-log-keep-changes-together)
                       add-log-keep-changes-together)
                  (goto-char (point-max))
                (forward-paragraph))) ; paragraph delimits entries for file
            (point)))
    (goto-char beg)
    (forward-line -1)
    ;; Now insert the new line for this entry.
    (cond ((re-search-forward "^\\s *\\*\\s *$" bound t)
           ;; Put this file name into the existing empty entry.
           (if entry
               (insert entry)))
          ((let (case-fold-search)
             (re-search-forward
              (concat (regexp-quote (concat "* " entry))
                      ;; Don't accept `foo.bar' when
                      ;; looking for `foo':
                      "\\(\\s \\|[(),:]\\)")
              bound t))
           ;; Add to the existing entry for the same file.
           (re-search-forward "^\\s *$\\|^\\s \\*")
           (goto-char (match-beginning 0))
           ;; Delete excess empty lines; make just 2.
           (while (and (not (eobp)) (looking-at "^\\s *$"))
             (delete-region (point) (line-beginning-position 2)))
           (insert-char ?\n 2)
           (forward-line -2)
           (indent-relative-maybe))
          (t
           ;; Make a new entry.
           (if tla-log-insert-last
               (progn
                 (goto-char (point-max))
                 (re-search-backward "^.")
                 (end-of-line)
                 (insert "\n\n* ")
                 )
             (forward-line 1)
             (while (looking-at "\\sW")
               (forward-line 1))
             (while (and (not (eobp)) (looking-at "^\\s *$"))
               (delete-region (point) (line-beginning-position 2)))
             (insert-char ?\n 3)
             (forward-line -2)
             (indent-to left-margin)
             (insert "* "))
           (if entry (insert entry))))
    (if narrowing (widen))
    ;; Now insert the function name, if we have one.
    ;; Point is at the entry for this file,
    ;; either at the end of the line or at the first blank line.
    (if defun
        (progn
          ;; Make it easy to get rid of the function name.
          (undo-boundary)
          (unless (save-excursion
                    (beginning-of-line 1)
                    (looking-at "\\s *$"))
            (insert ?\ ))
          ;; See if the prev function name has a message yet or not
          ;; If not, merge the two entries.
          (let ((pos (point-marker)))
            (if (and (skip-syntax-backward " ")
                     (skip-chars-backward "):")
                     (looking-at "):")
                     (progn (delete-region (+ 1 (point)) (+ 2 (point))) t)
                     (> fill-column (+ (current-column) (length defun) 3)))
                (progn (delete-region (point) pos)
                       (insert ", "))
              (goto-char pos)
              (insert "("))
            (set-marker pos nil))
          (insert defun "): "))
      ;; No function name, so put in a colon unless we have just a star.
      (unless (save-excursion
                (beginning-of-line 1)
                (looking-at "\\s *\\(\\*\\s *\\)?$"))
        (insert ": ")))))

(defvar tla-changes-cookie nil
  "ewoc cookie for the changes buffer.

Element should look like

 (file \"filename\" \"M\" \"/\")
 (subtree \"name\" related-buffer changes?)
 (message \"doing such or such thing\")
")

(defun tla--changes-delete-messages (&optional immediate)
  "Removes messages from the ewoc list of changes. if IMMEDIATE is
non-nil, refresh the display too."
  (when tla-changes-cookie
    (ewoc-filter tla-changes-cookie
                 (lambda (elem)
                   (not (eq (car elem) 'message))))))

(defvar tla--changes-summary nil
  "Wether the current buffer display only a summary or a full diff.")

(defvar tla--changes-buffer-master-buffer nil
  "Master buffer for a nested *tla-changes* buffer")

(defvar tla--changes-summary nil
  "Wether the current buffer display only a summary or a full diff.")

;;;###autoload
(defun tla-changes (&optional summary against)
  "Run tla changes.
When called without a prefix argument: show the detailed diffs also.
When called with a prefix argument: do not show detailed diffs.
When AGAINST is non-nil, use it as comparison tree."
  (interactive "P")
  (let* ((root (tla--read-project-tree-maybe
                "Run tla changes in: "))
         (default-directory root)
         (buffer (tla--prepare-changes-buffer ;; TODO rename function
                  (or against
                      (list 'last-revision root))
                  (list 'local-tree root))))
    (with-current-buffer buffer
      (set (make-local-variable 'tla--changes-summary)
           summary))
    (when tla-switch-to-buffer-first
      (tla-switch-to-buffer buffer))
    (tla-save-some-buffers)
    (tla--run-tla-async
     '("inventory" "--nested" "--trees")
     :related-buffer buffer
     :finished
     `(lambda (output error status arguments)
        (let ((subtrees (delete ""
                                (split-string
                                 (with-current-buffer
                                     output (buffer-string)) "\n"))))
          (with-current-buffer ,buffer
            (let ((inhibit-read-only t))
              (ewoc-enter-last
               tla-changes-cookie
               (list 'message
                     (concat "* running tla changes in tree " ,root
                             "...\n\n")))
              (ewoc-refresh tla-changes-cookie))
            (dolist (subtree subtrees)
              (let ((buffer-sub (tla--get-buffer-create
                                 'changes subtree)))
                (with-current-buffer buffer-sub
                  (let ((inhibit-read-only t))
                    (erase-buffer))
                  (tla-changes-mode)
                  (set (make-local-variable
                        'tla--changes-buffer-master-buffer)
                       ,buffer))
                (ewoc-enter-last tla-changes-cookie
                                 (list 'subtree buffer-sub subtree
                                       nil))
                (tla--changes-internal
                 ,(not summary)
                 nil ;; TODO "against" what for a nested tree?
                 subtree
                 buffer-sub
                 ,buffer)))
            (tla--changes-internal ,(not summary)
                                   (quote ,against)
                                   ,root ,buffer nil)))))))

;;;###autoload
(defun tla-changes-last-revision (&optional summary)
  "Runs tla-changes against the last but one revision. The idea is
that running this command just after a commit should be equivalent to
running `tla-changes' just before the commit."
  (interactive "P")
  (let ((default-directory (tla--read-project-tree-maybe
                            "Review last patch in directory: ")))
    (tla-changes summary (tla--name-construct (tla-compute-direct-ancestor)))))

(defvar tla--changes-modified nil
  "MODIFIED revision for the changes currently displayed. Must be
buffer-local.

This variable has the form (type location), and can be either

'(revision (\"archive\" \"category\" \"branch\" \"version\"
            \"revision\"))

or

'(local-tree \"/path/to/local/tree\")

The value nil means we have no information about which local tree or
revision is used.")

(defvar tla--changes-base nil
  "BASE revision for the changes currently displayed. Must be
buffer-local.

The values for this variable can be the same as for
`tla--changes-modified', plus the values

'(last-revision \"/path/to/tree\"),
used by tla-changes to mean \"revision on which this local tree is
based\".

and

'(previous-revision (\"archive\" \"category\" \"branch\" \"version\"
                     \"revision\")),
used by commands like tla-get-changeset, and means that the changes
are against the previous revision.
")

(defun tla--changes-internal (diffs against root buffer master-buffer)
  "Runs tla changes in directory ROOT against tree AGAINST.
if DIFFS is non nil, show the detailed diffs also.
The output will be displayed in buffer BUFFER.

BUFFER must already be in changes mode, but mustn't contain any change
information. Only roots of subprojects are already in the ewoc.

If MASTER-BUFFER is non-nil, this run of tla changes is done in a
nested project of a bigger one. MASTER-BUFFER is the buffer in which
the root of the projects is displayed."
  (with-current-buffer buffer
    (tla--run-tla-async
     `("changes" ,(when diffs "--diffs")
       ,(case (car against)
          (local-tree
           (error "Can not run tla changes against a local tree"))
          (previous-revision (tla-compute-direct-ancestor
                              (cadr against)))
          (last-revision (if (string= (tla--uniquify-file-name
                                       (cadr against))
                                      (tla--uniquify-file-name
                                       (tla-tree-root)))
                             nil
                           (error "tla changes against last %s %s"
                                  "revision of local tree not"
                                  "implemented.")))
          (revision (tla--name-construct (cadr against)))))
     :finished
     `(lambda (output error status arguments)
        (if ,master-buffer
            (message "No changes in subtree %s" ,root)
          (message "No changes in %s" ,root))
        (with-current-buffer ,(current-buffer)
          (let ((inhibit-read-only t))
            (tla--changes-delete-messages)
            (ewoc-enter-last tla-changes-cookie
                             (list 'message (concat "* No changes in "
                                                    ,root ".\n\n")))
            (when ,master-buffer
              (with-current-buffer ,master-buffer
                (ewoc-map (lambda (x)
                            (when (and (eq (car x) 'subtree)
                                       (eq (cadr x) ,buffer))
                              (setcar (cdddr x) 'no-changes))
                            )
                          ;; (ewoc-refresh tla-changes-cookie)))
                          tla-changes-cookie)))
            (ewoc-refresh tla-changes-cookie))))
     :error
     `(lambda (output error status arguments)
        (if (/= 1 status)
            (tla--default-error-function output error status arguments))
        (tla-show-changes-buffer output nil ,buffer ,master-buffer)
        (when ,master-buffer
          (with-current-buffer ,master-buffer
            (ewoc-map (lambda (x)
                        (when (and (eq (car x) 'subtree)
                                   (eq (cadr x) ,buffer))
                          (setcar (cdddr x) 'changes))
                        )
                      tla-changes-cookie))))
     )))

(defun tla-changes-printer (elem)
  "ewoc pretty-printer"
  (cond
   ((eq (car elem) 'file)
    (insert (if (member (cadr elem) tla-buffer-marked-file-list) "* " "  ")
            (caddr elem) (cadddr elem) " ")
    (if (nth 4 elem) (insert (nth 4 elem) "\t"))
    (insert (cadr elem)))
   ((eq (car elem) 'subtree)
    (insert " T" (cond ((not (cadddr elem)) "?")
                       ((eq  (cadddr elem) 'changes) "M")
                       ((eq  (cadddr elem) 'no-changes) "-"))
            " " (caddr elem)))
   ((eq (car elem) 'message)
    (insert (cadr elem))))
  )

(defconst tla-verbose-format-spec
  '(("added files"    "A" " ")
    ("modified files" "M" " ")
    ("removed files"  "D" " "))
  "Internal variable used to parse the output of tla show-changeset"
  )

(defun tla-show-changes-buffer (buffer &optional verbose-format
                                       output-buffer no-switch)
  "Show the *tla-changes* buffer built from the *tla-process* buffer.
If VERBOSE-FORMAT is non-nil, the format of the *tla-process* buffer
should be the one of tla show-changeset.

Use OUTPUT-BUFFER to display changes if provided. That buffer must
already be in changes mode.

If NO-SWITCH is nil, don't switch to the created buffer."
  (let* ((root (with-current-buffer buffer
                 (tla-tree-root default-directory t)))
         (changes-buffer (or output-buffer (tla--get-buffer-create
                                            'changes root)))
         (header ""))
    (if (or no-switch tla-switch-to-buffer-first)
        (set-buffer changes-buffer)
      (tla-switch-to-buffer changes-buffer))
    (let (buffer-read-only)
      (tla--changes-delete-messages)
      (unless output-buffer
        (erase-buffer)
        (tla-changes-mode))
      (with-current-buffer buffer
        (if verbose-format
            (progn
              (goto-char (point-min))
              (while (re-search-forward
                      (concat "^\\* \\(" (regexp-opt
                                          (mapcar 'car tla-verbose-format-spec))
                              "\\)\n")
                      nil t)
                (let* ((elem (assoc (match-string 1)
                                    tla-verbose-format-spec))
                       (modif (cadr elem))
                       (dir (caddr elem)))
                  (if (string= modif "M")
                      (while (re-search-forward "^--- orig/\\(.*\\)$"
                                                nil t)
                        (let ((file (match-string 1)))
                          (with-current-buffer changes-buffer
                            (ewoc-enter-last tla-changes-cookie
                                             (list 'file file modif dir)))))
                    (while (looking-at "^$") (forward-line 1))
                    (while (looking-at
                            "^ +\\([^ ].*\\)$")
                      (let ((file (match-string 1)))
                        (with-current-buffer changes-buffer
                          (ewoc-enter-last tla-changes-cookie
                                           (list 'file file modif dir)))
                        (forward-line 1))))))
              (goto-char (point-min))
              (if (re-search-forward "^---" nil t)
                  (forward-line -1)
                (beginning-of-line)))
          (setq header (buffer-substring-no-properties
                        (goto-char (point-min))
                        (progn (re-search-forward "^[^*]" nil t)
                               (beginning-of-line)
                               (point))))
          (while (looking-at "^\\(.\\)\\([ /bfl>-]?\\) +\\([^\t\n]*\\)\\(\t\\(.*\\)\\)?$")
            (let ((file (match-string 3))
                  (modif (match-string 1))
                  (dir (match-string 2))
                  (newname (match-string 5)))
              (with-current-buffer changes-buffer
                (if newname
                    (ewoc-enter-last tla-changes-cookie
                                     (list 'file newname modif dir file))
                  (ewoc-enter-last tla-changes-cookie
                                   (list 'file file modif dir)))))
            (forward-line 1)))
        (let ((footer (concat
                       (tla--face-add (make-string  72 ?\ ) 'tla-separator)
                       "\n\n"
                       (buffer-substring-no-properties
                        (point) (point-max)))))
          (with-current-buffer changes-buffer
            (ewoc-set-hf tla-changes-cookie header footer)
            (if root (cd root)))))
      ))
  (toggle-read-only 1)
  (when (or global-font-lock-mode font-lock-mode)
    (let ((font-lock-verbose nil))
      (font-lock-fontify-buffer)))
  (if (ewoc-nth tla-changes-cookie 0)
      (goto-char (ewoc-location (ewoc-nth tla-changes-cookie 0)))))


;;;###autoload
(defun tla-delta (base modified &optional directory)
  "Runs tla delta BASE MODIFIED.
If DIRECTORY is a non-empty string, the delta is stored to it.
If DIRECTORY is ask, a symbol, ask the name of directory.
If DIRECTORY is nil or an empty string, just show the delta using --diffs."
  (interactive (list
                (tla--name-construct
                 (tla--read-archive-category-branch-version-revision-name
                  "Archive[BASE]: "
                  "Category[BASE]: "
                  "Branch[BASE]: "
                  "Version[BASE]: "
                  "Revision[BASE]: "))
                (tla--name-construct
                 (tla--read-archive-category-branch-version-revision-name
                  "Archive[MODIFIED]: "
                  "Category[MODIFIED]: "
                  "Branch[MODIFIED]: "
                  "Version[MODIFIED]: "
                  "Revision[MODIFIED]: "))
                (when current-prefix-arg
                  'ask)))

  (when (eq directory 'ask)
    (setq directory
          (tla--read-directory-name "Stored to: "
                                    (tla-tree-root default-directory t)
                                    (tla-tree-root default-directory t)
                                    nil
                                    "")))

  (when (and directory (stringp directory) (string= directory ""))
    (setq directory nil))

  (when (and directory (file-directory-p directory))
    (error "%s is already existed" directory))

  (let ((args
         (if directory
             (list "delta" base modified directory)
           (list "delta" "--diffs" base modified)))
        (run-dired-p (when directory 'ask)))
    (tla--run-tla-async args
                        :finished
                        `(lambda (output error status arguments)
                           (if ,directory
                               (tla--delta-show-directory ,directory ',run-dired-p)
                             (tla--delta-show-diff-on-buffer
                              output ,base ,modified))))))

(defun tla--delta-show-diff-on-buffer (output base modified)
  "Show the result of \"delta -diffs\""
  (with-current-buffer output
    (let ((no-changes
           ;; There were no changes if the last line of
           ;; the buffer is "* changeset report"
           (save-excursion
             (goto-char (point-max))
             (previous-line 1)
             (beginning-of-line)
             (looking-at "^* changeset report")))
          buffer)
      (if no-changes
          (message
           (concat "tla delta finished: "
                   "No changes in this arch working copy"))
        (setq buffer (tla--prepare-changes-buffer
                      (list 'revision
                            (tla--name-split base))
                      (list 'revision
                            (tla--name-split modified))))
        (tla-show-changes-buffer output nil buffer)
        (tla-switch-to-buffer buffer)
        (message "tla delta finished")))))

(defun tla--delta-show-directory (directory run-dired-p)
  (tla-show-changeset directory t)
  (tla--do-dired (concat directory "/..")  run-dired-p)
  (revert-buffer)
  (goto-char (point-min))
  (re-search-forward (concat
                      (regexp-quote (file-name-nondirectory directory))
                      "$"))
  (goto-char (match-beginning 0))
  (tla-flash-line))

;;;###autoload
(defun tla-get-changeset (revision justshow &optional destination
                                   without-diff)
  "When JUSTSHOW is non-nil, just show the diff. Otherwise, store
changeset in DESTINATION"
  (interactive "sRevision to view: \np")
  (let* ((dest (or destination
                   (make-temp-name "/tmp/tla-changeset")))
         (rev-list (tla--name-split revision))
         (buffer (and justshow
                      (tla--prepare-changes-buffer
                       (list 'previous-revision rev-list)
                       (list 'revision rev-list))))
         (tla-switch-to-buffer-mode
          (if tla-switch-to-changes-buffer
              tla-switch-to-buffer-mode 'show-in-other-window)))
    (when (and justshow tla-switch-to-buffer-first)
      (tla-switch-to-buffer buffer))
    (tla--run-tla-async
     (list "get-changeset" revision dest)
     :finished
     `(lambda (output error status arguments)
        (when ,justshow
          (tla-show-changeset ,dest ,without-diff ,buffer)
          (call-process "rm" nil nil nil "-rf"
                        ,dest))))))

(defun tla--prepare-changes-buffer (base modified)
  (with-current-buffer
      (tla--get-buffer-create 'changes ;; TODO
                              default-directory)
    (let ((inhibit-read-only t)) (erase-buffer))
    (tla-changes-mode)
    (set (make-local-variable 'tla--changes-base)     base)
    (set (make-local-variable 'tla--changes-modified) modified)
    (current-buffer)))

(defun tla-show-changeset (directory &optional without-diff buffer
                                     base modified)
  "Run tla changeset on DIRECTORY.
If prefix argument, WITHOUT-DIFF is non-nil, just show the summary.
BUFFER is the target buffer to output. If BUFFER is nil, create a new
one. "
  (interactive "DChangeset directory: \nP")
  (unless buffer
    (setq buffer (tla--prepare-changes-buffer base modified))
    (if tla-switch-to-buffer-first
        (tla-switch-to-buffer buffer)))
  (tla--run-tla-sync (list "show-changeset"
                           (unless without-diff
                             "--diffs")
                           directory)
                     :finished
                     `(lambda (output error status arguments)
                        (tla-show-changes-buffer output (not ',without-diff)
                                                 ,buffer
                                                 ,tla-switch-to-buffer-first)
                        (tla-post-switch-to-buffer))))

;;;###autoload
(defun tla-apply-changeset (changeset target &optional reverse)
  (interactive "DChangeset Directory: \nDTarget Directory: \nP")
  (if (file-directory-p changeset)
      (setq changeset (expand-file-name changeset))
    (error "%s is not directory" changeset))
  (if (file-directory-p target)
      (setq target (expand-file-name target))
    (error "%s is not directory" target))

  (or (tla-save-some-buffers target)
      (y-or-n-p
       "Apply-change may delete unsaved changes. Continue anyway? ")
      (error "Not applying"))
  (tla--apply-changeset-internal changeset target reverse)
  (when (y-or-n-p (format "Run inventory at `%s'? " target))
    (tla-inventory target)))

(defun tla--apply-changeset-internal (changeset target reverse)
  (let ((args (list changeset target)))
    (if reverse
        (setq args (cons "--reverse" args)))
    (if tla-use-forward-option
        (setq args (cons "--forward" args)))
    (setq args (cons "apply-changeset" args))
    (tla--run-tla-sync args
                       :finished `(lambda (output error status arguments)
                                    ;; (tla--show-last--process-buffer)
                                    (tla-show-changes-buffer output)
                                    (message "tla apply-changeset finished")
                                    (tla-revert-some-buffers ,target)))))


;;;###autoload
(defun tla-file-ediff-revisions (file &optional base modified)
  "View changes between BASE and MODIFIED in file FILE using ediff."
  (interactive (let ((version-list (tla-tree-version-list)))
                 (list (buffer-file-name)
                       (tla--name-construct
                        (append version-list
                                (list
                                 (apply 'tla--complete-revision
                                        version-list))))
                       (tla--name-construct
                        (append version-list
                                (list
                                 (apply 'tla--complete-revision
                                        version-list)))))))
  (tla--ediff-buffers
   (tla-file-get-revision-in-buffer file base)
   (tla-file-get-revision-in-buffer file modified)))

;;;###autoload
(defun tla-file-diff (file &optional revision)
  "Run tla file-diff on file FILE. In interactive mode, the file is
the current buffer's file."
  (interactive (list (buffer-file-name)))
  (let ()
    (tla--run-tla-async (list "file-diffs" file revision)
                        :finished
                        (lambda (output error status arguments)
                          (message "No changes in this arch working copy"))
                        :error
                        (lambda (output error status arguments)
                          (if (= 1 status)
                              (tla--show-last-process-buffer
                               'file-diff
                               'diff-mode)
                            (tla--default-error-function
                             output error status arguments))))))

(defvar tla-mine-string "TREE")
(defvar tla-his-string "MERGE-SOURCE")

;;;###autoload
(defun tla-view-conflicts (buffer)
  "*** WARNING : Use this function if you like, but M-x smerge-mode
RET is actually better for the same task ****

Graphical view of conflicts after tla star-merge --three-way. The
buffer given as an argument must be the content of a file with
conflicts markers like.

    <<<<<<< TREE
    my text
    =======
    his text
    >>>>>>> MERGE-SOURCE

Priority is given to your file by default. (This means all conflicts
will be rejected if you do nothing).
"
  (interactive (list (find-file (read-file-name "View conflicts in: "))))
  (let ((mine-buffer buffer)
        (his-buffer (get-buffer-create "*tla-his*")))
    (with-current-buffer his-buffer
      (erase-buffer)
      (insert-buffer mine-buffer)
      (goto-char (point-min))
      (while (re-search-forward (concat "^<<<<<<< "
                                        (regexp-quote tla-mine-string) "$")
                                nil t)
        (beginning-of-line)
        (delete-region (point) (progn
                                 (re-search-forward "^=======\n")))
        (re-search-forward
         (concat "^>>>>>>> "
                 (regexp-quote tla-his-string) "$"))
        (beginning-of-line)
        (delete-region (point) (1+ (line-end-position)))
        )
      )
    (with-current-buffer mine-buffer
      (goto-char (point-min))
      (while (re-search-forward (concat "^<<<<<<< "
                                        (regexp-quote tla-mine-string) "$")
                                nil t)
        (beginning-of-line)
        (delete-region (point) (1+ (line-end-position)))
        (re-search-forward "^=======$")
        (beginning-of-line)
        (delete-region (point) (progn
                                 (re-search-forward
                                  (concat "^>>>>>>> "
                                          (regexp-quote tla-his-string) "\n"))))
        ))
    (tla--ediff-buffers mine-buffer his-buffer)
    ))

(defun tla-file-get-revision-in-file (file &optional revision)
  "Get the last-committed version of FILE. Returns (file temporary).
temporary is non-nil when the file is temporary and should be
deleted."
  (case (car revision)
    (local-tree (list file nil))
    (previous-revision (tla-file-get-revision-in-file
                        file
                        (list 'revision
                              (tla-compute-direct-ancestor
                               (cadr revision)))))
    ((last-revision revision)
     (let* ((default-directory (if (eq (car revision) 'last-revision)
                                   (cadr revision)
                                 (tla-tree-root file)))
            (revision (if (eq (car revision) 'revision)
                          (tla--name-construct (cadr revision))))
            (original (progn (tla--run-tla-sync
                              (list "file-find" file revision)
                              :finished 'tla--null-handler)
                             (with-current-buffer tla--last-process-buffer
                               (goto-char (point-min))
                               (re-search-forward "^[^*]")
                               (buffer-substring-no-properties
                                (line-beginning-position)
                                (line-end-position)))))
            (original-to-be-removed nil)
            file-unmodified-p)
       (unless (file-exists-p original)
         ;; Probably tla is ran remotely or whatever. Well, get the
         ;; file using the old good tla file-diff | patch -R -o ...
         (setq original (make-temp-name "/tmp/tla-ediff")
               original-to-be-removed t)
         (tla--run-tla-sync (list "file-diffs" file revision)
                            :finished 'tla--null-handler
                            :error
                            (lambda (output error status arguments)
                              (if (not (eq status 1))
                                  (tla--default-error-function
                                   output error status arguments))))
         (with-current-buffer tla--last-process-buffer
           (if (= (point-min) (point-max))
               (setq file-unmodified-p t))
           (call-process-region (point-min) (point-max)
                                tla-patch-executable
                                nil nil nil
                                "-R" "-o" original file)))
       (list original file-unmodified-p original-to-be-removed)))))

(defun tla-file-revert (file &optional revision)
  "Reverts the file FILE to the last committed version. Warning: You
use version control to keep backups of your files. This function will
by definition not keep any backup in the archive.

Most of the time, you should not use this function. Call
`tla-file-ediff' instead, and undo the changes one by one with the key
`b', then save your buffer.

As a last chance, tla-file-revert keeps a backup of the last-saved in
~ backup file."
  (interactive (list (progn (when (and (buffer-modified-p)
                                       (or tla-do-not-prompt-for-save
                                           (y-or-n-p (format "Save buffer %s? "
                                                             (buffer-name
                                                              (current-buffer))))))
                              (save-buffer))
                            (buffer-file-name))))
  ;; set aside a backup copy
  (copy-file file (car (find-backup-file-name file)) t)

  ;; display diff
  (tla--run-tla-sync (list "file-diffs" file revision)
                     :finished
                     (lambda (output error status arguments)
                       (error "File %s is not modified!" (cadr arguments)))
                     :error
                     (lambda (output error status arguments)
                       (if (/= 1 status)
                           (tla--default-error-function
                            output error status arguments)
                         (tla--show-last-process-buffer
                          'file-diff
                          (lambda ()
                            (goto-char (point-min))
                            (let ((inhibit-read-only t))
                              (insert
                               (format "M %s\n" (cadr arguments))
                               "Do you really want to revert ALL the changes listed below?\n")
                              (if tla-highlight (font-lock-fontify-buffer)))
                            (diff-mode))))))

  (let* ((file-unmo-temp (tla-file-get-revision-in-file
                          file (list 'revision revision)))
         (original (car file-unmo-temp)))
    (unless (yes-or-no-p (format "Really revert %s? " file))
      (bury-buffer)
      (error "Not reverting file %s!" file))
    (bury-buffer)
    (copy-file original file t)
    (let ((buf (get-file-buffer file)))
      (when buf (with-current-buffer buf (revert-buffer))))))

(defun tla-undo (tree &optional archive category branch version revision)
  "Undo whole local TREE against ARCHIVE/CATEGORY-BRANCH-VERSION-REVISION.
If ARCHIVE is nil, default ARCHIVE/CATEGORY-BRANCH-VERSION-REVISION
associated with TREE."
  (interactive
   (if (y-or-n-p "Use default revision to undo? ")
       (list default-directory nil nil nil nil nil)
     (cons default-directory
           (tla--read-archive-category-branch-version-revision-name-for-tree
            default-directory "Undo against archive: "))))
  (tla--undo-internal tree archive category branch version revision))


(defun tla--undo-internal (tree &optional archive category branch version revision)
  (save-excursion (if archive
                      (tla-changes nil (tla--name-construct
                                        archive category branch version revision))
                    (tla-changes)))
  (when (yes-or-no-p
         (if archive
             (format "Revert whole local tree (%s) from `%s'?"
                     tree (tla--name-construct
                           archive category branch version revision))
           (format "Revert whole local tree (%s) from default revision" tree)))
    (let ((default-directory tree))
      (tla--run-tla-sync (if archive
                             (list "undo" (tla--name-construct
                                           archive category branch version revision))
                           (list "undo"))))
    (tla-revert-some-buffers tree)))

;;;###autoload
(defun tla-file-ediff (file &optional revision)
  "Interactive view of differences since last commit (or REVISION if
specified) using ediff."
  (interactive (list (progn (when (and (buffer-modified-p)
                                       (y-or-n-p (format "Save buffer %s? "
                                                         (buffer-name
                                                          (current-buffer)))))
                              (save-buffer))
                            (buffer-file-name))))
  (let ((original (tla-file-get-revision-in-buffer
                   file (or revision (list 'last-revision
                                           (tla-tree-root))))))
    (when (string= (with-current-buffer original (buffer-string))
                   (buffer-string))
      (error "No modification in this file"))
    (tla--ediff-buffers (or (get-file-buffer file)
                            (find-file-noselect file))
                        original)))

;;;###autoload
(defun tla-file-view-original (file &optional revision)
  "Gets the last-committed version of FILE in a buffer."
  (interactive (list (buffer-file-name)))
  (let ((original (tla-file-get-revision-in-buffer
                   file (or revision (list 'last-revision
                                           (tla-tree-root))))))
    (when (string= (with-current-buffer original (buffer-string))
                   (buffer-string))
      (message "No modification in this file"))
    (tla-switch-to-buffer original)))

(defun tla--buffer-for-rev (file revision)
  "Returns an empty buffer with a suitable name  for viewing REVISION
of FILE. REVISION may have one of the values described in the
docstring of `tla--changes-modified' or `tla--changes-base'."
  (let ((name (concat
               (file-name-nondirectory file)
               "(" (cond
                    ((eq (car revision) 'revision)
                     (tla--name-construct (cadr revision)))
                    ((eq (car revision) 'local-tree)
                     (cadr revision))
                    ((eq (car revision) 'last-revision) "original")
                    ((eq (car revision) 'previous-revision)
                     (tla--name-construct-semi-qualified
                      (tla-compute-direct-ancestor (cadr revision))))
                    (t ""))
               ")")))
    (get-buffer-create
     (create-file-buffer name))))

(defun tla-file-get-revision-in-buffer (file &optional revision)
  "Gets the last committed version of file in a buffer. Returned value
is the buffer.

REVISION can have any of the values described in the docstring of
`tla--changes-base' and `tla--changes-modified'"
  (let* ((default-directory (tla-tree-root))
         (file-unmo-temp (tla-file-get-revision-in-file file revision))
         (original (car file-unmo-temp))
         (original-to-be-removed (cadr file-unmo-temp)))
    (if (eq (car revision) 'local-tree)
        (find-file-noselect original)
      (let ((buffer-orig (tla--buffer-for-rev file revision)))
        (with-current-buffer buffer-orig
          (erase-buffer)
          (insert-file-contents original)
          (when original-to-be-removed
            (delete-file original)))
        buffer-orig))))

(defun tla-ediff-startup-hook ()
  ;; ediff-after-quit-hook-internal is local to an ediff session.
  (add-hook 'ediff-after-quit-hook-internal
            `(lambda ()
               (set-window-configuration
                ,tla-window-config))
            nil 'local))

;;;###autoload
(defun tla-commit ()
  "Runs tla commit.
Returns the exit status of tla"
  (interactive)
  (or (tla-save-some-buffers)
      (y-or-n-p
       "Commit with unsaved changes is a bad idea. Continue anyway? ")
      (error "Not committing"))
  (let* ((file-list (and (buffer-live-p tla-buffer-source-buffer)
                         (with-current-buffer tla-buffer-source-buffer
                           tla-buffer-marked-file-list)))
         arglist)
    (when file-list (setq arglist (append arglist (cons "--"
                                                        file-list))))
    ;; raises an error if commit isn't possible
    (tla--run-tla-sync
     (cons "commit"
           (cons (when tla-strict-commits "--strict")
                 (when file-list (cons "--"
                                       file-list)))))
    (let ((changes (tla--get-buffer 'changes (tla-tree-root))))
      (when changes
        (with-current-buffer changes
          (let ((inhibit-read-only t))
            (erase-buffer)
            (insert "* Just committed ! Please refresh buffer")))))
    )
  0)

;;;###autoload
(defun tla-rm (file)
  "Calls tla rm on file FILE. Prompts for confirmation before"
  (when (yes-or-no-p (format "Delete file %s? " file))
    (tla--run-tla-sync (list "rm" file)
                       :finished 'tla--null-handler)))

(defun tla-pristines ()
  (interactive)
  (tla--run-tla-sync '("pristines")))

;;;###autoload
(defun tla-changelog ()
  (interactive)
  (let ((default-directory (tla--read-project-tree-maybe)))
    (tla--run-tla-sync '("changelog")
                       :finished 'tla--null-handler)
    (tla--show-last-process-buffer 'changelog 'tla-changelog-mode)
    (goto-char (point-min))))

;;;###autoload
(defun tla-logs ()
  "Run tla logs"
  (interactive)
  (let ((default-directory (tla--read-project-tree-maybe))
;        (details (or tla-revisions-shows-date
;                     tla-revisions-shows-creator
;                     tla-revisions-shows-summary))
        )
    (tla--run-tla-async
     (list "logs" "--full"
                                        ;           (when details "--date")
                                        ;           (when details "--creator")
                                        ;           (when details "--summary"))
           )
     :finished
     `(lambda (output error status arguments)
        (let ((buffer (tla--get-buffer-create 'logs (tla-tree-root))))
          (tla-switch-to-buffer buffer)
          (tla-revision-list-mode)
          (tla--revisions-parse-list 'logs nil ;;,details
                                     nil ;; TODO (merges)
                                     output nil
                                     tla-revision-list-cookie)
          (set (make-local-variable 'tla-buffer-refresh-function)
               'tla-logs))))))

;; ;;;###autoload
;; (defun tla-tree-lint ()
;;   "Audit an arch source tree."
;;   (interactive)
;;   (let ((default-directory (tla--read-project-tree-maybe
;;                             "Run tla tree-lint in: ")))
;;     (tla--run-tla-async
;;      '("tree-lint")
;;      :finished
;;      `(lambda (output error status arguments)
;;         (message "No tree-lint warnings for %s." ,default-directory))
;;      :error
;;      (lambda (output error status arguments)
;;        (save-excursion
;;          (tla--show-last-process-buffer 'tree-lint))))))

;;;###autoload
(defun tla-help (command)
  "Runs tla COMMAND -H"
  (interactive
   (list (completing-read
          "Get help for: "
          (tla--run-tla-sync
           '("help")
           :finished
           `(lambda (output error status arguments)
              (with-current-buffer output
                (goto-char (point-min))
                (let (listcmd)
                  (while (re-search-forward
                          " *\\([^ ]*\\) : " nil t)
                    (setq listcmd
                          (cons (list (match-string 1))
                                listcmd)))
                  listcmd)))))))
  (tla--run-tla-sync (list command "-H")))

(defun tla-tree-version-list-tla ()
  "Returns the tree version, or nil if not in a project tree"
  (tla--run-tla-sync '("tree-version")
                     :finished
                     (lambda (output error status arguments)
                       (with-current-buffer output
                         (and
                          (goto-char (point-min))
                          (re-search-forward "\\(.*\\)/\\(.*\\)--\\(.*\\)--\\(.*\\)" nil t)
                          (list (match-string 1)
                                (match-string 2)
                                (match-string 3)
                                (match-string 4)))))))

(defun tla-tree-version-list (&optional location no-error)
  "Elisp implementation of `tla-tree-version-list-tla'
A string, LOCATION is used as a directory where \"/{arch}/++default-version\" is.
If NO-ERROR is non-nil, errors are not reported; just return nil."
  (let ((version-string (tla-tree-version location no-error)))
    (and version-string
         (string-match "\\(.*\\)/\\(.*\\)--\\(.*\\)--\\(.*\\)" version-string)
         (list (match-string 1 version-string)
               (match-string 2 version-string)
               (match-string 3 version-string)
               (match-string 4 version-string)))))

(defun tla-tree-root-tla ()
  "Run tla tree-root."
  (interactive)
  (tla--run-tla-sync '("tree-root")
                     :finished
                     `(lambda (output error status arguments)
                        (let ((result (tla--buffer-content output)))
                          (when ,(interactive-p)
                            (message "tla tree-root is: %s"
                                     result))
                          result))))

(defun tla-tree-version (&optional location no-error)
  "Equivalent of tla tree-version (but implemented in pure elisp)"
  (interactive (list nil nil))
  (let* ((tree-root (tla-tree-root location no-error))
         (default-version-file (when tree-root
                                 (expand-file-name
                                  "{arch}/++default-version"
                                  tree-root)))
         (version (and (boundp 'tla-buffer-version-name)
                       tla-buffer-version-name)))
    (if (and (null version)
             default-version-file
             (file-readable-p default-version-file))
        (with-temp-buffer
          (insert-file-contents default-version-file)
          (setq version (buffer-substring-no-properties (point-min)
                                                        (- (point-max) 1)))))
    (when (interactive-p)
      (message "%s" version))
    version))

;;;###autoload
(defun tla-my-id (&optional arg)
  "Run tla my-id.
When called without a prefix argument, just print the my-id from tla.
When called with a prefix argument, ask for a new my-id.

The my-id should have the following format:

Your id is recorded in various archives and log messages
as you use arch. It must consist entirely of printable
characters and fit on one line. By convention, it should
have the form of an email address, as in this example:

Jane Hacker <jane.hacker@gnu.org>
"
  (interactive "P")
  (let ((id (tla--run-tla-sync '("my-id")
                               :finished
                               (lambda (output error status arguments)
                                 (tla--buffer-content output)))))
    (if arg
        ;; Set the user's ID
        (let ((new-id (read-string "New arch my-id: "
                                   id tla-my-id-history id)))
          (if (string= id new-id)
              (message "Id unchanged! Id = %s" new-id)
            (message "Setting id to: %s" new-id)
            (tla--run-tla-sync (list "my-id" new-id)
                               :finished (lambda (output error status arguments)
                                           (message "Id changed"))))
          new-id)
      (cond (id (when (interactive-p)
                  (message "Arch my-id: %s" id))
                id)
            (t (message (concat "Arch my-id has not been given yet. "
                                "Call `%s' with prefix arguments to set.")
                        this-command))))))

(defun tla-set-my-id ()
  "Set tla's my-id. "
  (interactive)
  (tla-my-id 1))

;;
;; Library
;;

;;;###autoload
(defun tla-my-revision-library (&optional arg)
  "Run tla my-revision-library.
When called without a prefix argument, just print the my-revision-library from tla.
When called with a prefix argument, ask for a new my-revision-library.

my-revision-library specifies a path, where the revision library is stored to
speed up tla. For example ~/tmp/arch-lib.

You can configure the parameters for the library via tla-library-config."
  (interactive "P")
  (let ((result (tla--run-tla-sync '("my-revision-library")
                                   :finished 'tla--status-handler
                                   :error 'tla--null-handler))
        (rev-lib (tla--get-process-output)))
    (when (eq 0 result)
      (if arg
          (tla--library-add-interactive rev-lib)
        (if (and rev-lib (string= "" rev-lib))
            (message "Arch my-revision-library has not been given yet. Call `%s' with prefix arguments to set."
                     this-command)
          (when (interactive-p) (message "Arch my-revision-library: %s" rev-lib)))
        rev-lib))))

(defun tla--library-add-interactive (&optional old-rev-lib)
  (unless old-rev-lib (setq old-rev-lib ""))
  (let ((new-rev-lib (expand-file-name (tla--read-directory-name
                                        "New arch revision library: " old-rev-lib))))
    (if (not (string= old-rev-lib new-rev-lib))
        (progn
          (message "Setting my-revision-library to: %s" new-rev-lib)
          (tla--library-add new-rev-lib))
      old-rev-lib)))

(defun tla-library-delete (rev-lib)
  "unregister revision library REV-LIB"
  (interactive (list (tla--read-revision-library)))
  (tla--run-tla-sync (list "my-revision-library" "--delete" rev-lib)
                     :finished (lambda (output error status arguments)
                                 (message "Library %s removed."
                                          rev-lib))))

(defun tla--library-add (new-rev-lib)
  (let ((dir-attr (file-attributes new-rev-lib)))
    (unless dir-attr
      (make-directory new-rev-lib t))
    (tla--run-tla-sync (list "my-revision-library" new-rev-lib)
                       :finished
                       (lambda (output error status arguments)
                         (message (tla--buffer-content output))))
    new-rev-lib))

(defun tla--revision-library-list ()
  "Runs tla my-revision-library, and parses the output into a list of
revision libraries"
  (tla--run-tla-sync '("my-revision-library")
                     :finished
                     'tla--output-buffer-split-handler))

(defvar tla--library-history nil)

(defun tla--read-revision-library (&optional prompt)
  "Reads a revision library from keyboard"
  (let ((list-lib (tla--revision-library-list)))
    (if (null (cdr list-lib))
        (car list-lib)
      (completing-read (or prompt
                           (format "Revision library (default %s): "
                                   (car list-lib)))
                       (mapcar 'list (tla--revision-library-list))
                       nil t nil tla--library-history
                       (car list-lib)))))

(defun tla-library-config (&optional arg)
  "Run tla library-config.
When called without a prefix argument, just print the config.
When called with a prefix argument, let the user change the config."
  (interactive "P")
  (let ((rev-lib (tla--read-revision-library))
        (config-param (when arg
                        (completing-read "tla library config "
                                         (mapcar 'list '("--greedy"
                                                         "--sparse"
                                                         "--non-greedy"
                                                         "--non-sparse"))
                                         nil t "--"))))
    (tla--run-tla-sync (list "library-config" config-param rev-lib)
                       :finished 'tla--null-handler)
    (message (tla--get-process-output))))

(defun tla-library-add (archive category branch version revision)
  (tla--show-last-process-buffer)
  (tla--run-tla-async `("library-add"
                        ,(tla--name-construct archive category
                                              branch version
                                              revision))))

(defun tla-library-find (archive category branch version revision
                                 &optional silent)
  "Run tla library-find.
If the revision is found, return the path for it. Else return nil."
  (if (zerop (tla--run-tla-sync (list "library-find" (when silent "--silent")
                                      (tla--name-construct
                                       archive category branch
                                       version revision))
                                :finished 'tla--status-handler
                                :error 'tla--status-handler))
      (tla--get-process-output)))

;; completing-read: tagline, explicit, names, implicit
(defvar tla-id-tagging-method-history nil)
;;;###autoload
(defun tla-id-tagging-method (arg)
  "View or change the id-tagging method.
When called without a prefix argument: show the actual tagging method.
When called with a prefix argument: Ask the user for the new tagging method."
  (interactive "P")
  (let ((tm (progn (tla--run-tla-sync '("id-tagging-method")
                                      :finished
                                      (lambda (output error status arguments)
                                        (tla--buffer-content output)))))
        (new-tagging-method))
    (if arg
        (progn
          (setq new-tagging-method
                (completing-read
                 (format "New id tagging method (default %s): " tm)
                 (mapcar 'list '("tagline" "explicit" "names" "implicit"))
                 nil t nil
                 tla-id-tagging-method-history
                 tm))
          (when (not (string= tm new-tagging-method))
            (message "Setting tagging method to: %s" new-tagging-method)
            (tla--run-tla-sync (list "id-tagging-method"
                                     new-tagging-method)
                               :finished 'tla--null-handler)))
      (message "Arch id tagging method: %s" tm))))

(defun tla-archive-mirror (archive &optional category branch version from)
  "Synchronize the mirror for the archive ARCHIVE. Limit to
CATEGORY--BRANCH--VERSION. If FROM is provided, mirror from it."
  (interactive (tla--read-archive-name))
  (let ((name (tla--name-construct-semi-qualified category branch version)))
    (when (string= name "") (setq name nil))
    (tla--run-tla-async (list "archive-mirror"
                              archive
                              name
                              from)
                        :finished `(lambda (output error status arguments)
                                     (message "tla archive-mirror finished"))
                        )))

(defun tla-star-merge (from &optional to-tree)
  ;;(message "run tla star-merge %s" from)
  (or (tla-save-some-buffers (or to-tree default-directory))
      (y-or-n-p
       "Update may delete unsaved changes. Continue anyway? ")
      (error "Not updating"))
  (let* ((default-directory (or to-tree default-directory))
         (arglist '())
         (buffer (tla--prepare-changes-buffer
                  (list 'last-revision default-directory)
                  (list 'local-tree default-directory))))
    (when tla-switch-to-buffer-first
      (tla-switch-to-buffer buffer))
    (when tla-three-way-merge (add-to-list 'arglist "--three-way"))
    (when tla-use-forward-option (add-to-list 'arglist "--forward"))
    (tla--run-tla-async `("star-merge" ,@arglist ,from)
                        :finished `(lambda (output error status arguments)
                                     ;; (tla--show-last--process-buffer)
                                     (tla-show-changes-buffer
                                      output nil ,buffer)
                                     (message "tla star-merge finished")
                                     (tla-revert-some-buffers ,to-tree))
                        :error `(lambda (output error status arguments)
                                  (case status
                                    ;; 2 stands for an error.
                                    (2 (tla--default-error-function
                                        output error status arguments))
                                    ;; How about other status?
                                    (otherwise (tla-show-changes-buffer output)
                                               output nil ,buffer))))))

(defun tla-replay (from &optional to-tree arg)
  "Replay the revision FROM into tree TO-TREE.  If FROM is a string, it should
be a fully qualified revision.  If FROM is a list, it should be a list of
fully qualified revisions to be replayed.

If ARG is non-nil, replay all the version instead of the revision only."
  (interactive (list
                (tla--name-construct
                 (if current-prefix-arg
                     (tla--read-archive-category-branch-version-name
                      "Replay patch: Archive ")
                   (tla--read-archive-category-branch-version-revision-name
                    "Replay patch: Archive ")))
                (tla--read-directory-name "Replay in tree: ")
                current-prefix-arg))
  (or (tla-save-some-buffers (or to-tree default-directory))
      (y-or-n-p
       "Replay may delete unsaved changes. Continue anyway? ")
      (error "Not replaying"))
  (tla--show-last-process-buffer)
  (when to-tree (cd to-tree))
  (tla--run-tla-async `("replay"
                        ,(when tla-use-forward-option "--forward")
                        ,@(if (listp from)
                              from
                            (list from)))
                      :finished `(lambda (output error status arguments)
                                   (tla-show-changes-buffer output)
                                   (message "tla replay finished")
                                   (tla-revert-some-buffers ,to-tree))
                      :error (lambda (output error status arguments)
                               (tla-show-changes-buffer output))))

(defun tla-sync-tree (from &optional to-tree)
  (interactive (list
                (tla--name-construct
                 (tla--read-archive-category-branch-version-revision-name
                  "Sync tree with revision: Archive "))
                (tla--read-directory-name "Sync tree: ")))
  (or (tla-save-some-buffers (or to-tree default-directory))
      (y-or-n-p
       "Update may delete unsaved changes. Continue anyway? ")
      (error "Not updating"))
  (tla--show-last-process-buffer)
  (when to-tree (cd to-tree))
  (tla--run-tla-async `("sync-tree" ,from)
                      :finished `(lambda (output error status arguments)
                                   (tla--show-last-process-buffer)
                                   (message "tla sync-tree finished")
                                   (tla-revert-some-buffers ,to-tree))
                      :error (lambda (output error status arguments)
                               (tla-show-changes-buffer output))))

(defun tla-tag (source-revision tag-version)
  "Runs tla tag --setup"
  (tla--run-tla-sync (list "tag" "--setup"
                           source-revision tag-version)))

;;
;; Xtla bookmarks
;;

(defvar tla-bookmarks-loaded nil
  "Whether tla-bookmarks have been loaded from file")

(defvar tla-bookmarks-alist nil
  "Alist containing xtla bookmarks.")

(defvar tla-bookmarks-show-details nil
  "Whether tla-bookmarks should show bookmark details")

(defvar tla-bookmarks-cookie nil
  "ewoc dll")

(defvar tla-missing-buffer-todolist nil
  "List of (kind info) which can be
 (separator \"label\" bookmark \"local-tree\")
 (changes \"local-tree\")
 (missing \"local-tree\" \"location\" \"bookmark-name\")")

(defvar tla-bookmarks-marked-list nil
  "list of marked bookmarks")

(defun tla-bookmarks-load-from-file (&optional force)
  (when (or force (not tla-bookmarks-loaded))
    (let ((file (expand-file-name tla-bookmarks-file-name)))
      (save-excursion
        (unless (file-exists-p file)
          (with-temp-buffer
            (insert "()")
            (write-file file)))
        (unless (file-readable-p file)
          (error "Xtla bookmark file not readable"))
        (with-temp-buffer
          (insert-file-contents file)
          (setq tla-bookmarks-alist (read (current-buffer))
                tla-bookmarks-loaded t))))))

(defun tla-bookmarks-save-to-file ()
  "Saves `tla-bookmarks-alist' to a file"
  (let ((print-readably t)
        print-level print-length
        (file (expand-file-name tla-bookmarks-file-name)))
    (with-temp-buffer
      (insert (pp-to-string tla-bookmarks-alist))
      (write-file file))))

(defun tla-bookmarks-toggle-details (&optional val)
  "toggles the value of `tla-bookmarks-toggle-details'"
  (interactive "P")
  (let ((current-bookmark (ewoc-locate tla-bookmarks-cookie)))
    (setq tla-bookmarks-show-details
          (if val
              (if (> val 0) t
                (if (< val 0) nil
                  (not tla-bookmarks-show-details)))
            (not tla-bookmarks-show-details)))
    (ewoc-refresh tla-bookmarks-cookie)
    (tla-bookmarks-cursor-goto current-bookmark)))

(defvar tla-bookmarks-align 19
  "Position, in chars, of the `:' when displaying the bookmarks
buffer.")

(defun tla-bookmarks-printer (element)
  "Pretty printer used by ewoc, printing an entry of the bookmark
list"
  (insert (if (member element tla-bookmarks-marked-list) " *" "  "))
  (tla--insert-right-justified (concat (car element) ": ")
                               (- tla-bookmarks-align 2)
                               'tla-bookmark-name)
  (insert (tla--face-add (tla--name-construct
                          (cdr (assoc 'location (cdr element))))
                         'tla-revision-name
                         tla-bookmarks-entry-map
                         tla-bookmarks-entry-menu
                         ))
  (when tla-bookmarks-show-details
    (newline)
    (insert-char ?\  tla-bookmarks-align)
    (insert (cdr (assoc 'timestamp (cdr element))))
    (newline)
    (let ((notes (assoc 'notes (cdr element))))
      (when notes
        (insert-char ?\  tla-bookmarks-align)
        (insert (cdr notes))
        (newline)))
    (let ((nickname (assoc 'nickname (cdr element))))
      (when nickname
        (tla--insert-right-justified "nickname: " tla-bookmarks-align)
        (insert (cadr nickname))
        (newline)))
    (let ((partners (assoc 'partners (cdr element))))
      (when partners
        (tla--insert-right-justified "partners: " tla-bookmarks-align)
        (insert (cadr partners))
        (dolist (x (cddr partners))
          (insert ",\n")
          (insert-char ?\  tla-bookmarks-align)
          (insert x))
        (newline)))
    (let ((local-tree (assoc 'local-tree (cdr element))))
      (when local-tree
        (tla--insert-right-justified "local trees: " tla-bookmarks-align)
        (insert (cadr local-tree))
        (dolist (x (cddr local-tree))
          (insert ", " x ))
        (newline)))
    (let ((groups (assoc 'groups (cdr element))))
      (when groups
        (tla--insert-right-justified "Groups: " tla-bookmarks-align)
        (insert (cadr groups))
        (dolist (x (cddr groups))
          (insert ", " x ))
        (newline)))))

(defvar tla-revision-list-cookie nil
  "ewoc cookie for tla-bookmark-missing")

(defun tla-bookmarks-read-local-tree (&optional bookmark arg)
  "Reads a local tree from keyboard for bookmark BOOKMARK, and
possibly add it to the bookmarks. If arg is non-nil, user will be
prompted anyway. Otherwise, just use the default if it exists."
  (let* ((bookmark (or bookmark
                       (ewoc-data (ewoc-locate
                                   tla-bookmarks-cookie))))
         (local-trees (assoc 'local-tree (cdr bookmark))))
    (cond
     ((not local-trees)
      (let ((dir (tla--read-directory-name
                  (format "Local tree for \"%s\": "
                          (car bookmark)))))
        (when (y-or-n-p "Add this tree in your bookmarks? ")
          (tla-bookmarks-add-tree bookmark dir))
        dir))
     (arg
      ;; multiple local trees.
      (let ((dir (completing-read
                  (format "Local tree for \"%s\": "
                          (car bookmark))
                  (mapcar #'(lambda (x) (cons x nil))
                          (cdr local-trees))
                  nil nil nil nil (cadr local-trees))))
        (when (and (not (member dir (cdr local-trees)))
                   (y-or-n-p "Add this tree in your bookmarks? "))
          (tla-bookmarks-add-tree bookmark dir))
        (when (and (not (string=
                         dir (cadr local-trees)))
                   (y-or-n-p "Make this the default? "))
          (tla-bookmarks-delete-tree bookmark dir)
          (tla-bookmarks-add-tree bookmark dir))
        dir))
     (t (cadr local-trees)))))

(defun tla-bookmarks-missing (&optional arg)
  "Show the missing patches from your partners.
The missing patches are received via tla missing.
Additionally the local changes in your working copy are also shown."
  (interactive "P")
  (let ((list (or tla-bookmarks-marked-list
                  (list (ewoc-data (ewoc-locate
                                    tla-bookmarks-cookie))))))
    (set-buffer (tla--get-buffer-create 'missing))
    (tla-revision-list-mode)
    (set (make-local-variable 'tla-buffer-refresh-function)
         'tla-missing-refresh)
    (let ((tla-bookmarks-missing-buffer-list-elem
           (mapcar
            #'(lambda (elem)
                (cons
                 elem
                 (tla-bookmarks-read-local-tree elem arg)))
            list)))
      (set (make-local-variable 'tla-missing-buffer-todolist)
           (reverse
            (apply 'append
                   (mapcar (lambda (elem)
                             (tla-bookmarks-missing-elem
                              (car elem) arg (cdr elem) t t))
                           tla-bookmarks-missing-buffer-list-elem))))
      (tla-missing-refresh))))

(defvar tla--nb-active-processes 1
  "nummber of active processes in this buffer. Used internally as a
counter to launch a global handler when all processes have finished.")

(defun tla-missing-refresh (&optional arg)
  (interactive)
  (set (make-local-variable 'tla--nb-active-processes) 1)
  (let ((buffer-read-only nil))
    (erase-buffer)
    (set (make-local-variable 'tla-revision-list-cookie)
         (ewoc-create 'tla-revision-list-printer))
    (tla--kill-process-maybe (current-buffer))
    (dolist (item tla-missing-buffer-todolist)
      (case (car item)
        (missing
         ;; This item is a version that we want to check for missing patches.
         ;; ITEM is of the form:
         ;; (missing <local tree> <fully qualified version> [bookmark name])
         (let* ((local-tree (nth 1 item))
                (version (nth 2 item))
                (bookmark-name (nth 3 item))
                (text (if bookmark-name
                          (format "Missing patches from partner %s:"
                                  bookmark-name)
                        (concat "Missing patches from archive " version)))
                (node (ewoc-enter-last tla-revision-list-cookie
                                       (list 'separator (concat
                                                         text)
                                             'partner))))
           (ewoc-enter-last tla-revision-list-cookie
                            '(entry-change "Checking for missing patches..."))
           (let ((default-directory local-tree))
             (cd default-directory)
             (setq tla--nb-active-processes
                   (+ tla--nb-active-processes 1))
             (tla--run-tla-async
              `("missing" "--full" ;;"--summary" "--creator" "--date"
                ,version)
              :finished
              `(lambda (output error status arguments)
                 (when (and (tla--get-buffer 'missing)
                            (buffer-live-p (tla--get-buffer 'missing)))
                   (with-current-buffer (tla--get-buffer-create 'missing)
                     (when (ewoc-p tla-revision-list-cookie)
                       (let* ((cookie tla-revision-list-cookie)
                              (to-delete (ewoc-next cookie ,node)))
                         (tla--revisions-parse-list
                          'missing nil
                          nil ;; TODO (merges)
                          output ,node cookie
                          'tla-revision-compute-merged-by
                          )
                         (ewoc--node-delete to-delete)
                         (ewoc-refresh tla-revision-list-cookie))))))
              :error
              `(lambda (output error status arguments)
                 (when (and (tla--get-buffer 'missing)
                            (buffer-live-p (tla--get-buffer 'missing)))
                   (with-current-buffer (tla--get-buffer-create 'missing)
                     (when (ewoc-p tla-revision-list-cookie)
                       (let* ((cookie tla-revision-list-cookie)
                              (to-delete (ewoc-next cookie ,node)))
                         (setcdr (ewoc-data to-delete) '("Error in tla process"))))))
                 (message "Abnormal exit with code %d!\n%s" status
                          (tla--buffer-content error)))))))
        (separator
         ;; This item is a separator -- the name of a bookmark.
         ;; ITEM is of the form:
         ;; (separator <text> bookmark <local tree>)
         (let* ((text (nth 1 item))
                (local-tree (nth 3 item)))
           (ewoc-enter-last tla-revision-list-cookie
                            (list 'separator
                                  text
                                  'bookmark
                                  local-tree))))
        (changes
         ;; This item is a local-tree that should be checked for changes.
         ;; ITEM is of the form:
         ;; (changes <local tree>)
         (let ((default-directory (nth 1 item)))
           (tla--run-tla-async
            '("changes")
            :error `(lambda (output error status arguments)
                      (tla-bookmarks-missing-parse-changes
                       output ,(ewoc-nth tla-revision-list-cookie
                                         -1)))
            :finished 'tla--null-handler))))
      (ewoc-set-hf tla-revision-list-cookie ""
                   (concat "\n" (tla--face-add "end."
                                               'tla-separator)))))
  ;; If all processes have been run synchronously,
  ;; tla--nb-active-processes is 1 now, and we should run the
  ;; callback.
  (setq tla--nb-active-processes
        (- tla--nb-active-processes 1))
  (when (zerop tla--nb-active-processes)
    (tla-revision-compute-merged-by))
  )

(defun tla--revision-ewoc-map (function ewoc-list)
  "like (ewoc-map FUNCTION EWOC-LIST), but call FUNCTION only on
'entry-patch nodes. The argument passed to FUNCTION is a struct of
type tla--revisions."
  (ewoc-map (lambda (elem)
              (when (eq (car elem) 'entry-patch)
                (funcall function (caddr elem))))
            ewoc-list))

(defvar tla-revision-merge-by-computed nil
  "Non-nil when the \"merged-by\" field have been computed")

(defun tla-revision-compute-merged-by ()
  "In a revision list buffer, with revisions containing the
\"merges:\" information, compute another field \"merged-by:\",
containing the reverse information. If revision-A is a merge of
revision-B, then, you'll get
revision-A
  merges: revision-B
revision-B
  merged-by: revision-A"
  (interactive)
  (tla--revision-ewoc-map (lambda (elem)
                            (setf (tla--revision-merged-by elem) nil))
                          tla-revision-list-cookie)
  (tla--revision-ewoc-map 'tla--set-merged-patches
                          tla-revision-list-cookie)
  (tla--revision-ewoc-map (lambda (elem)
                            (unless (tla--revision-merged-by elem)
                              (setf (tla--revision-merged-by elem) 'nobody)))
                          tla-revision-list-cookie)
  (set (make-local-variable 'tla-revision-merge-by-computed) t)
  )

(defun tla--set-merged-patches (rev)
  (dolist (merged-rev (tla--revision-merges rev))
    (tla--revision-ewoc-map
     `(lambda (rev-to-fill)
        (when (equal (tla--name-construct
                      (tla--revision-revision rev-to-fill))
                     merged-rev)
          (setf (tla--revision-merged-by rev-to-fill)
                (cons ,(tla--name-construct
                        (tla--revision-revision rev))
                      (tla--revision-merged-by rev-to-fill)))))
     tla-revision-list-cookie)))

(defun tla-bookmarks-missing-elem (data arg local-tree header
                                        &optional changes-too)
  "Show missing patches for one element"
  (let* ((partners (assoc 'partners (cdr data)))
         (location (cdr (assoc 'location (cdr data)))))
    (tla-switch-to-buffer (tla--get-buffer-create 'missing))
    (cd local-tree)
    (let ((item '()))
      (add-to-list 'item
                   `(separator
                     ,(format "Bookmark %s (%s):"
                              (car data)
                              (tla--name-construct location))
                     bookmark
                     ,local-tree))
      (when changes-too
        (add-to-list 'item `(changes ,local-tree)))
      (dolist (partner (cons (tla--name-construct
                              (cdr (assoc 'location (cdr data)))) ; Me
                             (cdr partners))) ; and my partners
        (let ((bookmark-name
               (some (lambda (bookmark)
                       (and (string= partner
                                     (tla--name-construct
                                      (cdr (assoc 'location bookmark))))
                            (car bookmark)))
                     tla-bookmarks-alist)))
          (add-to-list 'item `(missing ,local-tree ,partner ,bookmark-name))))
      item)))

(defun tla--read-field (field)
  "Must be called from a log file buffer. Returns the content of the
field FIELD. FIELD is just the name of the field, without trailing
\": \""
  (save-excursion
    (goto-char (point-min))
    (if (re-search-forward (concat "^" field ": ") nil t)
        (buffer-substring-no-properties
         (point) (progn
                   (re-search-forward "^[^ \t]")
                   (- (point) 2))) ;; back to the end of the last line
      ;; of the field.
      "")))

(defun tla--revisions-parse-list (type details merges buffer
                                       parent-node cookie
                                       &optional callback)
  "Parses a list of revisions. TYPE can be either 'logs, 'missing, but
could be extended in the future.

DETAILS must be non-nil if the buffer contains date, author and
summary.
MERGES must be non-nil if the buffer contains list of merged patches
for each revision.
BUFFER is the buffer to parse.

PARENT-NODE is an ewoc node to which the new items will be appened. If
nil, append at the end of the ewoc list.
COOKIE must be the ewoc list containing PARENT-NODE.
"
  (with-current-buffer (ewoc-buffer cookie)
    (set (make-local-variable 'tla-revision-merge-by-computed) nil))
  (let ((last-node parent-node)
        revision date creator summary merges)
    (with-current-buffer (with-current-buffer buffer
                           (clone-buffer))
      (goto-char (point-min))
      (re-search-forward ".*/.*--.*--.*--.*" nil t)
      (beginning-of-line)
      (while (progn (> (point-max) (point)))
        (setq revision (buffer-substring-no-properties
                        (point) (line-end-position)))
        (forward-line 1)
        (when details
          (cond ((or (eq type 'logs)
                     (eq type 'missing))
                 (re-search-forward " +\\(.*[^ ]\\)   *\\(.*\\)"
                                    (line-end-position))
                 (setq date (match-string-no-properties 1))
                 (setq creator (match-string-no-properties 2))
                 (forward-char 1))
                (t (error "type of list not managed (yet)")))
          (if (re-search-forward " +" (line-end-position) t)
              (progn (setq summary (buffer-substring-no-properties
                                    (point) (line-end-position)))
                     (forward-line 1))
            (setq summary "")))
        (let* ((rev-struct (make-tla--revision
                            :revision (tla--name-split revision)))
               (elem (list 'entry-patch nil
                           rev-struct)))
          (when (or tla-revisions-shows-summary
                    tla-revisions-shows-creator
                    tla-revisions-shows-date
                    tla-revisions-shows-merges
                    tla-revisions-shows-merged-by)
            (with-current-buffer (ewoc-buffer cookie)
              (setq tla--nb-active-processes
                    (+ tla--nb-active-processes 1))
              (tla--cat-log-any
               (tla--name-split revision)
               nil
               `(lambda (output error status arguments)
                  (with-current-buffer output
                    (setf (tla--revision-date ,rev-struct)
                          (tla--read-field "Standard-date"))
                    (setf (tla--revision-creator ,rev-struct)
                          (tla--read-field "Creator"))
                    (setf (tla--revision-summary ,rev-struct)
                          (tla--read-field "Summary"))
                    (setf (tla--revision-merges ,rev-struct)
                          (remove ,revision
                                  (split-string (tla--read-field
                                                 "New-patches")))))
                  (with-current-buffer ,(ewoc-buffer cookie)
                    (setq tla--nb-active-processes
                          (- tla--nb-active-processes 1))
                    (when (and ',callback
                               (zerop tla--nb-active-processes))
                      (funcall ',callback)))
                  (ewoc-refresh ,cookie)))))
          (if last-node
              (setq last-node
                    (ewoc-enter-after cookie last-node elem))
            (ewoc-enter-last cookie elem))))
      (kill-buffer (current-buffer)))
    (with-current-buffer (ewoc-buffer cookie)
      (setq tla--nb-active-processes (- tla--nb-active-processes 1))
      (when (and callback
                 (zerop tla--nb-active-processes))
        (funcall callback))))
  (ewoc-refresh cookie))

(defun tla-bookmarks-missing-parse-changes (buffer parent-node)
  (with-current-buffer buffer
    (let ((changes
           (progn (goto-char (point-min))
                  (when (re-search-forward "^[^\\*]" nil t)
                    (buffer-substring-no-properties
                     (line-beginning-position)
                     (point-max)))))
          (local-tree default-directory))
      (when changes
        (with-current-buffer (tla--get-buffer-create 'missing)
          (ewoc-enter-after tla-revision-list-cookie
                            parent-node
                            (list 'entry-change
                                  changes
                                  local-tree)))))))

(defun tla-bookmarks-open-tree ()
  "Open a local tree in a dired buffer."
  (interactive)
  (dired-other-window (tla-bookmarks-read-local-tree)))

(defun tla-bookmarks-inventory ()
  "Run tla-inventory on a local tree."
  (interactive)
  (let ((default-directory (tla-bookmarks-read-local-tree)))
    (tla-inventory nil t)))

(defun tla-bookmarks-changes ()
  "Runs tla-changes on a local tree."
  (interactive)
  (let ((default-directory (tla-bookmarks-read-local-tree)))
    (tla-changes nil nil)))

(defmacro tla-make-move-fn (ewoc-direction function cookie
                                           &optional only-unmerged)
  `(defun ,function ()
     (interactive)
     (let* ((elem (ewoc-locate ,cookie))
            (next (or (,ewoc-direction ,cookie elem) elem)))
       (while (and next
                   (if ,only-unmerged
                       (not (and (eq (car (ewoc-data next))
                                     'entry-patch)
                                 (eq (tla--revision-merged-by
                                      (caddr (ewoc-data next)))
                                     'nobody)))
                     (eq (car (ewoc-data next)) 'separator))
                   (,ewoc-direction ,cookie next))
         (setq next (,ewoc-direction ,cookie next)))
       (while (and next
                   (if ,only-unmerged
                       (not (and (eq (car (ewoc-data next))
                                     'entry-patch)
                                 (eq (tla--revision-merged-by
                                      (caddr (ewoc-data next)))
                                     'nobody)))
                     (eq (car (ewoc-data next)) 'separator)))
         (setq next (,(if (eq ewoc-direction 'ewoc-next)
                          'ewoc-prev
                        'ewoc-next) ,cookie next)))
       (when next (goto-char (ewoc-location next)))))
  )

(tla-make-move-fn ewoc-next tla-revision-next
                  tla-revision-list-cookie)

(tla-make-move-fn ewoc-prev tla-revision-prev
                  tla-revision-list-cookie)

(tla-make-move-fn ewoc-next tla-revision-next-unmerged
                  tla-revision-list-cookie t)

(tla-make-move-fn ewoc-prev tla-revision-prev-unmerged
                  tla-revision-list-cookie t)

;;;###autoload
(defun tla-bookmarks (&optional arg)
  "Displays xtla bookmarks in a buffer. Non-nil prefix argument to
reload the file from disk."
  (interactive "P")
  (tla-bookmarks-load-from-file arg)
  (pop-to-buffer "*tla-bookmarks*")
  (let ((pos (point)))
    (toggle-read-only -1)
    (erase-buffer)
    (set (make-local-variable 'tla-bookmarks-cookie)
         (ewoc-create 'tla-bookmarks-printer))
    (set (make-local-variable 'tla-bookmarks-marked-list) nil)
    (dolist (elem tla-bookmarks-alist)
      (ewoc-enter-last tla-bookmarks-cookie elem))
    (tla-bookmarks-mode)
    (if (equal pos (point-min))
        (if (ewoc-nth tla-bookmarks-cookie 0)
            (tla-bookmarks-cursor-goto (ewoc-nth tla-bookmarks-cookie 0))
          (message "You have no bookmarks, create some in the other buffers"))
      (goto-char pos))))


(defun tla-bookmarks-mode ()
  "Major mode to show xtla bookmarks.

You can add a bookmark with '\\<tla-bookmarks-mode-map>\\[tla-bookmarks-add]', and remove one with '\\[tla-bookmarks-delete]'. After
marking a set of files with '\\[tla-bookmarks-mark]', make them partners with '\\[tla-bookmarks-marked-are-partners]', and
you will then be able to use '\\[tla-bookmarks-missing]' to view the missing patches.

Commands:
\\{tla-bookmarks-mode-map}
"
  (interactive)
  (use-local-map tla-bookmarks-mode-map)
  (setq major-mode 'tla-bookmarks-mode)
  (setq mode-name "tla-bookmarks")
  (toggle-read-only 1)
  (run-hooks 'tla-bookmarks-mode-hook))

(defun tla-bookmarks-cursor-goto (ewoc-bookmark)
  "Move cursor to the ewoc location of EWOC-BOOKMARK"
  (interactive)
  (goto-char (ewoc-location ewoc-bookmark))
  (search-forward ":"))

(defun tla-bookmarks-next ()
  (interactive)
  (let* ((cookie tla-bookmarks-cookie)
         (elem (ewoc-locate cookie))
         (next (or (ewoc-next cookie elem) elem)))
    (tla-bookmarks-cursor-goto next)))

(defun tla-bookmarks-previous ()
  (interactive)
  (let* ((cookie tla-bookmarks-cookie)
         (elem (ewoc-locate cookie))
         (previous (or (ewoc-prev cookie elem) elem)))
    (tla-bookmarks-cursor-goto previous)))

(defun tla-bookmarks-move-down ()
  (interactive)
  (let* ((cookie tla-bookmarks-cookie)
         (elem (ewoc-locate cookie))
         (data (ewoc-data elem))
         (oldname (car data))
         (next (ewoc-next cookie elem)))
    (unless next
      (error "Can't go lower"))
    (tla--ewoc-delete cookie elem)
    (goto-char (ewoc-location
                (ewoc-enter-after cookie next data)))
    (let ((list tla-bookmarks-alist)
          newlist)
      (while list
        (if (string= (caar list) oldname)
            (progn
              (setq newlist (cons (car (cdr list)) newlist))
              (setq newlist (cons (car      list)  newlist))
              (setq list (cdr list)))
          (setq newlist (cons (car list) newlist)))
        (setq list (cdr list)))
      (setq tla-bookmarks-alist (reverse newlist)))
    (search-forward ":")))

(defun tla-bookmarks-move-up ()
  (interactive)
  (let* ((cookie tla-bookmarks-cookie)
         (elem (ewoc-locate cookie))
         (data (ewoc-data elem))
         (oldname (car data))
         (previous (ewoc-prev cookie elem)))
    (unless previous
      (error "Can't go upper"))
    (tla--ewoc-delete cookie elem)
    (goto-char (ewoc-location
                (ewoc-enter-before cookie previous data)))
    (let ((list tla-bookmarks-alist)
          newlist)
      (while list
        (if (string= (caar (cdr list)) oldname)
            (progn
              (setq newlist (cons (car (cdr list)) newlist))
              (setq newlist (cons (car      list)  newlist))
              (setq list (cdr list)))
          (setq newlist (cons (car list) newlist)))
        (setq list (cdr list)))
      (setq tla-bookmarks-alist (reverse newlist)))
    (search-forward ":")))

(defun tla-bookmarks-get (directory)
  (interactive (list (expand-file-name (tla--read-directory-name "Get in directory: "))))
  (let* ((elem (ewoc-data (ewoc-locate tla-bookmarks-cookie)))
         (location (cdr (assoc 'location elem))))
    (tla-get directory t
             (tla-archive-name  location)
             (tla-category-name location)
             (tla-branch-name  location)
             (tla-version-name location))))

(defun tla-bookmarks-goto ()
  (interactive)
  (let* ((elem (ewoc-data (ewoc-locate tla-bookmarks-cookie)))
         (location (cdr (assoc 'location elem)))
         (archive  (tla-archive-name  location))
         (category (tla-category-name location))
         (branch   (tla-branch-name  location))
         (version  (tla-version-name location)))
    (cond (version  (tla-revisions archive category branch version))
          (branch   (tla-versions  archive category branch))
          (category (tla-branches  archive category))
          (archive  (tla-categories archive))
          (t (error "Nothing specified for this bookmark")))))

(defun tla-bookmarks-goto-by-mouse (event)
  (interactive "e")
  (mouse-set-point event)
  (tla-bookmarks-goto))

(defun tla-bookmarks-star-merge (arg)
  (interactive "P")
  (let* ((elem (ewoc-data (ewoc-locate tla-bookmarks-cookie)))
         (location (cdr (assoc 'location elem)))
         (local-tree (tla--read-directory-name "Star-merge into: ")))
    (tla-star-merge (tla--name-construct location)
                    local-tree)))

(defun tla-bookmarks-replay (arg)
  (interactive "P")
  (let* ((elem (ewoc-data (ewoc-locate tla-bookmarks-cookie)))
         (location (cdr (assoc 'location elem)))
         (local-tree (tla--read-directory-name "Replay into: ")))
    (tla-replay (tla--name-construct location)
                local-tree)))

(defun tla-bookmarks-add-elem (name info)
  "Internal function. Adds the association (name . info) to the list
of bookmarks, and saves it"
  (when (assoc name tla-bookmarks-alist)
    (error (concat "Already got a bookmark " name)))
  (let ((elem (cons name info)))
    (tla--add-to-list 'tla-bookmarks-alist elem t)
    (tla-bookmarks-save-to-file)
    (ewoc-enter-last tla-bookmarks-cookie elem)
    ))

(defun tla-bookmarks-add (name revision-spec)
  "Adds a bookmark named NAME for location ARCHIVE/CATEGORY--BRANCH--VERSION"
  (interactive (let* ((fq (tla--read-archive-category-branch-version-name))
                      (n  (read-string (format "Name of the bookmark for `%s': "
                                               (tla--name-construct fq)))))
                 (list n fq)))
  (unless (get-buffer "*tla-bookmarks*")
    (tla-bookmarks))
  (with-current-buffer "*tla-bookmarks*"
    (let* ((info (list (cons 'location
                             revision-spec)
                       (cons 'timestamp (current-time-string)))))
      (tla-bookmarks-add-elem name info))))

(defun tla-bookmarks-mark ()
  "Marks bookmark at point"
  (interactive)
  (let ((pos (point)))
    (add-to-list 'tla-bookmarks-marked-list
                 (ewoc-data (ewoc-locate tla-bookmarks-cookie)))
    (ewoc-refresh tla-bookmarks-cookie)
    (goto-char pos))
  (tla-bookmarks-next))

(defun tla-bookmarks-unmark ()
  "Marks bookmark at point"
  (interactive)
  (let ((pos (point)))
    (setq tla-bookmarks-marked-list
          (delq (ewoc-data (ewoc-locate tla-bookmarks-cookie))
                tla-bookmarks-marked-list))
    (ewoc-refresh tla-bookmarks-cookie)
    (goto-char pos))
  (tla-bookmarks-next))

(defun tla-bookmarks-unmark-all ()
  "Unmarks all bookmarks in current buffer"
  (interactive)
  (let ((pos (point)))
    (setq tla-bookmarks-marked-list nil)
    (ewoc-refresh tla-bookmarks-cookie)
    (goto-char pos)))

(defun tla-bookmarks-marked-are-partners ()
  "All marked bookmarks becomes mutual partners"
  (interactive)
  (let ((list-arch (mapcar
                    #'(lambda (x)
                        (format "%s"
                                (tla--name-construct
                                 (cdr (assoc 'location x)))))
                    tla-bookmarks-marked-list)))
    (dolist (book tla-bookmarks-marked-list)
      (let ((myloc (tla--name-construct
                    (cdr (assoc 'location book)))))
        (message myloc)
        (dolist (arch list-arch)
          (unless (string= myloc arch)
            (tla-bookmarks-add-partner book arch t))))))
  (tla-bookmarks-save-to-file)
  (save-window-excursion
    (tla-bookmarks)))

(defun tla-bookmarks-cleanup-local-trees ()
  "Remove fields local-tree from bookmarks when they don't exist"
  (interactive)
  (dolist (book tla-bookmarks-alist)
    (let ()
      (dolist (local-tree (cdr (assoc 'local-tree book)))
        (when (and (not (file-exists-p local-tree))
                   (or tla-bookmarks-cleanup-dont-prompt
                       (y-or-n-p
                        (format
                         "Remove tree %s from bookmarks %s? "
                         local-tree
                         (car book)))))
          (tla-bookmarks-delete-tree book local-tree t)))))
  (tla-bookmarks-save-to-file)
  (save-window-excursion
    (tla-bookmarks)))

(defun tla-bookmarks-delete (elem &optional force)
  "Deletes bookmark at point"
  (interactive (list (ewoc-locate tla-bookmarks-cookie)))
  (let* ((data (ewoc-data elem)))
    (when (or force
              (yes-or-no-p (format "Delete bookmark \"%s\"? " (car data))))
      (tla--ewoc-delete tla-bookmarks-cookie elem)
      (let ((list tla-bookmarks-alist)
            newlist)
        (while list
          (unless (string= (caar list) (car data))
            (setq newlist (cons (car list) newlist)))
          (setq list (cdr list)))
        (setq tla-bookmarks-alist (reverse newlist)))
      ;; TODO could be optimized
      (tla-bookmarks-save-to-file)
      )))

(defun tla-bookmarks-find-bookmark (location)
  "Finds the bookmark whose location is LOCATION"
  (let ((list tla-bookmarks-alist)
        result)
    (while list
      (when (string= (tla--name-construct
                      (cdr (assoc 'location (cdar list))))
                     location)
        (setq result (car list))
        (setq list nil))
      (setq list (cdr list)))
    result))

(defmacro tla-bookmarks-make-add-fn (name field message-already message-add)
  `(defun ,name (bookmark value &optional dont-save)
     "Adds the directory VALUE to the list of local trees of bookmark
BOOKMARK."
     (let ((local-trees (assoc ,field (cdr bookmark))))
       (if local-trees
           (if (member value (cdr local-trees))
               (message ,message-already)
             (progn
               (message ,message-add)
               (setcdr local-trees (cons value
                                         (cdr local-trees)))))
         (progn
           (message ,message-add)
           (setcdr bookmark (cons (list ,field value)
                                  (cdr bookmark)))))
       (unless dont-save
         (tla-bookmarks-save-to-file)
         (save-window-excursion
           (tla-bookmarks)))))
  )

(tla-bookmarks-make-add-fn tla-bookmarks-add-tree
                           'local-tree
                           "Local tree already in the list"
                           "Local tree added to your bookmarks")

(tla-bookmarks-make-add-fn tla-bookmarks-add-partner
                           'partners
                           "Partner already in the list"
                           "Partner added to your bookmarks")

(tla-bookmarks-make-add-fn tla-bookmarks-add-group
                           'groups
                           "Group already in the list"
                           "Group added to your bookmarks")

(tla-bookmarks-make-add-fn tla-bookmarks-add-nickname
                           'nickname
                           "Nickname already in the list"
                           "Nickname added to your bookmark")

(defmacro tla-bookmarks-make-delete-fn (name field)
  `(defun ,name (bookmark value &optional dont-save)
     "Deletes the directory VALUE to the list of local trees of bookmark
BOOKMARK."
     (let ((local-trees (assoc ,field (cdr bookmark))))
       (when local-trees
         (let ((rem-list (delete value (cdr (assoc ,field
                                                   bookmark)))))
           (if rem-list
               (setcdr local-trees rem-list)
             ;; Remove the whole ('field ...)
             (setcdr bookmark (delq local-trees (cdr bookmark))))))
       (unless dont-save
         (tla-bookmarks-save-to-file)
         (save-window-excursion
           (tla-bookmarks)))))
  )

(tla-bookmarks-make-delete-fn tla-bookmarks-delete-tree
                              'local-tree)

(tla-bookmarks-make-delete-fn tla-bookmarks-delete-partner
                              'partners)

(tla-bookmarks-make-delete-fn tla-bookmarks-delete-group
                              'groups)

(tla-bookmarks-make-delete-fn tla-bookmarks-delete-nickname
                              'nickname)

(defun tla-bookmarks-add-partner-interactive ()
  "Adds a partner to the current or marked bookmarks"
  (interactive)
  (let ((bookmarks (or tla-bookmarks-marked-list
                       (list (ewoc-data (ewoc-locate
                                         tla-bookmarks-cookie)))))
        (partner (tla--name-construct
                  (tla--read-archive-category-branch-version-name))))
    (dolist (bookmark bookmarks)
      (tla-bookmarks-add-partner bookmark partner t))
    (tla-bookmarks-save-to-file)
    (save-window-excursion
      (tla-bookmarks))))

(defun tla-bookmarks-add-partners-from-file ()
  "Adds a partner to the current or marked bookmarks"
  (interactive)
  (let ((bookmarks (or tla-bookmarks-marked-list
                       (list (ewoc-data (ewoc-locate
                                         tla-bookmarks-cookie))))))
    (dolist (bookmark bookmarks)
      (let ((partners (tla-partner-list
                       (tla-bookmarks-read-local-tree bookmark))))
        (dolist (partner partners)
          (tla-bookmarks-add-partner bookmark partner t))))
    (tla-bookmarks-save-to-file)
    (save-window-excursion
      (tla-bookmarks))))

(defun tla-bookmarks-write-partners-to-file ()
  "Adds the partners recorded in the bookmarks to the partner file."
  (interactive)
  (let ((bookmarks (or tla-bookmarks-marked-list
                       (list (ewoc-data (ewoc-locate
                                         tla-bookmarks-cookie))))))
    (dolist (bookmark bookmarks)
      (let* ((local-tree (tla-bookmarks-read-local-tree bookmark))
             (partners (tla-partner-list local-tree)))
        (with-current-buffer
            (tla-partner-find-partner-file local-tree)
          (dolist (partner (cdr (assoc 'partners (cdr bookmark))))
            (unless (member partner partners)
              (insert partner "\n")))
          (and (buffer-modified-p)
               (progn (switch-to-buffer (current-buffer))
                      (y-or-n-p (format "Save file %s? "
                                        (buffer-file-name))))
               (save-buffer)))))))


(defun tla-bookmarks-delete-partner-interactive ()
  "Deletes a partner from the current or marked bookmarks"
  (interactive)
  (let* ((bookmarks (or tla-bookmarks-marked-list
                        (list (ewoc-data (ewoc-locate
                                          tla-bookmarks-cookie)))))
         (choices (apply 'append
                         (mapcar #'(lambda (x)
                                     (cdr (assoc 'partners
                                                 (cdr x))))
                                 bookmarks)))
         (choices-alist (mapcar #'(lambda (x) (list x)) choices))
         (partner (completing-read "Partner to remove: " choices-alist)))
    (dolist (bookmark bookmarks)
      (tla-bookmarks-delete-partner bookmark partner t))
    (tla-bookmarks-save-to-file)
    (save-window-excursion
      (tla-bookmarks))))

(defun tla-bookmarks-add-tree-interactive ()
  "Adds a local tree to the current or marked bookmarks"
  (interactive)
  (let ((bookmarks (or tla-bookmarks-marked-list
                       (list (ewoc-data (ewoc-locate
                                         tla-bookmarks-cookie)))))
        (local-tree (tla--read-directory-name "Local tree to add: ")))
    (unless (file-exists-p (concat local-tree "/{arch}"))
      (error (concat local-tree " is not an arch local tree.")))
    (dolist (bookmark bookmarks)
      (tla-bookmarks-add-tree bookmark local-tree t))
    (tla-bookmarks-save-to-file)
    (save-window-excursion
      (tla-bookmarks))))

(defun tla-bookmarks-delete-tree-interactive ()
  "Adds a local tree to the current or marked bookmarks"
  (interactive)
  (let* ((bookmarks (or tla-bookmarks-marked-list
                        (list (ewoc-data (ewoc-locate
                                          tla-bookmarks-cookie)))))
         (choices (apply 'append
                         (mapcar #'(lambda (x)
                                     (cdr (assoc 'local-tree
                                                 (cdr x))))
                                 bookmarks)))
         (choices-alist (mapcar #'(lambda (x) (list x)) choices))
         (local-tree (completing-read "Local tree to remove: " choices-alist)))
    (unless (file-exists-p (concat local-tree "/{arch}"))
      (error (concat local-tree " is not an arch local tree.")))
    (dolist (bookmark bookmarks)
      (tla-bookmarks-delete-tree bookmark local-tree t))
    (tla-bookmarks-save-to-file)
    (save-window-excursion
      (tla-bookmarks))))

(defun tla-bookmarks-list-groups ()
  "Returns the list of groups currently used by bookmarks"
  (let ((list (apply 'append
                     (mapcar #'(lambda (x)
                                 (cdr (assoc 'groups
                                             (cdr x))))
                             tla-bookmarks-alist)))
        result)
    ;; Make elements unique
    (dolist (elem list)
      (add-to-list 'result elem))
    result))

(defun tla-bookmarks-add-group-interactive ()
  "Add a group entry in the current or marked bookmarks"
  (interactive)
  (let* ((bookmarks (or tla-bookmarks-marked-list
                        (list (ewoc-data (ewoc-locate
                                          tla-bookmarks-cookie)))))
         (group (completing-read "Group of bookmarks: "
                                 (mapcar #'(lambda (x) (list x))
                                         (tla-bookmarks-list-groups)))))
    (dolist (bookmark bookmarks)
      (tla-bookmarks-add-group bookmark group t)))
  (tla-bookmarks-save-to-file)
  (save-window-excursion
    (tla-bookmarks)))


(defun tla-bookmarks-delete-group-interactive ()
  "Deletes a group of bookmark entry from the current or marked bookmarks"
  (interactive)
  (let* ((bookmarks (or tla-bookmarks-marked-list
                        (list (ewoc-data (ewoc-locate
                                          tla-bookmarks-cookie)))))
         (choices (apply 'append
                         (mapcar #'(lambda (x)
                                     (cdr (assoc 'groups
                                                 (cdr x))))
                                 bookmarks)))
         (choices-alist (mapcar #'(lambda (x) (list x)) choices))
         (group (completing-read "Group to remove: " choices-alist)))
    (dolist (bookmark bookmarks)
      (tla-bookmarks-delete-group bookmark group t)))
  (tla-bookmarks-save-to-file)
  (save-window-excursion
    (tla-bookmarks)))

(defun tla-bookmarks-select-by-group (group)
  "Select all bookmarks in the group GROUP"
  (interactive (list (completing-read "Group to select: "
                                      (mapcar #'(lambda (x) (list x))
                                              (tla-bookmarks-list-groups)))))
  (dolist (bookmark tla-bookmarks-alist)
    (when (member group (cdr (assoc 'groups bookmark)))
      (add-to-list 'tla-bookmarks-marked-list bookmark))
    )
  (ewoc-refresh tla-bookmarks-cookie))

(defun tla-bookmarks-add-nickname-interactive ()
  "Adds a nickname to the current bookmark"
  (interactive)
  (let* ((bookmark (ewoc-data (ewoc-locate
                               tla-bookmarks-cookie)))
         (prompt (format "Nickname for %s: " (tla--name-construct
                                              (cdr (assoc 'location bookmark))))))
    (tla-bookmarks-add-nickname bookmark (read-string prompt) t)
    (tla-bookmarks-save-to-file)
    (save-window-excursion
      (tla-bookmarks))))

(defun tla-bookmarks-delete-nickname-interactive ()
  "Deletes the nickname of the current bookmark"
  (interactive)
  (let* ((bookmark (ewoc-data (ewoc-locate
                               tla-bookmarks-cookie)))
         (nickname (cadr (assoc 'nickname bookmark))))
    (tla-bookmarks-delete-nickname bookmark nickname t)
    (tla-bookmarks-save-to-file)
    (save-window-excursion
      (tla-bookmarks))))

(defvar tla-buffer-bookmark nil
  "Bookmark manipulated in current buffer")

(defun tla-bookmarks-edit ()
  "Edit bookmark at point"
  (interactive)
  (let* ((elem (ewoc-locate tla-bookmarks-cookie))
         (data (ewoc-data elem)))
    (pop-to-buffer (concat "*xtla bookmark " (car data) "*"))
    (erase-buffer)
    (emacs-lisp-mode)
    (make-local-variable 'tla-buffer-bookmark)
    (setq tla-buffer-bookmark elem)
    (insert ";; Edit the current bookmark. C-c C-c to finish\n\n")
    (pp data (current-buffer))
    (goto-char (point-min)) (forward-line 2) (forward-char 2)
    (local-set-key [(control ?c) (control ?c)]
                   #'(lambda () (interactive)
                       (goto-char (point-min))
                       (let* ((newval (read (current-buffer)))
                              (elem tla-buffer-bookmark)
                              (oldname (car (ewoc-data elem))))
                         (kill-buffer (current-buffer))
                         (pop-to-buffer "*tla-bookmarks*")
                         (setcar (ewoc-data elem) (car newval))
                         (setcdr (ewoc-data elem) (cdr newval))
                         (let ((list tla-bookmarks-alist)
                               newlist)
                           (while list
                             (if (string= (caar list) oldname)
                                 (setq newlist (cons newval newlist))
                               (setq newlist (cons (car list) newlist)))
                             (setq list (cdr list)))
                           (setq tla-bookmarks-alist (reverse newlist)))
                         (tla-bookmarks-save-to-file)
                         (save-excursion (tla-bookmarks))
                         )))))

;;
;; Archives
;;
;;;###autoload
(defun tla-archives ()
  (interactive)
  (tla--archive-tree-build-archives)
  (tla-switch-to-buffer "*tla-archives*")
  (let ((a-list tla--archive-tree)
        (my-default-archive (tla-my-default-archive))
        defaultp
        archive-name
        archive-location
        p)
    (toggle-read-only -1)
    (erase-buffer)
    (while a-list
      (setq archive-name (caar a-list)
            archive-location (cadar a-list)
            a-list (cdr a-list)
            defaultp (string= archive-name my-default-archive))
      (if defaultp (setq p (point)))
      (tla-archives-insert-item archive-name archive-location defaultp))
    (delete-backward-char 1)
    (when p (goto-char p))
    (tla-archive-list-mode)))

(defun tla-archives-insert-item (archive location defaultp)
  (let ((start-pos (point))
        overlay)
    (insert (if defaultp "*" " ")
            "  "
            (tla--face-add-with-condition
             defaultp
             archive 'tla-marked 'tla-archive-name))
    (newline)
    (insert "      " location)
    (newline)
    (setq overlay (make-overlay start-pos (point)))
    (overlay-put overlay 'category 'tla-default-button)
    (overlay-put overlay 'keymap tla-archive-archive-map)
    (overlay-put overlay 'tla-archive-info archive)))

(defun tla-archives-goto-archive-by-name (name)
  (unless (string= (buffer-name) "*tla-archives*")
    (error "tla-archives-goto-archive-by-name can only be called in *tla-archives* buffer"))
  (goto-char (point-min))
  (search-forward name)
  (beginning-of-line))

(defun tla-get-archive-info (&optional property)
  "Returns the archive at point in an archive list buffer"
  (unless property
    (setq property 'tla-archive-info))
  (let ((overlay (car (overlays-at (point)))))
    (when overlay
      (overlay-get overlay property))))

(defun tla-my-default-archive (&optional new-default)
  "Set or get the default archive.
When called with a prefix argument: Ask the user for the new default archive.
When called with a string argument: Set the default archive to this string.
When called with no argument: return the name of the default argument.
When called interactively, with no argument: Show the name of the default archive."
  (interactive "P")
  (when (or (numberp new-default) (and (listp new-default) (> (length new-default) 0)))
    (setq new-default (car (tla--read-archive-name))))
  (cond ((stringp new-default)
         (message "Setting arch default archive to: %s" new-default)
         (tla--run-tla-sync (list "my-default-archive" new-default)
                            :finished 'tla--null-handler))
        (t
         (tla--run-tla-sync '("my-default-archive")
                            :finished
                            `(lambda (output error status arguments)
                               (let ((result (tla--buffer-content output)))
                                 (when ,(interactive-p)
                                   (message "Default arch archive: %s"
                                            result))
                                 result))))))

(defun tla-whereis-archive (&optional archive)
  "Call tla whereis-archive.
"
  (interactive "P")
  (let ((location))
    (unless archive
      (setq archive (tla--get-archive (tla--get-info-at-point))))
    (setq location
          (tla--run-tla-sync (list "whereis-archive" archive)
                             :finished
                             (lambda (output error status arguments)
                               (tla--buffer-content output))))
    (when (interactive-p)
      (message "archive location for %s: %s" archive location))
    location))

(defun tla-register-archive (location &optional archive)
  "Register arch archive.
LOCATION should be either a local directory or a remote path.
When ffap is available the url at point is suggested for LOCATION.
ARCHIVE is the name is archive. If ARCHIVE is not given or an empty string,
the default name is used."
  (interactive (list (read-string "Location: " (ffap-url-at-point))
                     (read-string "Archive (empty for default): ")))
  (if (and archive (eq 0 (length archive)))
      (setq archive nil))
  (let ((archive-registered nil)
        (tla-response nil))
    (tla--run-tla-sync (list "register-archive" archive location)
                       :finished
                       (lambda (output error status arguments)
                         (setq tla-response (tla--get-process-output))
                         (setq archive-registered t))
                       :error
                       (lambda (output error status arguments)
                         (setq tla-response (tla--get-error-output))
                         (when (= status 2) ;; already registered
                           (setq archive-registered t))))
    (when archive-registered
      (tla-archives)
      (tla-archives-goto-archive-by-name
       (progn
         (message tla-response) ; inform the user about the response from tla
         (if (string-match ".+: \\(.+\\)" tla-response)
             (match-string-no-properties 1 tla-response)
           archive)))
      (tla-flash-line))))

(defun tla-make-archive (name location &optional signed listing)
  "Create a new arch archive.
NAME is the global name for the archive.  It must be an
email address with a fully qualified domain name, optionally
followed by \"--\" and a string of letters, digits, periods
and dashes.
LOCATION specifies the path, where the archive should be created.

Examples for name are:
foo.bar@flups.com--public
foo.bar@flups.com--public-2004

If SIGNED is non-nil, the archive will be created with --signed.
If LISTING is non-nil, the archive will be created with --listing
 (Usefull for http mirrors)."
  (interactive
   (list (read-string "Archive name: ")
         (read-string "Location: ")
         (y-or-n-p "Sign the archive? ")
         (y-or-n-p "Create .listing files? ")))
  (setq location (expand-file-name location))
  (if (file-directory-p location)
      (error "directory is existing: %s" location))
  (tla--run-tla-sync (list "make-archive"
                           (when listing "--listing")
                           (when signed "--signed")
                           name location)
                     :error (lambda (output error status arguments)
                              (tla--show-error-buffer error)
                              (tla--show-last-process-buffer)
                              (error (format "tla failed: exits-status=%s" status))))
  (tla-archives))

(defun tla-mirror-archive (&optional archive location mirror signed
                                     listing)
  "Creates a mirror for the archive ARCHIVE, at location LOCATION,
using the registration name MIRROR for the mirror. The archive will be
signed if MIRROR is non-nil, and .listing files (usefull for HTTP
mirrors) will be created if LISTING is provided."
  (interactive)
  (let* ((archive (or archive (car (tla--read-archive-name "Archive to mirror: "))))
         (location (or location (read-string "Location of the mirror: ")))
         (mirror (or mirror (read-string "Name of the mirror: "
                                         (concat archive
                                                 "-MIRROR"))))
         (signed (or signed (y-or-n-p "Sign mirror? ")))
         (listing (or listing (y-or-n-p "Create .listing files? "))))
    (tla--run-tla-sync (list "make-archive"
                             (when listing "--listing")
                             (when signed "--signed")
                             "--mirror"
                             archive mirror location))
    (tla-archives)))

;;
;; Categories
;;
(defun tla-categories (archive)
  (interactive (tla--read-archive-name))
  (unless archive
    (setq archive (tla-my-default-archive)))
  (tla--archive-tree-build-categories archive)
  (tla-switch-to-buffer "*tla-categories*")
  (let ((list (cddr (tla--archive-tree-get-archive archive)))
        category start-pos overlay)
    (toggle-read-only -1)
    (erase-buffer)
    ;; TODO: button to invoke tla-archives.
    (insert (format "Archive: %s\n%s\n" archive
                    (make-string (+ (length archive)
                                    (length "Archive: ")) ?=)))
    (save-excursion
      (while list
        (setq category (car (car list))
              start-pos (point)
              list (cdr list))
        (insert "   " (tla--face-add category 'tla-category-name))
        (newline)
        (setq overlay (make-overlay start-pos (point)))
        (overlay-put overlay 'category 'tla-default-button)
        (overlay-put overlay 'keymap   tla-category-category-map)
        (overlay-put overlay 'tla-category-info category)
        )
      (delete-backward-char 1)))
  (tla-category-list-mode)
  (set (make-local-variable 'tla-buffer-archive-name)
       archive))

(defun tla-make-category (archive category)
  "Make new category."
  (interactive (tla--read-archive-category-name))
  (tla--run-tla-sync (list "make-category" "-A" archive category))
  (let ((tla-buffer-archive-name archive))
    (run-hooks 'tla-make-category-hook)))

;;
;; Branches
;;
(defun tla-branches (archive category)
  (interactive (tla--read-archive-category-name))
  (tla--archive-tree-build-branches archive category)
  (tla-switch-to-buffer "*tla-branches*")
  (let ((list (cdr (tla--archive-tree-get-category archive category)))
        alength
        clength
        branch
        start-pos
        overlay)
    (toggle-read-only -1)
    (erase-buffer)
    ;; TODO: button to invoke tla-categories and tla-archives
    (setq alength (+ (length archive)  (length "Archive: "))
          clength (+ (length category) (length "Category: ")))
    (insert (format "Archive: %s\nCategory: %s\n%s\n" archive category
                    (make-string (max alength clength) ?=)))
    (save-excursion
      (while list
        (setq branch (car (car list))
              start-pos (point)
              list (cdr list))
        (insert "   " (tla--face-add branch 'tla-branch-name))
        (newline)
        (setq overlay (make-overlay start-pos (point)))
        (overlay-put overlay 'category 'tla-default-button)
        (overlay-put overlay 'keymap    tla-branch-branch-map)
        (overlay-put overlay 'tla-branch-info branch))
      (delete-backward-char 1)))
  (tla-branch-list-mode)
  (set (make-local-variable 'tla-buffer-archive-name)
       archive)
  (set (make-local-variable 'tla-buffer-category-name)
       category))

(defun tla-make-branch (archive category branch)
  "Make new branch."
  (interactive (tla--read-archive-category-branch-name))
  (tla--run-tla-sync (list "make-branch"
                           (tla--name-construct
                            archive category branch)))
  (let ((tla-buffer-archive-name archive)
        (tla-buffer-category-name category))
    (run-hooks 'tla-make-branch-hook)))

;;
;; Versions
;;
(defun tla-versions (archive category branch)
  (interactive (tla--read-archive-category-branch-name))
  (tla--archive-tree-build-versions archive category branch)
  (tla-switch-to-buffer "*tla-versions*")
  (let ((list (cdr (tla--archive-tree-get-branch
                    archive category branch)))
        alength
        clength
        blength
        version
        start-pos
        overlay)
    (toggle-read-only -1)
    (erase-buffer)
    ;; TODO: button to invoke tla-categories and tla-archives
    (setq alength (+ (length archive)  (length "Archive: "))
          clength (+ (length category) (length "Category: "))
          blength (+ (length branch)   (length "Branch: ")))
    (insert (format "Archive: %s\nCategory: %s\nBranch: %s\n%s\n"
                    archive category branch
                    (make-string (max alength clength blength) ?=)))
    (save-excursion
      (while list
        (setq version (car (car list))
              start-pos (point)
              list (cdr list))
        (insert "   " (tla--face-add version 'tla-version-name))
        (newline)
        (setq overlay (make-overlay start-pos (point)))
        (overlay-put overlay 'category 'tla-default-button)
        (overlay-put overlay 'keymap   tla-version-version-map)
        (overlay-put overlay 'tla-version-info version))
      (delete-backward-char 1)))
  (tla-version-list-mode)
  (set (make-local-variable 'tla-buffer-archive-name) archive)
  (set (make-local-variable 'tla-buffer-category-name) category)
  (set (make-local-variable 'tla-buffer-branch-name) branch))

(defun tla-make-version (archive category branch version)
  "Make new version."
  (interactive (tla--read-archive-category-branch-version-name))
  (tla--run-tla-sync (list "make-version"
                           (tla--name-construct
                            archive category branch version)))
  (let ((tla-buffer-archive-name archive)
        (tla-buffer-category-name category)
        (tla-buffer-branch-name branch))
    (run-hooks 'tla-make-version-hook)))

;;
;; Revisions
;;

;; elem should be
;; ('separator "string" kind)
;; or
;; ('entry-patch nil revision) Where "revision" is of tla--revision
;; struct type.
;; ('entry-change "changes")
;; The second element tells if the element is marked or not.
(defun tla-revision-list-printer (elem)
  (let ()
    (case (car elem)
      (entry-patch
       (let* ((struct (caddr elem))
              (merged-by (tla--revision-merged-by struct))
              (unmerged (eq merged-by 'nobody)))
         (insert (if (cadr elem) " *" "  ")
                 ;; The revision is in library?
                 (if (and tla-revisions-shows-library
                          (apply 'tla--revlib-tree-get-revision
                                 (tla--revision-revision struct)))
                     ;;
                     ;; (apply 'tla-library-find
                     ;;       (append (caddr elem) '(t))

                     "L " "  ")
                 (tla--face-add (tla--name-construct
                                 (tla--revision-revision struct))
                                (if unmerged 'tla-unmerged
                                  'tla-revision-name)
                                tla-revision-revision-map
                                tla-revision-revision-menu)
                 (if unmerged (tla--face-add "  [NOT MERGED]"
                                             'tla-unmerged)
                   ""))
         (let ((summary (tla--revision-summary struct))
               (creator (tla--revision-creator struct))
               (date (tla--revision-date struct)))
           (when (and summary tla-revisions-shows-summary)
             (insert "\n      " summary))
           (when (and creator tla-revisions-shows-creator)
             (insert "\n      " creator))
           (when (and date tla-revisions-shows-date)
             (insert "\n      " date)))
         (when (and tla-revisions-shows-merges
                    (tla--revision-merges struct)
                    (not (null (car (tla--revision-merges struct)))))
           (insert "\n      Merges:")
           (dolist (elem (tla--revision-merges struct))
             (insert "\n        " elem)))
         (when tla-revisions-shows-merged-by
           (cond ((null merged-by) nil)
                 ((listp merged-by)
                  (insert "\n      Merged-by:")
                  (dolist (elem merged-by)
                    (insert "\n        " elem)))))))
      (entry-change (insert  (cadr elem)))
      (separator
       (case (caddr elem)
         (partner (insert "\n" (tla--face-add (cadr elem)
                                              'tla-separator)))
         (bookmark (insert "\n" (tla--face-add
                                 (concat "*** "
                                         (cadr elem)
                                         " ***")
                                 'tla-separator) "\n")))))))

(defun tla-tree-revisions ()
  "Calls `tla-revisions' in the current tree"
  (interactive)
  (let* ((default-directory (tla--read-project-tree-maybe
                             "Run tla revisions in: "))
         (version (tla-tree-version-list)))
    (unless version
      (error "not in a project tree"))
    (apply 'tla-revisions version)))

;;;###autoload
(defun tla-revisions (archive category branch version
                              &optional update-display from-revlib)
  (interactive (tla--read-archive-category-branch-version-name))
  ;; TODO: Consdider the case where (and update-display from-revlib) is t.
  (unless (and update-display
               (or (tla--revisions-tree-contains-details
                    archive category branch version)
                   (not (or tla-revisions-shows-summary
                            tla-revisions-shows-creator
                            tla-revisions-shows-date))))
    (if from-revlib
        (tla--revlib-tree-build-revisions archive category branch version)
      (tla--archive-tree-build-revisions archive category branch version)))
  (tla-switch-to-buffer "*tla-revisions*")
  (let ((list (cdr (if from-revlib
                       (tla--revlib-tree-get-version
                        archive category branch version)
                     (tla--archive-tree-get-version
                      archive category branch version))))
        first
        separator
        revision
        summary
        creator
        date)
    (tla-revision-list-mode)
    (toggle-read-only -1)
    (set (make-local-variable 'tla-buffer-refresh-function)
         'tla-revision-refresh)
    (setq separator (tla--face-add
                     (make-string
                      (max (+ (length archive)  (length "Archive: "))
                           (+ (length category) (length "Category: "))
                           (+ (length branch)   (length "Branch: "))
                           (+ (length version)  (length "Version: ")))
                      ?\ )
                     'tla-separator))
    (ewoc-set-hf tla-revision-list-cookie
                 (tla--revisions-header archive category branch version
                                        from-revlib separator)
                 (concat "\n" separator))
    (if tla-revisions-shows-library
        (tla--revlib-tree-build-revisions
         archive category branch version nil t))
    (while list
      (setq revision (car (car list))
            summary (car (cdr (car list)))
            creator (car (cddr (car list)))
            date (car (cdddr (car list)))
            list (cdr list))
      (ewoc-enter-last tla-revision-list-cookie
                       (list 'entry-patch nil
                             (make-tla--revision
                              :revision (list archive
                                              category
                                              branch
                                              version
                                              revision)
                              :summary summary
                              :creator creator
                              :date    date)))
      (if first
          (goto-char first)
        (goto-char (point-min))
        (re-search-forward "^$")
        (forward-line 1)
        (setq first (point)))
      (sit-for 0)))

  (set (make-local-variable 'tla-buffer-archive-name) archive)
  (set (make-local-variable 'tla-buffer-category-name) category)
  (set (make-local-variable 'tla-buffer-branch-name) branch)
  (set (make-local-variable 'tla-buffer-version-name) version)
  (toggle-read-only t))

(defun tla--revisions-header (archive category branch version from-revlib separator)
  (concat
   "Version: "
   (tla--face-add archive  'tla-archive-name)  "/"
   (tla--face-add category 'tla-category-name) "--"
   (tla--face-add branch   'tla-branch-name)   "--"
   (tla--face-add version  'tla-version-name)  "\n"
   "In Revision Library: " (tla--face-add (if from-revlib "Yes" "No")  'bold)
   "\n"
   separator "\n"))


;;;###autoload
(defun tla-missing (local-tree location)
  "Search in directory LOCAL-TREE for missing patches from LOCATION.
If the current buffers default directory is in an arch managed tree use that
one unless called with a prefix arg.  In all other cases prompt for the local
tree and the location."
  (interactive (let ((dir
                      (or (if (not current-prefix-arg)
                              (tla-tree-root nil t))
                          (expand-file-name
                           (tla--read-directory-name
                            "Search missing patches in directory: "
                            default-directory default-directory t nil)))))
                 (list dir
                       (let ((default-directory dir))
                         (if current-prefix-arg
                             (tla--read-archive-category-branch-version-name
                              "From location: ")
                           (tla-tree-version))))))
  (pop-to-buffer (tla--get-buffer-create 'missing))
  (tla-revision-list-mode)
  (set (make-local-variable 'tla-buffer-refresh-function)
       'tla-missing-refresh)
  (set (make-local-variable 'tla-missing-buffer-todolist)
       `((missing ,local-tree ,(tla--name-construct location) nil)))
  (tla-missing-refresh))


;;
;; Rbrowse interface
;;
;; TODO: Use tree-widget.
(defun tla-browse-archive (archive)
  (interactive (tla--read-archive-name))
  (unless archive
    (setq archive (tla-my-default-archive)))
  (tla--run-tla-sync (list "rbrowse" "-A" archive)))

;;
;; Get
;;
(defun tla-get (directory run-dired-p archive category branch
                          &optional version revision)
  "Run tla get."
  ;; run-dired-p => t, nil, ask
  (interactive (let* ((l (tla--read-archive-category-branch-version-revision-name))
                      (name (tla--name-construct-semi-qualified (remove nil l)))
                      (d (tla--read-directory-name (format "Store \"%s\" to: " name))))
                 (cons d (cons 'ask l))))
  (setq directory (expand-file-name directory))
  (let* ((name (tla--get-revision
                (tla--name-construct
                 (list archive category branch version revision))))
         (result (tla--run-tla-async
                  (list "get" "-A" archive name directory)
                  :finished `(lambda (output error status arguments)
                               (let ((i (tla--status-handler output error status arguments)))
                                 (when (zerop i)
                                   (tla--get-do-bookmark ,directory ,archive ,category ,branch ,version)
                                   (tla--do-dired ,directory ',run-dired-p)))))))))

(defun tla--get-do-bookmark (directory archive category branch version)
  (let ((bookmark (tla-bookmarks-find-bookmark
                   (tla--name-construct
                    archive category branch version))))
    (when bookmark
      (tla-bookmarks-add-tree bookmark directory))))

(defun tla--do-dired (directory run-dired-p)
  (case run-dired-p
    (ask (when (y-or-n-p (format "Run dired at %s? " directory))
           (dired directory)))
    ('nil nil)
    (t (dired directory))))

;;
;; Cacherev
;;
;; TODO:
;; - run this asynchronous
;; - provide the way to run interactively
;; - show progress
;;
(defun tla-cache-revision (archive category branch version revision)
  (interactive (tla--read-archive-category-branch-version-revision-name))
  (let ((result (tla--run-tla-async (list "cacherev"
                                          (tla--name-construct
                                           archive category branch version revision)))))
    ;;    (tla--show-last-process-buffer)
    result))

;;
;; Add
;;
(defun tla-add (id &rest files)
  (interactive (let ((name
                      (read-file-name "Add file as source: "
                                      nil nil t
                                      (file-name-nondirectory (or
                                                               (buffer-file-name) ""))))
                     (id (read-string "id (empty for default): ")))
                 (list id name)))
  (if (and id (string= id ""))
      (setq id nil))
  (setq files (mapcar 'expand-file-name files))
  (if id
      (tla--run-tla-sync `("add" "--id" ,id . ,files)
                         :finished 'tla--null-handler)
    (tla--run-tla-sync `("add" . ,files)
                       :finished 'tla--null-handler)))

;;
;; Remove
;;
(defun tla-remove (only-id &rest files)
  (interactive (let* ((name
                       (read-file-name "Remove file: "
                                       nil nil t
                                       (file-name-nondirectory (or
                                                                (buffer-file-name) ""))))
                      (only-id (not (y-or-n-p (format
                                               "Delete the \"%s\" locally also? "
                                               name)))))
                 (list only-id name)))
  (setq files (mapcar 'expand-file-name files))
  (dolist (f files)
    (when (equal 0 (tla--run-tla-sync (list "id" "--explicit" f)
                                      :finished 'tla--status-handler
                                      :error 'tla--status-handler))
      (tla--run-tla-sync (list "delete-id" f)
                         :finished 'tla--status-handler))
    (unless only-id
      (delete-file f))))

;;
;; Move
;;
(defun tla-move (from to only-id)
  (interactive
   (list (read-file-name "Move file: "
                         nil nil t
                         (file-name-nondirectory
                          (or (buffer-file-name) "")))
         nil nil))
  (setq to (or to (read-file-name (format "Move file %S to: " from)
                                  nil nil nil (file-name-nondirectory from)))
        only-id (if (eq only-id 'ask)
                    (not (y-or-n-p "Move the file locally also? "))
                  only-id)
        from (expand-file-name from)
        to   (expand-file-name to))
  (let ((buffer (get-file-buffer from))
        (cmd (if only-id "move-id" "mv")))
    (if buffer
        (save-excursion
          (set-buffer buffer)
          (set-visited-file-name to)))
    (tla--run-tla-sync (list cmd from to)
                       :finished 'tla--status-handler)))

;;
;; Update
;;
(defun tla-update (tree)
  (interactive (list (expand-file-name
                      (tla--read-directory-name "Update tree: " nil nil nil ""))))
  (or (tla-save-some-buffers tree)
      (y-or-n-p
       "Update may delete unsaved changes. Continue anyway? ")
      (error "Not updating"))
  (cd tree)
  (tla--run-tla-sync (list "update" (when tla-use-forward-option "--forward")))
  (tla-revert-some-buffers tree))

;;
;; Import
;;
;;;###autoload
(defun tla-start-project ()
  "Start a new project.
Prompts for the root directory of the project and the fully
qualified version name to use.  Sets up and imports the tree and
displays an inventory buffer to allow the project's files to be
added and committed."
  (interactive)
  (let* ((base (tla--read-directory-name "Directory containing files to import: "
                                         (or default-directory
                                             (getenv "HOME"))))
         (l (tla--read-archive-category-branch-version-name))
         (project (if (member nil l)
                      (error "Need a fully qualified version name")
                    (tla--name-construct l))))
    (let ((default-directory (concat base "/")))
      (tla--run-tla-sync (list "init-tree" project))
      (tla--run-tla-sync (list "import" "--setup"))
      (tla-inventory base t))))

(defvar tla-partner-file-precious "/{arch}/+partner-versions"
  "Precious version of the partner file. We strongly suggest keeping
the default value since this is a convention used by other tla
front-ends like Aba.")

(defvar tla-partner-file-source "{arch}/=partner-versions"
  "Precious version of the partner file. We strongly suggest keeping
the default value since this is a convention used by other tla
front-ends like Aba.")

;; ----------------------------------------------------------------------------
;; xtla partner stuff
;; ----------------------------------------------------------------------------
(defun tla-partner-find-partner-file (&optional local-tree)
  "Do find-file tla-partners file and return the buffer. If the file
`tla-partner-file-precious' exists, it is used in priority. Otherwise,
use `tla-partner-file-source'. The precious one is meant for user
configuration, whereas the source one is used for project-wide
configuration."
  (interactive)
  (let ((default-directory (or local-tree default-directory)))
    (let* ((partner-file
            (cond ((file-exists-p (concat (tla-tree-root)
                                          tla-partner-file-precious))
                  (concat (tla-tree-root) tla-partner-file-precious))
                  (t (concat (tla-tree-root)
                             tla-partner-file-source))))
           (buffer-visiting (find-buffer-visiting partner-file)))
      (if buffer-visiting
          (with-current-buffer buffer-visiting
            (if (buffer-modified-p)
                (if (progn (switch-to-buffer (current-buffer))
                           (y-or-n-p (format "Save file %s? "
                                             (buffer-file-name))))
                    (save-buffer)
                  (revert-buffer)))
            buffer-visiting)
      (find-file-noselect partner-file)))))

(defun tla-partner-read (&optional prompt)
  (tla--name-construct
   (tla--read-archive-category-branch-version-name
    (when prompt (concat prompt "[Archive name] "))
    (when prompt (concat prompt "[Category name] "))
    (when prompt (concat prompt "[Branch name] "))
    (when prompt (concat prompt "[Version name] "))
    )))

(defun tla-partner-add (partner)
  "Add a partner for this xtla working copy.
Return nil if PARTNER is alerady in partners file.
For example: Franz.Lustig@foo.bar--public/tla--main--0.1"
  (interactive (list (tla-partner-read)))
  (let ((list (tla-partner-list)))
    (if (member partner list)
        nil
      (with-current-buffer (tla-partner-find-partner-file)
        (goto-char (point-min))
        (insert partner)
        (newline)
        (save-buffer))
      partner)))

(defun tla-partner-list (&optional local-tree)
  "Reads the partner list from partner files in LOCAL-TREE, or
`default-directory' if local-tree is nil"
  (let ((default-directory (or local-tree default-directory)))
    (with-current-buffer (tla-partner-find-partner-file)
      (let ((partners (split-string (buffer-substring (point-min) (point-max)) "\n")))
        (remove "" partners)))))

(defun tla-partner-select (&optional prompt including-self)
  "Select a partner from either the xtla partner file or the
bookmarks. If INCLUDING-SELF is non-nil and a user gives an empty
string from minibuffer, this function can return `self'.

This is just a wrapper calling the function defined in
`tla-partner-select-function'."
  (funcall tla-partner-select-function prompt including-self))

(defvar tla-partner-select-history nil)
(defvar tla-partner-select-history-including-self nil)
(defun tla-partner-select-from-file (&optional prompt including-self)
  "Select a partner from the xtla partner file
If INCLUDING-SELF is non-nil and a user gives an empty string from minibuffer,
this function can return `self'."
  (unless prompt
    (setq prompt "Enter xtla partner: "))
  (save-window-excursion
    (let ((partners (tla-partner-list)))
      (if including-self
          (cond ((= 0 (length partners))
                 (if (y-or-n-p "Select self as default partner? ")
                     (progn
                       (setq tla-partner-select-was-interactive nil)
                       'self)
                   (setq tla-partner-select-was-interactive t)
                   (tla-partner-add (tla-partner-read prompt))))
                (t
                 (setq tla-partner-select-was-interactive t)
                 (completing-read
                  prompt
                  (mapcar 'list partners) nil nil nil
                  'tla-partner-select-history-including-self
                  'self)))
        (cond ((= 0 (length partners))
               (setq tla-partner-select-was-interactive t)
               (tla-partner-add (tla-partner-read prompt)))
              ((= 1 (length partners))
               (setq tla-partner-select-was-interactive nil)
               (car partners))
              (t
               (setq tla-partner-select-was-interactive t)
               (completing-read (format "%s (default %s) " prompt (car partners))
                                (mapcar 'list partners) nil nil nil
                                'tla-partner-select-history (car partners)
                                )))))))


;; (tla-partner-popup-menu)
(defun tla-partner-popup-menu (&optional prompt)
  (let ((list (tla-partner-list)))
    (popup-menu (easy-menu-create-menu prompt
                                       (mapcar
                                        (lambda (item)
                                          (let ((v (make-vector 3 nil)))
                                            (aset v 0 item) ; name
                                            (aset v 1 item) ; callback
                                            (aset v 2 t) ; enable
                                            ;;(aset v 3 :style)
                                            ;;(aset v 4 'radio)
                                            ;;(aset v 5 :selected)
                                            ;;(aset v 6 (if ...))
                                            v))
                                        list)))))

;; ----------------------------------------------------------------------------
;; tla-inventory-mode:
;; ----------------------------------------------------------------------------

(defun tla-inventory-mode ()
  "Major Mode to show the inventory of a tla working copy.

This allows you to view the list of files in your local tree. You can
display only some particular kinds of files with 't' keybindings:
'\\<tla-inventory-mode-map>\\[tla-inventory-toggle-source]' to toggle show sources,
'\\[tla-inventory-toggle-precious]' to toggle show precious, ...

Use '\\[tla-inventory-mark-file]' to mark files, and '\\[tla-inventory-unmark-file]' to unmark.
If you commit from this buffer (with '\\[tla-inventory-edit-log]'), then, the list of selected
files in this buffer at the time you actually commit with
\\<tla-log-edit-mode-map>\\[tla-log-edit-done].

Commands:
\\{tla-inventory-mode-map}
"
  (interactive)
  ;;  don't kill all local variables : this would clear the values of
  ;;  tla-inventory-display-*, and refresh wouldn't work well anymore.
  ;;  (kill-all-local-variables)
  (use-local-map tla-inventory-mode-map)
  (set (make-local-variable 'tla-buffer-refresh-function)
       'tla-inventory)
  (make-local-variable 'tla-buffer-marked-file-list)
  (easy-menu-add tla-inventory-mode-menu)
  (tla--install-buffer-menu)
  (setq major-mode 'tla-inventory-mode)
  (setq mode-name "tla-inventory")
  (setq mode-line-process 'tla-mode-line-process)
  (set (make-local-variable 'tla-get-file-info-at-point-function)
       'tla-inventory-get-file-info-at-point)
  (toggle-read-only 1)
  (run-hooks 'tla-inventory-mode-hook))

(defun tla-inventory-cursor-goto (ewoc-inv)
  "Move cursor to the ewoc location of EWOC-BOOKMARK"
  (interactive)
  (if ewoc-inv
      (progn (goto-char (ewoc-location ewoc-inv))
             (forward-char 6))
    (goto-char (point-min))))

(defun tla-inventory-next ()
  (interactive)
  (let* ((cookie tla-inventory-cookie)
         (elem (ewoc-locate cookie))
         (next (or (ewoc-next cookie elem) elem)))
    (tla-inventory-cursor-goto next)))

(defun tla-inventory-previous ()
  (interactive)
  (let* ((cookie tla-inventory-cookie)
         (elem (ewoc-locate cookie))
         (previous (or (ewoc-prev cookie elem) elem)))
    (tla-inventory-cursor-goto previous)))

(defun tla-inventory-edit-log (&optional insert-changelog)
  "Wrapper around `tla-edit-log', setting the source buffer to current
buffer."
  (interactive "P")
  (tla-edit-log insert-changelog (current-buffer)))

(defun tla-inventory-add-files (files)
  "Create explicit inventory ids for FILES."
  (interactive
   (list
    (if tla-buffer-marked-file-list
        (progn
          (unless (y-or-n-p (if (eq 1 (length tla-buffer-marked-file-list))
                                (format "Add %s? "
                                        (car tla-buffer-marked-file-list))
                              (format "Add %s files? "
                                      (length tla-buffer-marked-file-list))))
            (error "Not adding any file"))
          tla-buffer-marked-file-list)
      (list (read-file-name "Add file: " default-directory
                            nil nil
                            (tla--get-file-info-at-point))))))
  (apply 'tla-add nil files)
  (tla-inventory))

(defun tla-inventory-remove-files (files id-only)
  "Remove explicit inventory ids of FILES.
If ID-ONLY is nil, remove the files as well."
  (interactive
   (let ((read-files
          (if tla-buffer-marked-file-list
              (progn
                (unless (y-or-n-p
                         (format "Remove %d MARKED file%s? "
                                 (length tla-buffer-marked-file-list)
                                 (if (< (length tla-buffer-marked-file-list) 2)
                                        "" "s")))
                  (error "Not removing any files"))
                tla-buffer-marked-file-list)
            (list (read-file-name "Remove file: " default-directory
                                  nil nil
                                  (tla--get-file-info-at-point))))))
     (list read-files (not (y-or-n-p (format "Delete %d %sfile%s also locally? "
                                             (length read-files)
                                             (if tla-buffer-marked-file-list "MARKED " "")
                                             (if (< (length read-files) 2) "" "s")))))))
  (apply 'tla-remove id-only files)
  (tla-inventory))

(defun tla-inventory-delete-files (files no-questions)
  "Delete files locally.
This is here for convenience to delete left over, temporary files or files
avoiding a commit or conflicting with tree-lint.

It is not meant to delete tla managed files, i.e. files with IDs will be
passed to `tla-inventory-remove-files'!

When called with a prefix arg, do not annoy the user with questions, but just
delete the files."
  (interactive
   (list
    (if tla-buffer-marked-file-list
        (progn
          (or current-prefix-arg
              (unless (yes-or-no-p
                       (format "Delete %d files permanently? "
                               (length tla-buffer-marked-file-list)))
                (error "Not deleting any files")))
          tla-buffer-marked-file-list)
      (if (or current-prefix-arg
              (yes-or-no-p (format "Delete file %S permanently? "
                                   (tla--get-file-info-at-point))))
          (list (tla--get-file-info-at-point))))
    current-prefix-arg))
  (while files
    (let ((f (car files)))
      (if (= 0 (tla--run-tla-sync (list "id" f)
                                  :finished 'tla--status-handler
                                  :error 'tla--status-handler))
          (if (or no-questions
                  (y-or-n-p (format (concat "File %s is arch managed! "
                                            "Delete it with its id?") f)))
              (tla-inventory-remove-files (list f) nil))
        (if (file-directory-p f)
            (condition-case nil
                (delete-directory f)
              (file-error
               (if (or no-questions
                       (y-or-n-p (format "Delete non-empty directory %S? " f)))
                   (dired-recursive-delete-directory f))))
          (delete-file f))))
    (setq files (cdr files)))
  (tla-inventory))

(defun tla-inventory-move ()
  "Rename file at the current point and update its inventory id if present."
  (interactive)
  (if (eq 0 (tla-move (tla--get-file-info-at-point) nil 'ask))
      (tla-generic-refresh)
    (tla--show-last-process-buffer)))

(defun tla-inventory-revert ()
  "Reverts file at point"
  (interactive)
  (let* ((file (tla--get-file-info-at-point))
         (absolute (if (file-name-absolute-p file)
                       file
                     (expand-file-name
                      (concat default-directory "/" file)))))
    (tla-file-revert absolute)))

(defun tla-inventory-undo (specify-revision)
  "Undo whole local tree associated with the current inventory buffer.
If prefix arg, SPECIFY-REVISION is non-nil, read a revision and use it to undo."
  (interactive "P")
  (let* ((tree (tla-tree-root default-directory t))
         (revision (if specify-revision
                       (tla--read-archive-category-branch-version-revision-name-for-tree
                        tree "Undo against archive: ")
                     (list nil nil nil nil nil))))
    (apply 'tla--undo-internal (cons tree revision))))


(defun tla-file-has-conflict-p (file-name)
  (let ((rej-file-name (concat default-directory
                               (file-name-nondirectory file-name)
                               ".rej")))
    (file-exists-p rej-file-name)))

(defun tla-inventory-find-file ()
  (interactive)
  (let* ((file (tla--get-file-info-at-point)))
    (cond
     ((not file)
      (error "no file at point"))
     ((car (file-attributes file))      ; file is a directory
      (tla-inventory (expand-file-name file)))
     (t
      (find-file file)
      (when (tla-file-has-conflict-p file)
        (smerge-mode t))))))

(defun tla-inventory-parent-directory ()
  "Go to parent directory in inventory mode"
  (interactive)
  (tla-inventory (expand-file-name "..")))

(defun tla-inventory-find-file-other-window ()
  (interactive)
  (let ((file (tla--get-file-info-at-point)))
    (if file
        (progn
          (find-file-other-window file)
          (when (tla-file-has-conflict-p file)
            (smerge-mode t)))
      (error "no file at point"))))

(defun tla-inventory-view-file ()
  (interactive)
  (let ((file (tla--get-file-info-at-point)))
    (if file
        (view-file-other-window file)
      (error "no file at point"))))

(defun tla-inventory-mirror ()
  (interactive)
  (let ((tree-version (tla-tree-version-list)))
    (tla-archive-mirror (tla-archive-name  tree-version)
                        (tla-category-name tree-version)
                        (tla-branch-name   tree-version)
                        (tla-version-name  tree-version))))

(defun tla-inventory-star-merge ()
  "Run tla star-merge.
Either use a partner in (concat (tla-tree-root) \"/++tla-partners\")
Or ask the user for the merge partner."
  (interactive)
  (let ((merge-partner (tla-partner-select "Star-merge with: ")))
    (when (or tla-partner-select-was-interactive
              (y-or-n-p (format "Star-merge with %s ? " merge-partner)))
      (tla-star-merge merge-partner))))

(defun tla-inventory-changes (summary)
  "Run tla changes.
A prefix argument decides whether the user is asked for a diff partner
and whether only a summary without detailed diffs will be shown.

When called without a prefix argument: Show the changes for your tree.
When called with C-u as prefix: Ask the user for a diff partner via tla-partner-select.
When called with a negative prefix: Show only a summary of the changes.
When called with C--C-u as prefix: Ask the user for a diff partner, show only change summary."
  (interactive "P")
  (let* ((ask-for-compare-partner (and summary (listp summary)))
         (compare-partner (if ask-for-compare-partner
                              (tla-partner-select "Compare with (default is your tree): " t)
                            'self)))
    (if (eq 'self compare-partner)
        (setq compare-partner nil)
      (setq compare-partner (list 'revision (tla--name-split compare-partner))))
    (when (listp summary)
      (setq summary (car summary)))
    (tla-changes summary compare-partner)))

(defun tla-inventory-replay ()
  "Run tla replay.
Either use a partner in (concat (tla-tree-root) \"/++tla-partners\")
Or ask the user for the merge partner."
  (interactive)
  (let ((merge-partner (tla-partner-select "Replay from: ")))
    (when (or tla-partner-select-was-interactive
              (y-or-n-p (format "Replay from %s ? " merge-partner)))
      (tla-replay merge-partner))))

(defun tla-inventory-update ()
  "Run tla update."
  (interactive)
  (tla-update default-directory))

(defun tla-inventory-missing (&optional arg)
  "Run tla missing in default-directory.
With an prefix ARG, do this for the archive of one of your partners."
  (interactive "P")
  (if arg
      (let ((missing-partner (tla-partner-select "Check missing against: ")))
        (when (or tla-partner-select-was-interactive
                  (y-or-n-p (format "Check missing against %s ? " missing-partner)))
          (tla-missing default-directory missing-partner)))
    (tla-missing default-directory (tla-tree-version))))

(defun tla-inventory-file-ediff (&optional file)
  (interactive (list (caddr (ewoc-data (ewoc-locate tla-inventory-cookie)))))
  (tla-file-ediff file))

(defun tla-inventory-find-file-by-mouse (event)
  (interactive "e")
  (mouse-set-point event)
  (tla-inventory-find-file))

(defun tla-cat-log (revision-spec)
  (interactive "sRevision spec: ")
  (tla--run-tla-sync (list "cat-log" revision-spec)
                     :finished 'tla--finish-function-without-buffer-switch)
  (tla--show-last-process-buffer 'cat-log 'tla-cat-log-mode))

(defun tla-cat-archive-log (revision-spec)
  "Run cat-archive-log."
  (interactive "sRevision spec: ")
  (tla--run-tla-sync (list "cat-archive-log" revision-spec)
                     :finished 'tla--finish-function-without-buffer-switch)
  (tla--show-last-process-buffer 'cat-log 'tla-cat-log-mode))

(defun tla--maybe-save-log (revision)
  "Must be called from the buffer containing the log for REVISION.
Saves this buffer to the corresponding file in the log-library if
`tla-log-library-greedy' is non-nil."
  (if tla-log-library-greedy
      (let ((dir (expand-file-name
                  (concat tla-log-library "/" (car revision))))
            (file (tla--name-construct-semi-qualified (cdr revision))))
        (unless (file-directory-p dir)
          (make-directory dir))
        (let ((name (concat " *tla-log-rev-" (tla--name-construct
                                              revision) "*"))
              make-backup-files)
          (write-file (concat dir "/" file))
          (set-visited-file-name (concat dir "/" file))
          (set-buffer-modified-p nil)
          (rename-buffer name)
          (current-buffer)))
    (clone-buffer)))

(defun tla--cat-log-any (revision &optional tree async-handler)
  "Creates a buffer containing the log file for REVISION. Either call
cat-log, cat-archive-log, or read the log from the log library.

REVISION must be specified as a list. If TREE is provided, try a
cat-log in TREE preferably. Otherwise, try a cat-log in the local
directory. If both are impossible, run cat-archive-log. (same result,
but needs to retrieve something from the archive).

Call the function ASYNC-HANDLER in the created buffer, with arguments
 (output error status arguments)."
  ;;  (message "tla-cat-log-any %S" revision)
  ;; See if the log is in the log library
  (when tla-log-library-greedy
    (if (not (file-directory-p tla-log-library))
        (make-directory tla-log-library)))
  (let* ((lib-log (concat tla-log-library "/"
                          (tla--name-construct revision)))
         (buffer
          (or (get-file-buffer lib-log)
              (when (file-exists-p lib-log)
                (let* ((name (concat " *tla-log-rev-"
                                     (tla--name-construct revision) "*")))
                  (or (get-buffer name)
                      ;; Surprisingly, (rename-buffer) didn't rename
                      ;; anything here. Solution: Create a buffer with
                      ;; the right name, and simulate a find-file.
                      (with-current-buffer
                          (get-buffer-create name)
                        (insert-file-contents lib-log)
                        (set-visited-file-name lib-log)
                        (rename-buffer name)
                        (set-buffer-modified-p nil)
                        (current-buffer))))))))
    (if buffer
        (if async-handler
            (funcall async-handler buffer nil 0 "cat-log")
          buffer)
      ;; Try a cat-log
      (let* ((revision-string (tla--name-construct revision)))
        (let ((run-mode (if async-handler 'tla--run-tla-async 'tla--run-tla-sync))
              (handler (if async-handler
                           `(lambda (output error status arguments)
                              (with-current-buffer output
                                (tla--maybe-save-log ',revision))
                              (funcall ,async-handler output error status
                                       arguments))
                         `(lambda (output error status arguments)
                            (with-current-buffer output
                              (tla--maybe-save-log ',revision))))))
          (tla--run-tla-sync ;; Anyway, tla cat-log is fast, so, no
           ;; need for an asynchronous process. For some reason,
           ;; running it asynchronously caused a random bug when
           ;; running tla remotely.
           (list "cat-log" revision-string)
           :finished handler
           ;; cat-log failed: cat-archive-log is needed
           :error `(lambda (output error status arguments)
                     (funcall ',run-mode
                              (list "cat-archive-log"
                                    ,revision-string)
                              :finished ',handler))))))))

(defun tla-inventory-delta ()
  "Run tla delta.
Use the head revision of the version associated with the current inventory buffer
as modified tree. Give the base tree interactively."
  (interactive)
  (let* ((modified (tla-tree-version-list))
         (modified-revision (apply 'tla--version-head modified))
         (modified-fq (tla--name-construct
                       (tla-archive-name modified)
                       (tla-category-name modified)
                       (tla-branch-name modified)
                       (tla-version-name modified)
                       modified-revision))
         (base (tla--read-archive-category-branch-version-revision-name
                (format "Archive for delta to %s(HEAD) from: " modified-fq)
                (format "Category for delta to %s(HEAD) from: " modified-fq)
                (format "Branch for delta to %s(HEAD) from: " modified-fq)
                (format "Version for delta to %s(HEAD) from: " modified-fq)
                (format "Revision for delta to %s(HEAD) from: " modified-fq)))
         (base-fq (tla--name-construct base)))
    (tla-delta base-fq modified-fq 'ask)))


(defun tla-inventory-apply-changeset (reverse)
  "Apply changeset to the tree visited by the current inventory buffer."
  (interactive "P")
  (let ((inventory-buffer (current-buffer))
        (target (tla-tree-root))
        (changeset (expand-file-name (tla--read-directory-name
                                      "Changeset Directory: "
                                      nil nil t))))
    (tla-show-changeset changeset t)
    (when (y-or-n-p (format "Apply the changeset%s? "
                            (if reverse " in REVERSE" "")))
      (tla-apply-changeset changeset target reverse)
      (with-current-buffer inventory-buffer
        (tla-generic-refresh)))))

(defun tla-inventory-add-to-exclude (=tagging-method)
  "Add the file/directory under the point as an exclude entry
of .arch-inventory or =tagging-method file."
  (interactive "P")
  (tla--inventory-add-to-* "exclude" =tagging-method))

(defun tla-inventory-add-to-junk (=tagging-method)
  "Add the file/directory under the point as a junk entry
of .arch-inventory or =tagging-method file."
  (interactive "P")
  (tla--inventory-add-to-* "junk" =tagging-method))

(defun tla-inventory-add-to-backup (=tagging-method)
  "Add the file/directory under the point as a backup entry
of .arch-inventory or =tagging-method file."
  (interactive "P")
  (tla--inventory-add-to-* "backup" =tagging-method))

(defun tla-inventory-add-to-precious (=tagging-method)
  "Add the file/directory under the point as a precious entry
of .arch-inventory or =tagging-method file."
  (interactive "P")
  (tla--inventory-add-to-* "precious" =tagging-method))

(defun tla-inventory-add-to-unrecognized (=tagging-method)
  "Add the file/directory under the point as an unrecognized entry
of .arch-inventory or =tagging-method file."
  (interactive "P")
  (tla--inventory-add-to-* "unrecognized" =tagging-method))

(defun tla--inventory-add-to-* (category =tagging-method)
  "Added marked files or a file under the point to =tagging-method or .arch-inventory file.
If =TAGGING-METHOD is t, entries for the files are added to =tagging-method.
Else, they are added to .arch-inventory.
CATEGORY is one of the following strings: \"unrecognized\", \"precious\", \"backup\", \"junk\" or \"exclude\"."
  (let ((files
         (if tla-buffer-marked-file-list
             (if (y-or-n-p (format "Really want to make %d file%s %s? in %s file"
                                   (length tla-buffer-marked-file-list)
                                   (if (eq 1 (length tla-buffer-marked-file-list))
                                       "" "s")
                                   (upcase category)
                                   (if =tagging-method
                                       "=tagging-method" ".arch-inventory")))
                 tla-buffer-marked-file-list
               (error "making the files to %s is canceled" (upcase category)))
           (list (tla--get-file-info-at-point))))
        (point (point))
        (basedir (expand-file-name default-directory)))
    ;; Write down
    (save-excursion
      (mapc (lambda (file)
              (if =tagging-method
                  (tla-edit-=tagging-method)
                (tla-edit-.arch-inventory
                 (concat basedir (file-name-directory file))))
              (tla--inventory-file-add-file
               category (file-name-nondirectory file))
              (save-buffer)) files))
    ;; Keep the position
    (prog1
        (tla-generic-refresh)
      (if (< point (point-max))
          (goto-char point)))))

;; ----------------------------------------------------------------------------
;; tla-cat-log-mode:
;; ----------------------------------------------------------------------------
(defun tla-cat-log-mode ()
  "Major Mode to show a specific log message
Commands:
\\{tla-cat-log-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (use-local-map tla-cat-log-mode-map)
  (set (make-local-variable 'font-lock-defaults)
       '(tla-cat-log-font-lock-keywords t))
  (setq major-mode 'tla-cat-log-mode)
  (setq mode-name "tla-cat-log")
  (toggle-read-only 1)
  (run-hooks 'tla-cat-log-mode-hook))

;; Obsolete
(defun tla-log-merges (revision &optional callback)
  "Return a list that will contain patches merged by REVISION.  When the list
  has been filled in, CALLBACK is called with no arguments."
  (let ((merges (list "")))
    (tla--cat-log-any
     revision nil
     `(lambda (output error status args)
        (with-current-buffer output
          (goto-char (point-min))
          (unwind-protect
              (let ((list (split-string
                           (buffer-substring-no-properties
                            (re-search-forward "^New-patches: ")
                            (progn (re-search-forward "^[^\t ]")
                                   (beginning-of-line) (point))))))
                (setq list
                      (remove (tla--name-construct
                               ',revision)
                              list))
                (setcar ',merges (car list))
                (setcdr ',merges (cdr list)))
            (when ',callback (funcall ',callback))
            (kill-buffer nil)))))
    merges))

;; ----------------------------------------------------------------------------
;; tla-log-edit-mode:
;; ----------------------------------------------------------------------------
(defun tla-log-edit-next-field ()
  "Go to next field in a log edition"
  (interactive)
  (let ((in-field (string-match "^\\([A-Z][A-Za-z]*\\(: ?\\)?\\)?$"
                                (buffer-substring
                                 (line-beginning-position) (point)))))
    (if (and in-field
             (string-match "^[A-Z][A-Za-z]*: $"
                           (buffer-substring
                            (line-beginning-position) (point))))
        (forward-line))
    (if in-field (beginning-of-line) (forward-line 1))
    (or (and (looking-at "^[A-Z][a-zA-Z]*: ")
             (goto-char (match-end 0)))
        (and (looking-at "^[A-Z][a-zA-Z]*:$")
             (goto-char (match-end 0))
             (progn (insert " ") t))
        (goto-char (point-max)))))

(defun tla-log-goto-field (field)
  "Go to FIELD in a log file."
  (goto-char (point-min))
  (re-search-forward field)
  (save-excursion
    (if (not (looking-at " "))
        (insert " ")))
  (forward-char 1))

(defun tla-log-goto-summary ()
  "Go to the Summary field in a log file."
  (interactive)
  (tla-log-goto-field "^Summary:"))

(defun tla-log-goto-keywords ()
  "Go to the Keywords field in a log file."
  (interactive)
  (tla-log-goto-field "^Keywords:"))

(defun tla-log-goto-body ()
  "Go to the Body in a log file."
  (interactive)
  (goto-char (point-min))
  (forward-line 3))

(defun tla-log-kill-body ()
  "Kill the content of the log file body."
  (interactive)
  (tla-log-goto-body)
  (kill-region (point) (point-max)))

;;;###autoload
(add-to-list 'auto-mode-alist '("\\+\\+log\\." . tla-log-edit-mode))
;;;###autoload
(define-derived-mode tla-log-edit-mode text-mode "tla-log-edit"
  "Major Mode to edit xtla log messages.
Commands:
\\{tla-log-edit-mode-map}
"
  (use-local-map tla-log-edit-mode-map)
  (easy-menu-add tla-log-edit-mode-menu)
  (tla--install-buffer-menu)
  (set (make-local-variable 'font-lock-defaults)
       '(tla-log-edit-font-lock-keywords t))
  (setq fill-column 73)
  (run-hooks 'tla-log-edit-mode-hook))

(defun tla-log-edit-abort ()
  (interactive)
  (bury-buffer)
  (set-window-configuration tla-pre-commit-window-configuration))

(defun tla-log-edit-done ()
  (interactive)
  (save-buffer)
  (let ((log-buffer (current-buffer)))
    (if (equal (tla-commit) 0)
        (kill-buffer log-buffer))))

(defun tla-archive-maintainer-name (archive)
  "Return the maintainer name for a given ARCHIVE.
This functions looks in the bookmarks file for the nickname field and returns it.
If the nickname field is not present, just return ARCHIVE."
  (tla-bookmarks-load-from-file)
  (dolist (elem tla-bookmarks-alist)
    (when (string= (cadr (assoc 'location elem)) archive)
      (return (or (cadr (assoc 'nickname elem)) archive)))))

(defun tla-archive-maintainer-id (archive &optional shorter)
  "Return my-id substring from ARCHIVE.
If SHORTER is non-nil, return login name part of the my-id substring.
E.g. If ARCHIVE is x@y.z--a, the result is x@y.z.
If SHORTER is non-nil, the result is x."
  (if (string-match "\\(\\(.+\\)@.+\\)--.+" archive)
      (if shorter
          (match-string 2 archive)
        (match-string 1 archive))))


(defun tla-merge-summary-line (mergelist)
  (concat "Merged from "
          (mapconcat '(lambda (m)
                        (format "%s (patch%s%s)" (car m)
                                (cadr m)
                                (if (caddr m) (concat "-" (caddr m)) "")))
                     mergelist", ")))

(defun tla-merge-summary-line-for-log ()
  (save-excursion
    (let ((rev-list)
          (maintainer)
          (archive)
          (rev)
          (patch-list))
      (goto-char (point-min))
      (while (re-search-forward "^ \\* \\(.+@.+--.+/.+--.+\\)$" nil t)
        (setq rev-list (tla--name-split (match-string-no-properties 1)))
        (setq archive (tla-archive-name rev-list))
        (setq maintainer (or (tla-archive-maintainer-name archive)
                             (tla-archive-maintainer-id archive t)))
        (setq rev (cadr (split-string (tla-revision-name rev-list) "-")))
        (add-to-list 'patch-list (list maintainer rev)))
      (add-to-list 'patch-list (list nil nil))
      (setq patch-list (nreverse patch-list))
      (let ((last-maintainer)
            (patch-start)
            (patch-end)
            (rev patch-list)
            (mergelist))
        (while rev
          (if (string= (caar rev) last-maintainer)
              (setq patch-end (cadar rev))
            (when last-maintainer
              (tla--add-to-list 'mergelist (list last-maintainer patch-start patch-end) t))
            (setq last-maintainer (caar rev)
                  patch-start (cadar rev)
                  patch-end nil))
          (setq rev (cdr rev)))
        (tla-merge-summary-line mergelist)))))

(defun tla-log-edit-insert-log-for-merge (arg)
  "Insert the output of tla log-for-merge at POINT.
When called with a prefix argument:
Create a standard Merged from line as Summary."
  (interactive "P")
  (tla--run-tla-sync '("log-for-merge")
                     :finished
                     `(lambda (output error status arguments)
                        (let ((content (tla--buffer-content
                                        output)))
                          (if (= 0 (length content))
                              (error "There was no merge!"))
                          (with-current-buffer ,(current-buffer)
                            (let ((on-summary-line
                                   (= 1 (count-lines (point-min) (point))))
                                  (old-pos (point)))
                              (if on-summary-line
                                  (tla-log-goto-body)
                                (goto-char old-pos))
                              (insert content)))
                          (when arg
                            (tla-log-goto-summary)
                            (delete-region (point) (line-end-position))
                            (insert
                             (with-current-buffer output
                               (tla-merge-summary-line-for-log)))
                            (tla-log-goto-keywords)
                            (delete-region (point) (line-end-position))
                            (insert "merge"))))))


;; ----------------------------------------------------------------------------
;; tla-log-edit-insert-keywords:
;; ----------------------------------------------------------------------------

(defvar tla-log-edit-keywords-marked-list)
(defvar tla-log-edit-keywords-cookie)
(defvar tla-log-edit-keywords-log-buffer)

(defun tla-log-edit-keywords-printer (elem)
  (insert (if (member elem tla-log-edit-keywords-marked-list) "* " "  ")
          elem))

(defun tla-log-edit-keywords (arg)
  "Adds keywords listed in `tla-log-edit-keywords'.
When called with a prefix argument, delete all current keywords."
  (interactive "P")
  (let ((current-keywords
         (save-excursion
           (tla-log-goto-keywords)
           (buffer-substring (point) (line-end-position))))
        (log-buffer (current-buffer))
        keywords)
    (setq current-keywords (replace-regexp-in-string "," " " current-keywords nil t)
          current-keywords (mapcar (lambda (k) (format "%s" k))
                                   (read (concat "(" current-keywords ")"))))
    (switch-to-buffer " *tla-log-keywords*")
    (toggle-read-only 0)
    (erase-buffer)
    (make-local-variable 'tla-log-edit-keywords)
    (make-local-variable 'tla-log-edit-keywords-marked-list)
    (make-local-variable 'tla-log-edit-keywords-cookie)
    (make-local-variable 'tla-log-edit-keywords-log-buffer)
    (setq tla-log-edit-keywords-log-buffer
          log-buffer
          tla-log-edit-keywords-marked-list
          current-keywords
          tla-log-edit-keywords-cookie
          (ewoc-create 'tla-log-edit-keywords-printer
                       "List of keywords from `tla-log-edit-keywords':"
                       (format "type C-c C-c to insert the marked keywords to the buffer\n%s"
                               (buffer-name log-buffer))))

    (while current-keywords
      (add-to-list 'tla-log-edit-keywords (car current-keywords))
      (setq current-keywords (cdr current-keywords)))

    (setq keywords tla-log-edit-keywords)

    (while keywords
      (add-to-list 'tla-log-edit-keywords (car keywords))
      (ewoc-enter-last tla-log-edit-keywords-cookie (car keywords))
      (setq keywords (cdr keywords))))

  (use-local-map tla-log-edit-keywords-mode-map)
  (setq major-mode 'tla-log-edit-keywords-mode)
  (setq mode-name "tla-log-keywords")
  (toggle-read-only 1)
  (message "Type C-c C-c to finish.")
  (goto-char (point-min))
  (forward-line 1))

(defun tla-log-edit-keywords-cursor-goto (elem)
  (interactive)
  (goto-char (ewoc-location elem))
  (re-search-forward "^"))

(defun tla-log-edit-keywords-next ()
  (interactive)
  (let* ((cookie tla-log-edit-keywords-cookie)
         (elem (ewoc-locate cookie))
         (next (or (ewoc-next cookie elem) elem)))
    (tla-log-edit-keywords-cursor-goto next)))

(defun tla-log-edit-keywords-previous ()
  (interactive)
  (let* ((cookie tla-log-edit-keywords-cookie)
         (elem (ewoc-locate cookie))
         (previous (or (ewoc-prev cookie elem) elem)))
    (tla-log-edit-keywords-cursor-goto previous)))

(defun tla-log-edit-keywords-mark ()
  (interactive)
  (let ((pos (point)))
    (add-to-list 'tla-log-edit-keywords-marked-list
                 (ewoc-data (ewoc-locate tla-log-edit-keywords-cookie)))
    (ewoc-refresh tla-log-edit-keywords-cookie)
    (goto-char pos))
  (tla-log-edit-keywords-next))

(defun tla-log-edit-keywords-unmark ()
  (interactive)
  (let ((pos (point)))
    (setq tla-log-edit-keywords-marked-list
          (delete (ewoc-data (ewoc-locate tla-log-edit-keywords-cookie))
                  tla-log-edit-keywords-marked-list))
    (ewoc-refresh tla-log-edit-keywords-cookie)
    (goto-char pos))
  (tla-log-edit-keywords-next))

(defun tla-log-edit-keywords-unmark-all ()
  (interactive)
  (let ((pos (point)))
    (setq tla-log-edit-keywords-marked-list nil)
    (ewoc-refresh tla-log-edit-keywords-cookie)
    (goto-char pos)))

(defun tla-log-edit-keywords-mark-all ()
  (interactive)
  (let ((pos (point)))
    (setq tla-log-edit-keywords-marked-list tla-log-edit-keywords)
    (ewoc-refresh tla-log-edit-keywords-cookie)
    (goto-char pos)))

(defun tla-log-edit-keywords-toggle-mark ()
  (interactive)
  (let ((pos (point)))
    (if (member (ewoc-data (ewoc-locate tla-log-edit-keywords-cookie))
                tla-log-edit-keywords-marked-list)
        (tla-log-edit-keywords-unmark)
      (tla-log-edit-keywords-mark))
    (ewoc-refresh tla-log-edit-keywords-cookie)
    (goto-char pos)))

(defun tla-log-edit-keywords-insert ()
  "Insert marked keywords into log buffer."
  (interactive)
  (let ((keywords tla-log-edit-keywords-marked-list))
    (switch-to-buffer tla-log-edit-keywords-log-buffer)
    (kill-buffer " *tla-log-keywords*")
    (save-excursion
      (tla-log-goto-keywords)
      (delete-region (point) (line-end-position))
      (insert (mapconcat 'identity (reverse keywords) ", ")))))

;; ----------------------------------------------------------------------------
;; tla-archive-list-mode:
;; ----------------------------------------------------------------------------
(defun tla-archive-mirror-archive ()
  "Mirrors archive at point"
  (interactive)
  (let ((archive-info (tla-get-archive-info)))
    (when archive-info
      (tla-mirror-archive archive-info))))

(defun tla-archive-synchronize-archive ()
  "Synchronizes the mirror for the archive at point"
  (interactive)
  (let ((archive-info (tla-get-archive-info)))
    (when archive-info
      (tla-archive-mirror archive-info))))

(defun tla-archive-list-mode ()
  "Major Mode to show arch archives:
\\{tla-archive-list-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (use-local-map tla-archive-list-mode-map)
  (easy-menu-add tla-archive-list-mode-menu)
  (tla--install-buffer-menu)
  (setq major-mode 'tla-archive-list-mode)
  (setq mode-name "tla-archives")

  (toggle-read-only 1)
  (set-buffer-modified-p nil)
  (set (make-local-variable 'tla-get-revision-info-at-point-function)
       'tla--get-archive-info-at-point)
  (run-hooks 'tla-archive-list-mode-hook))

(defun tla--get-archive-info-at-point ()
  "Get archive information."
  (list 'archive (tla-get-archive-info)))

(defun tla-archive-select-default ()
  (interactive)
  (when (tla-get-archive-info)
    (let ((pos (point)))
      (tla-my-default-archive (tla-get-archive-info))
      (tla-archives)
      (goto-char pos))))

(defun tla-archive-unregister-archive ()
  "Delete the registration of the selected archive."
  (interactive)
  (let ((archive (tla-get-archive-info)))
    (if archive
        (when (yes-or-no-p (format "Delete the registration of %s? " archive))
          (tla--run-tla-sync  (list "register-archive" "--delete" archive)
                              :finished
                              (lambda (output error status arguments)
                                (message "Deleted the registration of %s" archive)
                                (tla-archives))))
      (error "No archive under the point"))))

(defun tla-archive-list-categories ()
  (interactive)
  (let ((archive (tla-get-archive-info)))
    (if archive
        (tla-categories archive)
      (error "No archive under the point"))))

(defun tla-archive-list-categories-by-mouse (event)
  (interactive "e")
  (mouse-set-point event)
  (tla-archive-list-categories))

(defun tla-archive-browse-archive ()
  (interactive)
  (let ((archive (tla-get-archive-info)))
    (if archive
        (tla-browse-archive archive)
      (error "No archive under the point"))))

(defun tla-archive-next ()
  (interactive)
  (forward-line 2)
  (beginning-of-line))

(defun tla-archive-previous ()
  (interactive)
  (forward-line -2)
  (beginning-of-line))

(defun tla-save-archive-to-kill-ring ()
  "Save archive name associated with current buffer or under point to `kill-ring'"
  (interactive)
  (let ((archive (or (tla-get-archive-info)
                     tla-buffer-archive-name
                     (tla-archive-name (tla-tree-version-list nil 'no-error)))))
    (unless archive
      (error "No archive name associated with current buffer"))
    (kill-new archive)
    (if (interactive-p)
        (message "%s" archive))
    archive))

;; ----------------------------------------------------------------------------
;; tla-category-list-mode:
;; ----------------------------------------------------------------------------
(defun tla-category-list-mode ()
  "Major Mode to show arch categories:
\\{tla-category-list-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (use-local-map tla-category-list-mode-map)
  (easy-menu-add tla-category-list-mode-menu)
  (tla--install-buffer-menu)
  (setq major-mode 'tla-category-list-mode)
  (setq mode-name "tla-category")
  (add-hook 'tla-make-category-hook 'tla-category-update)

  (toggle-read-only 1)
  (set-buffer-modified-p nil)
  (set (make-local-variable 'tla-get-revision-info-at-point-function)
       'tla--get-category-info-at-point)
  (run-hooks 'tla-category-list-mode-hook))

(defun tla--get-category-info-at-point ()
  "Get archive/category--branch information."
  (let ((buffer-version (tla--name-construct
                         tla-buffer-archive-name
                         (tla-get-archive-info 'tla-category-info))))
    (list 'category buffer-version)))

(defun tla-category-list-branches ()
  (interactive)
  (let ((category (tla-get-archive-info 'tla-category-info)))
    (if category
        (tla-branches tla-buffer-archive-name category)
      (error "No category under the point"))))

(defun tla-category-list-branches-by-mouse (event)
  (interactive "e")
  (mouse-set-point event)
  (tla-category-list-branches))

(defun tla-category-make-category (category)
  (interactive "sCategory name: ")
  (tla-make-category tla-buffer-archive-name category))

(defun tla-category-update ()
  (interactive)
  (tla-categories tla-buffer-archive-name))

(defun tla-category-next ()
  (interactive)
  (forward-line 1)
  (beginning-of-line))

(defun tla-category-previous ()
  (interactive)
  (forward-line -1)
  (beginning-of-line)
  (unless (looking-at "^   ")
    (forward-line 1)))

(defun tla-category-mirror-archive ()
  (interactive)
  (let ((category (tla-get-archive-info 'tla-category-info)))
    (unless category
      (error "no category at point"))
    (tla-archive-mirror tla-buffer-archive-name
                        category)))


(defun tla-category-bookmarks-add-here (name)
  (interactive "sBookmark name: ")
  (tla-bookmarks-add name
                     (list tla-buffer-archive-name
                           (tla-get-archive-info 'tla-category-info)
                           nil nil nil))
  (message "bookmark %s added." name))

(defun tla-category-bookmarks-add (name)
  (interactive "sBookmark name: ")
  (tla-bookmarks-add name
                     (list tla-buffer-archive-name nil nil nil))
  (message "bookmark %s added." name))

;; ----------------------------------------------------------------------------
;; tla-branch-list-mode
;; ----------------------------------------------------------------------------
(defun tla-branch-list-mode ()
  "Major Mode to show arch branches:
\\{tla-branch-list-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (use-local-map tla-branch-list-mode-map)
  (easy-menu-add tla-branch-list-mode-menu)
  (tla--install-buffer-menu)
  (setq major-mode 'tla-branch-list-mode)
  (setq mode-name "tla-branch")
  (add-hook 'tla-make-branch-hook 'tla-branch-update)

  (toggle-read-only 1)
  (set-buffer-modified-p nil)
  (set (make-local-variable 'tla-get-revision-info-at-point-function)
       'tla--get-branch-info-at-point)
  (run-hooks 'tla-branch-list-mode-hook))

(defun tla--get-branch-info-at-point ()
  "Get archive/category--branch--version information."
  (let ((buffer-version (tla--name-construct
                         tla-buffer-archive-name
                         tla-buffer-category-name
                         (tla-get-archive-info 'tla-branch-info))))
    (list 'branch buffer-version)))

(defun tla-branch-make-branch (branch)
  (interactive "sBranch name: ")
  (tla-make-branch tla-buffer-archive-name
                   tla-buffer-category-name
                   branch))

(defun tla-branch-update ()
  (interactive)
  (tla-branches
   tla-buffer-archive-name
   tla-buffer-category-name))

(defun tla-branch-list-parent-category ()
  (interactive)
  (tla-categories tla-buffer-archive-name))

(defun tla-branch-list-versions ()
  (interactive)
  (let ((branch (tla-get-archive-info 'tla-branch-info)))
    (if branch
        (tla-versions tla-buffer-archive-name
                      tla-buffer-category-name
                      branch)
      (error "No branch under the point"))))

(defun tla-branch-list-versions-by-mouse (event)
  (interactive "e")
  (mouse-set-point event)
  (tla-branch-list-versions))

(defun tla-branch-mirror-archive ()
  (interactive)
  (let ((branch (tla-get-archive-info 'tla-branch-info)))
    (unless branch
      (error "No branch under the point"))
    (tla-archive-mirror tla-buffer-archive-name
                        tla-buffer-category-name
                        branch)))

(defun tla-branch-get-branch (directory)
  (interactive (list (expand-file-name
                      (tla--read-directory-name
                       (format "Restore \"%s\" to: "
                               (let ((branch
                                      (tla-get-archive-info 'tla-branch-info)))
                                 (unless branch
                                   (error "No branch under the point"))
                                 (tla--name-construct
                                  tla-buffer-archive-name
                                  tla-buffer-category-name
                                  branch)))))))
  (let ((branch (tla-get-archive-info 'tla-branch-info)))
    (if branch
        (tla-get directory
                 t
                 tla-buffer-archive-name
                 tla-buffer-category-name
                 branch)
      (error "No branch under the point"))))

(defun tla-branch-bookmarks-add-here (name)
  (interactive "sBookmark name: ")
  (tla-bookmarks-add name
                     (list tla-buffer-archive-name
                           tla-buffer-category-name
                           (tla-get-archive-info 'tla-branch-info)
                           nil nil))
  (message "bookmark %s added." name))

(defun tla-branch-bookmarks-add (name)
  (interactive "sBookmark name: ")
  (tla-bookmarks-add name
                     (list tla-buffer-archive-name
                           tla-buffer-category-name
                           nil nil nil))
  (message "bookmark %s added." name))




;; ----------------------------------------------------------------------------
;; tla-version-list-mode
;; ----------------------------------------------------------------------------
(defun tla-version-list-mode ()
  "Major Mode to show arch versions:
\\{tla-version-list-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (use-local-map tla-version-list-mode-map)
  (easy-menu-add tla-version-list-mode-menu)
  (tla--install-buffer-menu)
  (setq major-mode 'tla-version-list-mode)
  (setq mode-name "tla-version")
  (add-hook 'tla-make-version-hook 'tla-version-update)

  (toggle-read-only 1)
  (set-buffer-modified-p nil)
  (set (make-local-variable 'tla-get-revision-info-at-point-function)
       'tla--get-version-info-at-point)
  (run-hooks 'tla-version-list-mode-hook))

(defun tla--get-version-info-at-point ()
  "Get archive/category--branch--version--revision information."
  (let ((buffer-version (tla--name-construct
                         tla-buffer-archive-name
                         tla-buffer-category-name
                         tla-buffer-branch-name
                         (tla-get-archive-info 'tla-version-info))))
    (list 'version buffer-version)))

(defun tla-version-update ()
  (interactive)
  (tla-versions
   tla-buffer-archive-name
   tla-buffer-category-name
   tla-buffer-branch-name))

(defun tla-version-list-parent-branch ()
  (interactive)
  (tla-branches tla-buffer-archive-name
                tla-buffer-category-name))

(defun tla-version-list-revisions ()
  (interactive)
  (let ((version (tla-get-archive-info 'tla-version-info)))
    (if version
        (tla-revisions tla-buffer-archive-name
                       tla-buffer-category-name
                       tla-buffer-branch-name
                       version)
      (error "No version under the point"))))

(defun tla-version-list-revisions-by-mouse (event)
  (interactive "e")
  (mouse-set-point event)
  (tla-version-list-revisions))

(defun tla-version-make-version (version)
  (interactive "sVersion name: ")
  (tla-make-version tla-buffer-archive-name
                    tla-buffer-category-name
                    tla-buffer-branch-name
                    version))

(defun tla-version-bookmarks-add-here (name)
  (interactive "sBookmark name: ")
  (tla-bookmarks-add name
                     (list tla-buffer-archive-name
                           tla-buffer-category-name
                           tla-buffer-branch-name
                           (tla-get-archive-info 'tla-version-info)
                           nil))
  (message "bookmark %s added." name))

(defun tla-version-bookmarks-add (name)
  (interactive "sBookmark name: ")
  (tla-bookmarks-add name
                     (list tla-buffer-archive-name
                           tla-buffer-category-name
                           tla-buffer-branch-name
                           nil nil))
  (message "bookmark %s added." name))

(defun tla-version-get-version (directory)
  (interactive (list (expand-file-name
                      (tla--read-directory-name
                       (format "Restore \"%s\" to: "
                               (let ((version
                                      (tla-get-archive-info 'tla-version-info)))
                                 (unless version
                                   (error "No version under the point"))
                                 (tla--name-construct
                                  tla-buffer-archive-name
                                  tla-buffer-category-name
                                  tla-buffer-branch-name
                                  version)))))))
  (let ((version (tla-get-archive-info 'tla-version-info)))
    (if version
        (tla-get directory
                 t
                 tla-buffer-archive-name
                 tla-buffer-category-name
                 tla-buffer-branch-name
                 version)
      (error "No version under the point"))))


(defun tla-version-mirror-archive ()
  (interactive)
  (let ((version (tla-get-archive-info 'tla-version-info)))
    (if version
        (tla-archive-mirror tla-buffer-archive-name
                            tla-buffer-category-name
                            tla-buffer-branch-name
                            version))))

(defun tla-version-tag (to-archive to-category to-branch to-version)
  "Run tla tag from the current location in version buffer."
  (interactive
   (tla--read-archive-category-branch-version-name "Tag to: "))
  (let ((to-fq (tla--name-construct to-archive
                                    to-category
                                    to-branch
                                    to-version))
        from-fq
        (from-version (tla-get-archive-info 'tla-version-info)))
    (unless from-version
      (error "No version under the point"))
    (setq from-fq (tla--name-construct
                   tla-buffer-archive-name
                   tla-buffer-category-name
                   tla-buffer-branch-name
                   from-version))
    (tla--version-tag-internal from-fq to-fq)))


(defun tla--version-tag-internal (from-fq to-fq)
  (when (yes-or-no-p (format "Create a tag from `%s' to `%s'? " from-fq to-fq))
    (unless (tla-tag from-fq to-fq)
      (error "Fail to create a tag"))
    (when (y-or-n-p "Tag created. Get a copy of this revision? ")
      (apply 'tla-get (tla--read-directory-name "Get a copy in: ")
             'ask (tla--name-split to-fq)))))

;; ----------------------------------------------------------------------------
;; tla-revision-list-mode
;; ----------------------------------------------------------------------------
(defun tla-revision-list-mode ()
  "Major Mode to show arch revisions:
\\{tla-revision-list-mode-map}
"
  (interactive)
  (kill-all-local-variables)
  (toggle-read-only -1)
  (use-local-map tla-revision-list-mode-map)
  (easy-menu-add tla-revision-list-mode-menu)
  (tla--install-buffer-menu)
  (setq major-mode 'tla-revision-list-mode)
  (setq mode-name "tla-revision")
  (add-hook 'tla-make-revision-hook 'tla-revision-refresh)
  (erase-buffer)
  (set (make-local-variable 'tla-revision-list-cookie)
       (ewoc-create 'tla-revision-list-printer))
  (toggle-read-only 1)
  (set-buffer-modified-p nil)
  (set (make-local-variable 'tla-get-revision-info-at-point-function)
       'tla--get-revision-info-at-point)
  (setq mode-line-process 'tla-mode-line-process)
  (run-hooks 'tla-revision-list-mode-hook))

(defun tla--get-revision-info-at-point ()
  "Get archive/category--branch--version--revision--patch information."
  (let* ((elem (ewoc-data (ewoc-locate tla-revision-list-cookie)))
         (full (tla--revision-revision (caddr elem)))
         (buffer-version
          (tla--name-construct full)))
    (list 'revision buffer-version)))

(defun tla-revision-refresh ()
  (interactive)
  (tla-revisions
   tla-buffer-archive-name
   tla-buffer-category-name
   tla-buffer-branch-name
   tla-buffer-version-name))

(defun tla-revision-list-parent-version ()
  (interactive)
  (tla-versions tla-buffer-archive-name
                tla-buffer-category-name
                tla-buffer-branch-name))

(defun tla-revision-get-revision (directory archive category branch
                                            version revision)
  (interactive
   (let* ((elem (ewoc-data (ewoc-locate tla-revision-list-cookie)))
          (full (tla--revision-revision (caddr elem)))
          (revision (tla-revision-name full))
          (archive (tla-archive-name full))
          (category (tla-category-name full))
          (branch (tla-branch-name full))
          (version (tla-version-name full))
          dir)
     (unless revision
       (error "No revision under the point"))
     (setq dir (expand-file-name
                (tla--read-directory-name
                 (format "Restore \"%s\" to: "
                         (tla--name-construct
                          archive category branch version revision)))))
     (if (file-exists-p dir)
         (error "Directory %s already exists." dir))
     (list dir archive category branch version revision)))
  (if revision
      (tla-get directory t archive category branch version revision)
    (error "No revision under the point")))

(defun tla-revision-cache-revision (archive category branch version revision)
  "Create a cached revision for the revision at point"
  (interactive
   (let* ((elem (ewoc-data (ewoc-locate tla-revision-list-cookie)))
          (full (tla--revision-revision (caddr elem)))
          (revision (tla-revision-name full))
          (archive (tla-archive-name full))
          (category (tla-category-name full))
          (branch (tla-branch-name full))
          (version (tla-version-name full)))
     (unless revision
       (error "No revision under the point"))
     (list archive category branch version revision)))
  (if revision
      (tla-cache-revision archive category branch version revision)
    (error "No revision under the point")))

(defun tla-revision-add-to-library (archive category branch version revision)
  "Add the revision at point to library"
  (interactive
   (let* ((elem (ewoc-data (ewoc-locate tla-revision-list-cookie)))
          (full (tla--revision-revision (caddr elem)))
          (revision (tla-revision-name full))
          (archive (tla-archive-name full))
          (category (tla-category-name full))
          (branch (tla-branch-name full))
          (version (tla-version-name full)))
     (unless revision
       (error "No revision under the point"))
     (list archive category branch version revision)))
  (if revision
      (tla-library-add archive category branch version revision)
    (error "No revision under the point")))

(defun tla--revision-maybe-refresh ()
  "If the current ewoc doesn't contain creator, date, and summary, and
if these values should now be displayed, run the refresh function."
  (when (or tla-revisions-shows-date
            tla-revisions-shows-creator
            tla-revisions-shows-summary
            tla-revisions-shows-merges
            tla-revisions-shows-merged-by)
    (let ((stop nil)
          (ewoc-elem (ewoc-nth tla-revision-list-cookie 0)))
      (while (and ewoc-elem (not stop))
        (let ((elem (ewoc-data ewoc-elem)))
          (if (eq (car elem) 'entry-patch)
              (setq stop t)
            (setq ewoc-elem (ewoc-next tla-revision-list-cookie
                                       ewoc-elem)))))
      (when (and ewoc-elem
                 (null (tla--revision-summary (caddr (ewoc-data ewoc-elem)))))
        (tla-generic-refresh)))))

(defun tla-revision-toggle-date ()
  (interactive)
  (setq tla-revisions-shows-date (not tla-revisions-shows-date))
  (tla--revision-maybe-refresh)
  (ewoc-refresh tla-revision-list-cookie))

(defun tla-revision-toggle-summary ()
  (interactive)
  (setq tla-revisions-shows-summary (not tla-revisions-shows-summary))
  (tla--revision-maybe-refresh)
  (ewoc-refresh tla-revision-list-cookie))

(defun tla-revision-toggle-creator ()
  (interactive)
  (setq tla-revisions-shows-creator (not tla-revisions-shows-creator))
  (tla--revision-maybe-refresh)
  (ewoc-refresh tla-revision-list-cookie))

(defun tla-revision-toggle-library ()
  (interactive)
  (setq tla-revisions-shows-library (not tla-revisions-shows-library))
  (ewoc-refresh tla-revision-list-cookie))

(defun tla-revision-toggle-merges ()
  (interactive)
  (setq tla-revisions-shows-merges (not tla-revisions-shows-merges))
  (tla--revision-maybe-refresh)
  (ewoc-refresh tla-revision-list-cookie))

(defun tla-revision-toggle-merged-by ()
  (interactive)
  (setq tla-revisions-shows-merged-by
        (not tla-revisions-shows-merged-by))
  (when (and (not tla-revision-merge-by-computed)
             tla-revisions-shows-merged-by)
    (tla--revision-maybe-refresh)
    (tla-revision-compute-merged-by))
  (ewoc-refresh tla-revision-list-cookie))

(defun tla-revision-changeset (arg)
  "Gets and display the changeset at point in a revision list buffer.
If used with a prefix arg, don't include the diffs from the output."
  (interactive "P")
  (let* ((cookie tla-revision-list-cookie)
         (full (tla--revision-revision
                (caddr (ewoc-data (ewoc-locate cookie)))))
         (revision (tla--name-construct full)))
    (tla-get-changeset revision t nil arg)))

(defun tla-revision-store-delta (across-versions)
  (interactive "P")
  (tla-revision-delta across-versions t))

(defun tla-revision-delta (across-versions &optional stored-to-directory)
  "Runs tla delta from marked revision to revision at point.
If prefix-argument ACROSS-VERSIONS is nil, read a revision
in the current version. If ACROSS-VERSIONS is non-nil, read an archive,
a category, a branch, a version, and a revision to specify the revision.
If STORED-TO-DIRECTORY is nil, ask the user whether the changeset is stored
to or not. If STORED-TO-DIRECTORY is non-nil, don't ask the use and the changeset
is stored."
  (interactive "P")
  (let* ((modified
          (tla--revision-revision
           (caddr (ewoc-data (ewoc-locate tla-revision-list-cookie)))))
         (modified-fq (tla--name-construct modified))
         (base
          (let ((marked (tla-revision-marked-revisions)))
            (when (< 1 (length marked))
              (error "Delta can be run against one marked revision as the base revision."))
            (cond ((and marked (null (cdr marked)))
                   ;; use the marked revision
                   ;; (tla-revision-unmark-all)
                   (tla--revision-revision (car marked)))
                  (t
                   (tla--read-archive-category-branch-version-revision-name
                    (format "Archive for delta to %s from: " modified-fq)
                    (format "Category for delta to %s from: " modified-fq)
                    (format "Branch for delta to %s from: " modified-fq)
                    (format "Version for delta to %s from: " modified-fq)
                    (format "Revision for delta to %s from: "
                            (if across-versions
                                modified-fq
                              (tla-revision-name modified)))
                    (unless across-versions (tla-archive-name modified))
                    (unless across-versions (tla-category-name modified))
                    (unless across-versions (tla-branch-name modified))
                    (unless across-versions (tla-version-name modified))))))))

    (unless (tla-archive-name base)
      (error "Archive for the base is not specified"))
    (unless (tla-category-name base)
      (error "cateogory for the base is not specified"))
    (unless (tla-branch-name base)
      (error "branch for the base is not specified"))
    (unless (tla-version-name base)
      (error "version for the base is not specified"))
    (unless (tla-revision-name base)
      ;; No revision for modified is specified.
      ;; Use HEAD revision.
      (setcar (nthcdr 4 base)
              (tla--version-head
               (tla-archive-name base)
               (tla-category-name base)
               (tla-branch-name base)
               (tla-version-name base))))

    (when (or stored-to-directory
              (and (not stored-to-directory)
                   (y-or-n-p "Store the delta to a directory? ")))
      (setq stored-to-directory 'ask))

    (tla-delta (tla--name-construct base)
               modified-fq
               stored-to-directory)))

(defun tla-revision-bookmarks-add (name)
  (interactive "sBookmark name: ")
  (tla-bookmarks-add name
                     (list tla-buffer-archive-name
                           tla-buffer-category-name
                           tla-buffer-branch-name
                           tla-buffer-version-name
                           nil))
  (message "bookmark %s added." name))

(defun tla-revision-sync-tree (arg)
  "Unify a tree's patch log with the current revision.  With an argument, use
the latest version instead."
  (interactive "P")
  (let* ((last-inventory (tla--last-visited-inventory-buffer))
         (local-tree (or (if last-inventory
                             (with-current-buffer last-inventory
                               default-directory)
                           default-directory)))
         (current (ewoc-locate tla-revision-list-cookie)))
    (while (and current
                (not (and (eq (car (ewoc-data current))
                              'separator)
                          (eq (caddr (ewoc-data current))
                              'bookmark))))
      (setq current (ewoc-prev tla-revision-list-cookie current)))
    (when (and current
               (eq (car (ewoc-data current)) 'separator)
               (eq (caddr (ewoc-data current)) 'bookmark))
      (setq local-tree (cadddr (ewoc-data current))))
    (let ((to-tree (tla--read-directory-name "Sync with tree: " local-tree)))
      (let* ((elem (ewoc-data (ewoc-locate
                               tla-revision-list-cookie)))
             (full (tla--revision-revision (caddr elem))))
        (tla-sync-tree (tla--name-construct
                        (if arg (butlast full) full))
                       to-tree)))))

(defun tla-revision-star-merge-version ()
  "Runs star-merge for the version at point"
  (interactive)
  (tla-revision-star-merge t))

(defun tla-revision-star-merge (arg)
  "Run star-merge from the revision at point.  With an argument, merge all
missing revisions from this version."
  (interactive "P")
  (let* ((last-inventory (tla--last-visited-inventory-buffer))
         (local-tree (or (if last-inventory
                             (with-current-buffer last-inventory
                               default-directory)
                           default-directory)))
         (current (ewoc-locate tla-revision-list-cookie)))
    (while (and current
                (not (and (eq (car (ewoc-data current))
                              'separator)
                          (eq (caddr (ewoc-data current))
                              'bookmark))))
      (setq current (ewoc-prev tla-revision-list-cookie current)))
    (when (and current
               (eq (car (ewoc-data current)) 'separator)
               (eq (caddr (ewoc-data current)) 'bookmark))
      (setq local-tree (cadddr (ewoc-data current))))
    (let ((to-tree (tla--read-directory-name "Merge to tree: " local-tree)))
      (let* ((elem (ewoc-data (ewoc-locate
                               tla-revision-list-cookie)))
             (full (tla--revision-revision (caddr elem))))
        (tla-star-merge (tla--name-construct
                         (if arg (butlast full) full))
                        to-tree)))))

(defun tla-revision-replay-version ()
  "Calls `tla-revision-replay' with a prefix arg"
  (interactive)
  (tla-revision-replay t))

(defun tla-revision-replay (arg)
  "Run replay from the current location.  If there are marked revisions, these
are replayed.  Otherwise, if an argument is given, all missing
revisions from this version are replayed.  If there are no marked
revisions and no argument is given, the revision under the point
is replayed."
  (interactive "P")
  (let* ((last-inventory (tla--last-visited-inventory-buffer))
         (local-tree (or (if last-inventory
                             (with-current-buffer last-inventory
                               default-directory)
                           default-directory)))
         (current (ewoc-locate tla-revision-list-cookie)))
    (while (and current
                (not (and (eq (car (ewoc-data current))
                              'separator)
                          (eq (caddr (ewoc-data current))
                              'bookmark))))
      (setq current (ewoc-prev tla-revision-list-cookie current)))
    (when (and current
               (eq (car (ewoc-data current)) 'separator)
               (eq (caddr (ewoc-data current)) 'bookmark))
      (setq local-tree (cadddr (ewoc-data current))))
    (let ((to-tree (tla--read-directory-name "Replay to tree: " local-tree)))
      (if (tla-revision-marked-revisions)
          (let ((revisions (tla--revision-revision
                            (tla-revision-marked-revisions))))
            (tla-revision-unmark-all)
            (tla-replay (mapcar (lambda (revision)
                                  (tla--name-construct
                                   revision))
                                revisions)
                        to-tree))
        (let* ((elem (ewoc-data (ewoc-locate
                                 tla-revision-list-cookie)))
               (full (tla--revision-revision (caddr elem))))
          (tla-replay (tla--name-construct
                       (if arg (butlast full) full))
                      to-tree))))))

(defun tla-revision-mark-revision ()
  "Mark revision at point"
  (interactive)
  (let ((pos (point))
        (data (ewoc-data (ewoc-locate
                          tla-revision-list-cookie))))
    (setcar (cdr data) t)
    (ewoc-refresh tla-revision-list-cookie)
    (goto-char pos)
    (tla-revision-next)))

(defun tla-revision-marked-revisions ()
  "Return the revisions that are currently marked"
  (let ((acc '()))
    (ewoc-map #'(lambda (x) (when (and (eq (car x) 'entry-patch)
                                       (cadr x))
                              (push (caddr x) acc)))
              tla-revision-list-cookie)
    (nreverse acc)))

(defun tla-revision-unmark-revision ()
  "Unmark revision at point"
  (interactive)
  (let ((pos (point))
        (data (ewoc-data (ewoc-locate
                          tla-revision-list-cookie))))
    (setcar (cdr data) nil)
    (ewoc-refresh tla-revision-list-cookie)
    (goto-char pos)
    (tla-revision-next)))

(defun tla-revision-unmark-all ()
  "Unmark all revisions"
  (interactive)
  (let ((pos (point)))
    (ewoc-map #'(lambda (x) (when (and (eq (car x) 'entry-patch)
                                       (cadr x))
                              (setcar (cdr x) nil)))
              tla-revision-list-cookie)
    (ewoc-refresh tla-revision-list-cookie)
    (goto-char pos)))

(defun tla-revision-tag-from-head ()
  "Run tla tag from the newest revision in revision buffer."
  (interactive)
  (let* ((from (when tla-buffer-archive-name
                 (tla--name-construct tla-buffer-archive-name
                                      tla-buffer-category-name
                                      tla-buffer-branch-name
                                      tla-buffer-version-name))))
    (unless from (error "No head revision"))
    (tla--revision-tag-internal from)))

(defun tla-revision-tag-from-here ()
  "Run tla tag from the current location in revision buffer."
  (interactive)
  (let ((from (when tla-revision-list-cookie
                (let* ((elem (ewoc-data (ewoc-locate
                                         tla-revision-list-cookie))))
                  (tla--name-construct (caddr elem))))))
    (unless from (error "No revision here"))
    (tla--revision-tag-internal from)))

(defun tla--revision-tag-internal (from-fq)
  (let* ((to (tla--read-archive-category-branch-version-name "Tag to: "))
         (to-fq (tla--name-construct to)))
    (tla--version-tag-internal from-fq to-fq)))

(defun tla-revision-show-changeset ()
  (interactive)
  (let ((elem (ewoc-data (ewoc-locate
                          tla-revision-list-cookie))))
    (case (car elem)
      (entry-patch (tla-revision-cat-log))
      (entry-change (progn (cd (caddr elem)) (tla-changes))))))

(defun tla-revision-show-changeset-by-mouse (event)
  (interactive "e")
  (mouse-set-point event)
  (tla-revision-show-changeset))

(defun tla-revision-cat-log ()
  (interactive)
  (let* ((elem (ewoc-data (ewoc-locate
                           tla-revision-list-cookie)))
         (full (tla--revision-revision (caddr elem))))
    (tla-switch-to-buffer (tla--cat-log-any full))
    (tla-cat-log-mode)
    ))

(defun tla-revision-update ()
  "Runs tla update"
  (interactive)
  (let ((local-tree default-directory) ;; Default value
        (current (ewoc-locate tla-revision-list-cookie)))
    (while (and current
                (not (and (eq (car (ewoc-data current))
                              'separator)
                          (eq (caddr (ewoc-data current))
                              'bookmark))))
      (setq current (ewoc-prev tla-revision-list-cookie current)))
    (when (and current
               (eq (car (ewoc-data current)) 'separator)
               (eq (caddr (ewoc-data current)) 'bookmark))
      (setq local-tree (cadddr (ewoc-data current))))
    (let ((buffer (current-buffer)))
      (tla-update (tla--read-directory-name "Update tree: " local-tree))
      (pop-to-buffer buffer))
    (tla-generic-refresh)))

;; ----------------------------------------------------------------------------
;; tla-changes-mode
;; ----------------------------------------------------------------------------
(define-derived-mode tla-changes-mode fundamental-mode "tla-changes"
  "Major mode to display changesets. Derives from `diff-mode'.

Use '\\<tla-changes-mode-map>\\[tla-changes-mark-file]' to mark files, and '\\[tla-changes-unmark-file]' to unmark.
If you commit from this buffer (with '\\[tla-changes-edit-log]'), then, the list of selected
files in this buffer at the time you actually commit with
\\<tla-log-edit-mode-map>\\[tla-log-edit-done].

Commands:
\\{tla-changes-mode-map}
"
  (let ((diff-mode-shared-map (copy-keymap tla-changes-mode-map))
        major-mode mode-name)
    (diff-mode))

  (set (make-local-variable 'font-lock-defaults)
       (list 'tla-changes-font-lock-keywords t nil nil))
  (set (make-local-variable 'tla-get-file-info-at-point-function)
       'tla-changes-get-file-at-point)
  (set (make-local-variable 'tla-buffer-refresh-function)
       'tla-changes-generic-refresh)
  (set (make-local-variable 'tla-changes-cookie)
       (ewoc-create 'tla-changes-printer))
  (make-local-variable 'tla-buffer-marked-file-list)
  (easy-menu-add tla-changes-mode-menu)
  (tla--install-buffer-menu)
  (toggle-read-only 1)
  (set-buffer-modified-p nil))

(defun tla-changes-generic-refresh ()
  (interactive)
  (if (eq (car tla--changes-modified) 'local-tree)
      (tla-changes tla--changes-summary tla--changes-base)))

(defun tla-changes-jump-to-change (&optional other-file)
  "Jump to the corresponding file and location of the change."
  (interactive "P")
  (let* ((elem (ewoc-locate tla-changes-cookie))
         (data (ewoc-data elem)))
    (cond ((< (ewoc-location elem) (line-beginning-position))
           (diff-goto-source other-file))
          ((eq (car data) 'file)
           (find-file (cadr data)))
          ((eq (car data) 'subtree)
           (tla-switch-to-buffer (cadr data)))
          (t (error "not on a recognized location")))))

(defun tla-changes-diff-or-list ()
  "When in the list part of the buffer, jump to the corresponding
patch. When on a patch, jump to the corresponding entry in the list of
files."
  (interactive)
  (let* ((elem (ewoc-locate tla-changes-cookie))
         (data (ewoc-data elem)))
    (cond ((< (ewoc-location elem) (line-beginning-position))
           (let ((file (tla-changes-get-file-at-point))
                 (elem (ewoc-nth tla-changes-cookie 0)))
             (while (and elem
                         (or (not (eq (car (ewoc-data elem)) 'file))
                             (not (equal (expand-file-name
                                          (cadr (ewoc-data elem)))
                                         file))))
               (setq elem (ewoc-next tla-changes-cookie elem)))
             (if elem (goto-char (ewoc-location elem))
               (error "Can't find file in list"))
             ))
          ((eq (car data) 'file)
           (re-search-forward (concat "^--- orig/" (cadr data)))
           (diff-hunk-next))
          ((eq (car data) 'subtree)
           (tla-switch-to-buffer (cadr data)))
          (t (error "not on a recognized location")))))

(defun tla-changes-master-buffer ()
  "Jumps to the master *tla-changes* buffer for a nested changes
buffer"
  (interactive)
  (unless tla--changes-buffer-master-buffer
    (error "no master buffer"))
  (tla-switch-to-buffer tla--changes-buffer-master-buffer))

(defun tla-flash-line-on ()
  (cond
   ((functionp 'hl-line-mode)
    (hl-line-mode))
   ((functionp 'highline-on)
    (highline-on))))

(defun tla-flash-line-off ()
  (cond
   ((functionp 'hl-line-mode)
    (hl-line-mode))
   ((functionp 'highline-off)
    (highline-off))))

(defun tla-flash-line ()
  (tla-flash-line-on)
  (sit-for 1000)
  (tla-flash-line-off))

(defun tla-changes-view-source (&optional other-file)
  "Show the corresponding file and location of the change.
This function does not switch to the file, but it places the cursor
temporarily at the location of the change and will stay in the changes
buffer.  Thus you can quickly see more context on a specific change without
switching buffers."
  (interactive "P")
  (let ((diff-window (selected-window)))
    (save-excursion
      (diff-goto-source other-file)
      (recenter)
      (tla-flash-line)
      (select-window diff-window))))

(defun tla-changes-edit-log (&optional insert-changelog)
  "Wrapper around `tla-edit-log', setting the source buffer to current
buffer."
  (interactive "P")
  (tla-edit-log insert-changelog (current-buffer)))

(defun tla-changes-rm ()
  (interactive)
  (let ((file (tla--get-file-info-at-point)))
    (unless file
      (error "No file at point"))
    (tla-rm file)))

(defun tla-changes-mark-file ()
  (interactive)
  (let ((current (ewoc-locate tla-changes-cookie))
        (file (tla--get-file-info-at-point)))
    (add-to-list 'tla-buffer-marked-file-list file)
    (ewoc-refresh tla-changes-cookie)
    (goto-char (ewoc-location (or (ewoc-next tla-changes-cookie
                                             current)
                                  current)))))

(defun tla-changes-unmark-file ()
  (interactive)
  (let ((current (ewoc-locate tla-changes-cookie))
        (file (tla--get-file-info-at-point)))
    (setq tla-buffer-marked-file-list
          (delete file tla-buffer-marked-file-list))
    (ewoc-refresh tla-changes-cookie)
    (goto-char (ewoc-location (or (ewoc-next tla-changes-cookie
                                             current)
                                  current)))))

(defun tla-changes-diff ()
  "Runs tla file-diff on the file at point in *tla-changes*"
  (interactive)
  (let ((on-modified-file (tla--get-file-info-at-point)))
    (if on-modified-file
        (tla-file-diff on-modified-file)
      (error "not on a modified file"))))

(defun tla-changes-next ()
  (interactive)
  (let ((cur-location (ewoc-location (ewoc-locate tla-changes-cookie)))
        (next (ewoc-next tla-changes-cookie
                         (ewoc-locate tla-changes-cookie))))
    (cond
     ((> cur-location (point))
      (goto-char cur-location))
     (next
      (goto-char (ewoc-location next)))
     (t
      (diff-hunk-next)))))

(defun tla-changes-prev ()
  (interactive)
  (let* ((current (ewoc-locate tla-changes-cookie))
         (cur-location (ewoc-location current))
         (prev (ewoc-prev tla-changes-cookie current))
         (next (ewoc-next tla-changes-cookie current)))
    (cond (next
           (if prev (goto-char (ewoc-location prev))
             (goto-char cur-location)))
          ((condition-case nil (progn (diff-hunk-prev) t) (error nil)))
          ((> (line-beginning-position) cur-location)
           (goto-char cur-location))
          (prev
           (goto-char (ewoc-location prev)))
          (t
           (goto-char cur-location)))
    ))

(defun tla--changes-in-diff ()
  "Returns t if cursor is in the diffs section of the changes buffer"
  (save-excursion (re-search-backward "^--- orig" nil t)))


(defun tla-changes-ediff (&optional other-file)
  (interactive "P")
  (unless (and (car tla--changes-base)
               (car tla--changes-base))
    (error "No revision information to base ediff on."))
  (let ((on-modified-file (tla--get-file-info-at-point))
        (loc (point)))
    (if (and on-modified-file (not (tla--changes-in-diff)))
        (tla-file-ediff-revisions on-modified-file
                                  tla--changes-base
                                  tla--changes-modified)
      (re-search-backward "^--- orig/")
      (re-search-forward "^--- orig/")
      (let ((file (expand-file-name
                   (concat default-directory "/"
                           (buffer-substring-no-properties (point)
                                                           (line-end-position)))))
            (hunk 1))
        (diff-hunk-next)
        (while (<= (re-search-forward "\\(^[\\+-].*\n\\)+" nil t) loc)
          (setq hunk (1+ hunk)))
        (goto-char loc)
        (with-current-buffer
            (tla-file-ediff-revisions file tla--changes-base
                                      tla--changes-modified)
          (ediff-jump-to-difference hunk))))))

(defun tla-changes-get-file-at-point ()
  "Find file at point in *tla-changes*. Error when not on a file."
  (let ((elem (ewoc-locate tla-changes-cookie (point))))
    (or (when (and elem
                   (eq (car (ewoc-data elem)) 'file)
                   (>= (ewoc-location elem) (line-beginning-position)))
          (cadr (ewoc-data elem)))
        (let ((loc (diff-find-source-location)))
          (buffer-file-name (car loc))))))

(defun tla-changes-revert ()
  "Reverts file at point"
  (interactive)
  (let* ((file (tla--get-file-info-at-point))
         (absolute (if (file-name-absolute-p file)
                       file
                     (expand-file-name
                      (concat default-directory "/" file)))))
    (tla-file-revert absolute)))

;; ----------------------------------------------------------------------------
;; tla-changelog-mode
;; ----------------------------------------------------------------------------

(define-derived-mode tla-changelog-mode change-log-mode "tla-changelog"
  (set (make-local-variable 'font-lock-defaults)
       (list 'tla-changelog-font-lock-keywords
             t nil nil 'backward-paragraph))
  (use-local-map tla-changelog-mode-map)
  (toggle-read-only 1)
  (set-buffer-modified-p nil))

;; ----------------------------------------------------------------------------
;; tla-inventory-file-mode
;; ----------------------------------------------------------------------------
;;;###autoload
(defun tla-inventory-file-mode ()
  "Major Mode to edit tla inventory files (=tagging-method, .arch-inventory)."
  (interactive)
  (kill-all-local-variables)
  (set (make-local-variable 'font-lock-defaults)
       '(tla-inventory-file-font-lock-keywords t))
  (setq major-mode 'tla-inventory-file-mode
        mode-name "tla-inventory-file"
        comment-start "# ")
  (run-hooks 'tla-inventory-file-mode-hook))

(defun tla--inventory-file-jump-from-head (category)
  "Search CATEGORY from the head of the buffer."
  (let ((p (save-excursion (goto-char (point-min))
                           (re-search-forward
                            (concat "^" category) nil t))))
    (when p
      (goto-char p))))

(defun tla--inventory-file-jump-from-tail (category)
  "Search CATEGORY from the tail of the buffer."
  (let ((p (save-excursion (goto-char (point-max))
                           (re-search-backward
                            (concat "^" category) nil t))))
    (when p
      (goto-char p))))

(defun tla--inventory-file-add-file (category file)
  "Added TARGET to CATEGORY."
  (tla--inventory-file-jump-from-tail category)
  (save-excursion (open-line 1))
  (insert (format "%s ^(%s)$" category file)))

;; ----------------------------------------------------------------------------
;; Find file hook
;; ----------------------------------------------------------------------------
;; just 99% cut&paste from vc-follow-link in vc-hook.el, but this way there is
;; no need to load it thus avoiding interfering with VC ...
(defun tla-follow-link ()
  ;; If the current buffer visits a symbolic link, this function makes it
  ;; visit the real file instead.  If the real file is already visited in
  ;; another buffer, make that buffer current, and kill the buffer
  ;; that visits the link.
  (let* ((truename (abbreviate-file-name (file-truename buffer-file-name)))
         (true-buffer (find-buffer-visiting truename))
         (this-buffer (current-buffer)))
    (if (eq true-buffer this-buffer)
        (progn
          (kill-buffer this-buffer)
          ;; In principle, we could do something like set-visited-file-name.
          ;; However, it can't be exactly the same as set-visited-file-name.
          ;; I'm not going to work out the details right now. -- rms.
          (set-buffer (find-file-noselect truename)))
      (set-buffer true-buffer)
      (kill-buffer this-buffer))))

;;;###autoload
(defun tla-find-file-hook ()
  "Follow symlinked files/directories to the actual location of a file."
  (let (link file result)
    (when (and (if (boundp 'vc-ignore-vc-files)
                   (not vc-ignore-vc-files)
                 t)
               tla-follow-symlinks
               (setq file buffer-file-name)
               (not (string= (setq link (file-truename file)) file)))
      (setq file link
            result (cond ((equal tla-follow-symlinks 'tree)
                          (tla-tree-root file t))
                         ((equal tla-follow-symlinks 'id)
                          (= 0 (tla--run-tla-sync
                                (list "id" file)
                                :finished 'tla--status-handler
                                :error 'tla--status-handler)))))

      (if result
          (cond ((eq tla-follow-symlinks-mode 'warn)
                 (message
                  "Warning: symbolic link to arch-controlled source file: %s"
                  file))
                ((or (eq tla-follow-symlinks-mode 'follow)
                     (find-buffer-visiting file))
                 (tla-follow-link)
                 (message "Followed link to arch-controlled %s"
                          buffer-file-name))
                ((eq tla-follow-symlinks-mode 'ask)
                 (if (yes-or-no-p "Symbolic link to arch-controlled source file ; follow link? ")
                     (progn
                       (tla-follow-link)
                       (message "Followed link to arch-controlled %s"
                                buffer-file-name))
                   (message
                    "Warning: editing through the link bypasses version control")))
                (t (error "unknown mode for tla-follow-symlinks-mode=%s"
                          tla-follow-symlinks-mode)))
        ))))

;; ----------------------------------------------------------------------------
;; Misc functions
;; ----------------------------------------------------------------------------

;;;###autoload
(defun tla-insert-arch-tag ()
  "Insert a unique arch-tag in the current file."
  (interactive)
  (let ((uuid (shell-command-to-string "uuidgen"))
        (in-comment-p (nth 4 (parse-partial-sexp (point) (point-min))))
        (header "")
        (footer ""))
    (unless in-comment-p
      (setq header (if comment-start
                       (concat comment-start
                               (if (string-match " $" comment-start)
                                   "" " "))
                     "")
            footer (if (and comment-end (not (string= "" comment-end)))
                       (format "%s(do not change this comment)%s%s"
                               (make-string (length header) ?\ )
                               comment-end
                               (if (string-match "^ " comment-end)
                                   "" " "))
                     "")))
    (insert (concat header "arch-tag: " uuid footer))))

;;;###autoload
(defun tla-ediff-add-log-entry ()
  (interactive)
  (pop-to-buffer ediff-buffer-A)
  (tla-add-log-entry))

;;
;; Tree-lint mode
;;
(defvar tla--tree-lint-cookie nil
  "Ewoc cookie used in tree-lint mode.")

(define-derived-mode tla-tree-lint-mode fundamental-mode
  "tla-tree-lint"
  "Major mode to view tree-lint warnings.
Commands:
\\{tla-tree-lint-mode-map}
"
  (tla--install-buffer-menu)
  (let ((inhibit-read-only t))
    (erase-buffer))
  (set (make-local-variable 'tla-buffer-refresh-function)
       `(lambda () (interactive) (tla-tree-lint ,default-directory)))
  (set (make-local-variable 'tla--tree-lint-cookie)
       (ewoc-create 'tla--tree-lint-printer))
  (toggle-read-only t))

(defun tla--tree-lint-prepare-buffer (root)
  "Prepares the buffer to display the tree-lint warnings"
  (let ((buffer (tla--get-buffer-create 'tree-lint root)))
    (with-current-buffer buffer
      (tla-tree-lint-mode)
      (ewoc-enter-last
       tla--tree-lint-cookie
       (list 'message (format "Running tree-lint in %s ..."
                              root)))
      buffer)))

;;;###autoload
(defun tla-tree-lint (root)
  "Runs tla tree-lint in directory ROOT"
  (interactive
   (list (tla--read-project-tree-maybe "Run tla tree-lint in: ")))
  (let ((default-directory root)
        (buffer (tla--tree-lint-prepare-buffer root)))
    (when tla-switch-to-buffer-first
      (tla-switch-to-buffer buffer))
    (tla--run-tla-async
     '("tree-lint")
     :related-buffer buffer
     :finished
     `(lambda (output error status arguments)
        (message "No tree-lint warnings for %s." ,default-directory)
        (with-current-buffer ,buffer
          (let ((inhibit-read-only t))
            (erase-buffer)
            (ewoc-enter-last
             tla--tree-lint-cookie
             (list 'message (format "No tree-lint warnings for %s."
                                    ,default-directory))))))
     :error
     `(lambda (output error status arguments)
        (save-excursion
          (tla--tree-lint-parse-buffer output ,buffer))))))

(defconst tla--tree-lint-message-alist
  '(("^These files would be source but lack inventory ids"
     missing-file)
    ("^These explicit ids have no corresponding file:"
     id-without-file)
    ("^These files violate naming conventions:"
     unrecognized)
    ("^These symlinks point to nonexistent files:"
     broken-link)
    ))

(defun tla--tree-lint-message-type (message)
  "Returns a symbol saying which type of message the string MESSAGE
is."
  (let ((result nil)
        (iterator tla--tree-lint-message-alist))
    (while (and iterator (not result))
      (when (string-match (caar iterator) message)
        (setq result (cadar iterator)))
      (setq iterator (cdr iterator)))
    (or result 'unknown)))

(defun tla--tree-lint-parse-buffer (buffer output-buffer)
  "Parses the output of tla tree-lint in BUFFER to show in in the
tree-lint-mode buffer OUTPUT-BUFFER."
  (with-current-buffer output-buffer
    (let ((inhibit-read-only t))
      (erase-buffer)
      (insert (tla--face-add (format "Tree lint warnings in %s\n"
                                     default-directory)
                             'tla-message)))
    (setq tla--tree-lint-cookie
          (ewoc-create 'tla--tree-lint-printer)))
  (with-current-buffer buffer
    (goto-char (point-min))
    (let ((cookie (with-current-buffer output-buffer
                    tla--tree-lint-cookie)))
      (while (re-search-forward "^." nil t)
        (goto-char (line-beginning-position))
        (let* ((message (buffer-substring-no-properties
                         (point) (line-end-position)))
               (type (tla--tree-lint-message-type message)))
          (ewoc-enter-last cookie (list 'message message))
          (forward-line 2)
          (while (not (eq (char-after) ?\n))
            (ewoc-enter-last cookie
                             (list type (buffer-substring-no-properties
                                         (point)
                                         (line-end-position))))
            (forward-line 1))))
      (let ((inhibit-read-only t))
        (ewoc-refresh cookie))))
  (with-current-buffer output-buffer
    (tla--tree-lint-cursor-goto
     (ewoc-nth tla--tree-lint-cookie 0))))

(defun tla--tree-lint-printer (elem)
  "Ewoc printer for the tree-lint buffer"
  (case (car elem)
    (message (insert "\n" (tla--face-add (cadr elem) 'tla-messages) "\n"))
    (missing-file (tla--face-add (insert (cadr elem)) 'tla-to-add))
    (id-without-file (tla--face-add (insert (cadr elem)) 'tla-to-add))
    (unrecognized (insert (tla--face-add (cadr elem)
                                         'tla-unrecognized)))
    (broken-link (insert (tla--face-add (cadr elem)
                                        'tla-broken-link)))
    (unknown (insert (tla--face-add (cadr elem)
                                    'tla-unrecognized)))
    (t (error "unimplemented type of tree-lint error")))
  )

(defun tla--tree-lint-cursor-goto (ewoc-tree-lint)
  "Move cursor to the ewoc location of EWOC-TREE-LINT"
  (interactive)
  (if ewoc-tree-lint
      (progn (goto-char (ewoc-location ewoc-tree-lint))
             (re-search-forward "." nil t)
             (backward-char 1))
    (goto-char (point-min))))

(defun tla-tree-lint-next ()
  (interactive)
  (let* ((cookie tla--tree-lint-cookie)
         (elem (ewoc-locate cookie))
         (next (or (ewoc-next cookie elem) elem)))
    (tla--tree-lint-cursor-goto next)))

(defun tla-tree-lint-previous ()
  (interactive)
  (let* ((cookie tla--tree-lint-cookie)
         (elem (ewoc-locate cookie))
         (previous (or (ewoc-prev cookie elem) elem)))
    (tla--tree-lint-cursor-goto previous)))

(defun tla--tree-lint-select-files (msg-singular msg-plural msg-err
                                                 msg-prompt)
  "Get the list of files under cursor, and ask confirmation to the
user."
  (let* ((ewoc-elem (ewoc-locate tla--tree-lint-cookie))
         (elem (ewoc-data ewoc-elem)))
    (if (eq (car elem) 'message)
        (let ((list nil))
          (setq ewoc-elem
                (ewoc-next tla--tree-lint-cookie ewoc-elem))
          (setq elem (and ewoc-elem (ewoc-data ewoc-elem)))
          (while (and ewoc-elem (not (eq (car elem) 'message)))
            (add-to-list 'list (cadr elem))
            (setq ewoc-elem
                  (ewoc-next tla--tree-lint-cookie ewoc-elem))
            (setq elem (and ewoc-elem (ewoc-data ewoc-elem))))
          (progn
            (unless (y-or-n-p (if (eq 1 (length list))
                                  (format msg-singular
                                          (car list))
                                (format msg-plural
                                        (length list))))
              (error msg-err))
            list))
      (list (read-file-name msg-prompt default-directory
                            nil nil
                            (cadr elem))))))

(defun tla-tree-lint-add-files (files)
  "Prompts and add the file at point. If on a message field, add all
the files below this message."
  (interactive
   (list
    (tla--tree-lint-select-files "Add %s? "
                                 "Add %s files? "
                                 "Not adding any file"
                                 "Add file: ")))
  (apply 'tla-add nil files)
  (tla-tree-lint default-directory))

(defun tla-tree-lint-delete-files (files)
  "Prompts and delete the file at point. If on a message field, delete
all the files below this message."
  (interactive
   (list
    (tla--tree-lint-select-files "Delete %s? "
                                 "Delete %s files? "
                                 "Not deleting any file"
                                 "Delete file: ")))
  (mapcar 'delete-file files)
  (tla-tree-lint default-directory))

;; end tree-lint-mode

(defvar tla-arch-version nil
  "Version of tla version.")

(defun tla-arch-version ()
  "Returns the TLA (arch) version."
  (interactive)
  (setq tla-arch-version
        (tla--run-tla-sync '("-V")
                           :finished
                           (lambda (output error status arguments)
                             (tla--buffer-content output))))
  (if (interactive-p)
      (message tla-arch-version))
  tla-arch-version)

(defun tla-version ()
  "Returns the XTLA version."
  (interactive)
  (let ((version
         (or (when (locate-library "xtla-version")
               (load-library "xtla-version")
               (when (boundp 'tla-version)
                 tla-version))
             (let ((default-directory
                     (file-name-directory (locate-library "xtla"))))
               (defvar tla-version nil "Version of xtla")
               (tla--run-tla-sync '("logs" "-f" "-r")
                                  :finished
                                  (lambda (output error status arguments)
                                    (set-buffer output)
                                    (goto-char (point-min))
                                    (setq tla-version
                                          (buffer-substring-no-properties
                                           (point)
                                           (line-end-position))))
                                  :error
                                  (lambda (output error status arguments)
                                    (setq tla-version "unknown")))))))
    (if (not version)
        (progn
          (message "We did not find xtla-version.el nor the arch-tree containing xtla.el!")
          (sit-for 2)
          (message "Are you using a developer version of XTLA?")
          (sit-for 2))
      (if (interactive-p)
          (message tla-version))
      tla-version)))

;;;###autoload
(defun tla-submit-bug-report ()
  "Submit a bug report, with pertinent information to the xtla-el-dev list."
  (interactive)
  (require 'reporter)
  (delete-other-windows)
  (tla-version)
  (tla-arch-version)
  (reporter-submit-bug-report
   "xtla-el-dev@gna.org"
   (concat "XTLA " tla-version)
   (append
    ;; non user variables
    '(emacs-version
      tla-version
      tla-arch-version
      )
    ;; user variables
    (sort (apropos-internal "^tla-" 'user-variable-p)
          (lambda (v1 v2) (string-lessp (format "%s" v1) (format "%s" v2))))
    ;; see what the user had loaded
    (list 'features)
    )
   nil
   nil
   (concat
    "Please change the Subject header to a concise bug description or feature request.\n"
    "In this report, remember to cover the basics, that is, what you \n"
    "expected to happen and what in fact did happen.\n"
    "Please remove these instructions from your message."))

  ;; insert the backtrace buffer content if present
  (let ((backtrace (get-buffer "*Backtrace*")))
    (when backtrace
      (goto-char (point-max))
      (insert "\n\n")
      (insert-buffer-substring backtrace)))

  (goto-char (point-min))
  (mail-position-on-field "Subject")
  (insert "[BUG/FEATURE] "))

(provide 'xtla)

;; Local Variables:
;; arch-tag: f2eee8c5-0f20-4fc7-b1c1-6cef4dff8a5a
;; End:
